@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 1999-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 1999-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -25,19 +25,25 @@
 
 #include "os-interface.h"
 #include "nv-linux.h"
+#include "nv-caps-imex.h"
 
 #include "nv-time.h"
 
 #include <linux/mmzone.h>
 #include <linux/numa.h>
+#include <linux/cpuset.h>
 
 #include <linux/pid.h>
+#include <linux/pid_namespace.h>
+#if defined(CONFIG_LOCKDEP)
+#include <linux/lockdep.h>
+#endif // CONFIG_LOCKDEP
 
 extern char *NVreg_TemporaryFilePath;
 
-#define MAX_ERROR_STRING 512
+#define MAX_ERROR_STRING 528
 static char nv_error_string[MAX_ERROR_STRING];
-nv_spinlock_t nv_error_string_lock;
+static NV_DEFINE_SPINLOCK(nv_error_string_lock);
 
 extern nv_linux_state_t nv_ctl_device;
 
@@ -47,6 +53,8 @@ NvU32 os_page_size  = PAGE_SIZE;
 NvU64 os_page_mask  = NV_PAGE_MASK;
 NvU8  os_page_shift = PAGE_SHIFT;
 NvBool os_cc_enabled = 0;
+NvBool os_cc_sev_snp_enabled = 0;
+NvBool os_cc_snp_vtom_enabled = 0;
 NvBool os_cc_tdx_enabled = 0;
 
 #if defined(CONFIG_DMA_SHARED_BUFFER)
@@ -55,6 +63,8 @@ NvBool os_dma_buf_enabled = NV_TRUE;
 NvBool os_dma_buf_enabled = NV_FALSE;
 #endif // CONFIG_DMA_SHARED_BUFFER
 
+NvBool os_imex_channel_is_supported = NV_TRUE;
+
 void NV_API_CALL os_disable_console_access(void)
 {
     console_lock();
@@ -238,7 +248,20 @@ NV_STATUS NV_API_CALL os_release_semaphore
     return NV_OK;
 }
 
-typedef struct rw_semaphore os_rwlock_t;
+typedef struct
+{
+    struct rw_semaphore sem;
+
+#if defined(CONFIG_LOCKDEP)
+    /**
+     * A key of lock class. It would be registered to Lockdep validator so all
+     * instances' usages and dependencies will contribute to constructing correct
+     * locking rules and this lock will be tracked by the Lockdep validator.
+     *
+     */
+    struct lock_class_key key;
+#endif // CONFIG_LOCKDEP
+} os_rwlock_t;
 
 void* NV_API_CALL os_alloc_rwlock(void)
 {
@@ -247,11 +270,17 @@ void* NV_API_CALL os_alloc_rwlock(void)
     NV_STATUS rmStatus = os_alloc_mem((void *)&os_rwlock, sizeof(os_rwlock_t));
     if (rmStatus != NV_OK)
     {
-        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate rw_semaphore!\n");
+        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate a struct os_rwlock_t!\n");
         return NULL;
     }
 
-    init_rwsem(os_rwlock);
+    init_rwsem(&os_rwlock->sem);
+
+#if defined(CONFIG_LOCKDEP)
+    // Register the dynamically allocated key to Lockdep.
+    lockdep_register_key(&os_rwlock->key);
+    lockdep_set_class(&os_rwlock->sem, &os_rwlock->key);
+#endif // CONFIG_LOCKDEP
 
     return os_rwlock;
 }
@@ -259,6 +288,12 @@ void* NV_API_CALL os_alloc_rwlock(void)
 void NV_API_CALL os_free_rwlock(void *pRwLock)
 {
     os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
+
+#if defined(CONFIG_LOCKDEP)
+    // Unregister the dynamically allocated key.
+    lockdep_unregister_key(&os_rwlock->key);
+#endif // CONFIG_LOCKDEP
+
     os_free_mem(os_rwlock);
 }
 
@@ -270,7 +305,7 @@ NV_STATUS NV_API_CALL os_acquire_rwlock_read(void *pRwLock)
     {
         return NV_ERR_INVALID_REQUEST;
     }
-    down_read(os_rwlock);
+    down_read(&os_rwlock->sem);
     return NV_OK;
 }
 
@@ -282,7 +317,7 @@ NV_STATUS NV_API_CALL os_acquire_rwlock_write(void *pRwLock)
     {
         return NV_ERR_INVALID_REQUEST;
     }
-    down_write(os_rwlock);
+    down_write(&os_rwlock->sem);
     return NV_OK;
 }
 
@@ -290,7 +325,7 @@ NV_STATUS NV_API_CALL os_cond_acquire_rwlock_read(void *pRwLock)
 {
     os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
 
-    if (down_read_trylock(os_rwlock))
+    if (down_read_trylock(&os_rwlock->sem))
     {
         return NV_ERR_TIMEOUT_RETRY;
     }
@@ -302,7 +337,7 @@ NV_STATUS NV_API_CALL os_cond_acquire_rwlock_write(void *pRwLock)
 {
     os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
 
-    if (down_write_trylock(os_rwlock))
+    if (down_write_trylock(&os_rwlock->sem))
     {
         return NV_ERR_TIMEOUT_RETRY;
     }
@@ -313,13 +348,13 @@ NV_STATUS NV_API_CALL os_cond_acquire_rwlock_write(void *pRwLock)
 void NV_API_CALL os_release_rwlock_read(void *pRwLock)
 {
     os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
-    up_read(os_rwlock);
+    up_read(&os_rwlock->sem);
 }
 
 void NV_API_CALL os_release_rwlock_write(void *pRwLock)
 {
     os_rwlock_t *os_rwlock = (os_rwlock_t *)pRwLock;
-    up_write(os_rwlock);
+    up_write(&os_rwlock->sem);
 }
 
 NvBool NV_API_CALL os_semaphore_may_sleep(void)
@@ -338,14 +373,27 @@ NvBool NV_API_CALL os_is_administrator(void)
     return NV_IS_SUSER();
 }
 
-NvBool NV_API_CALL os_allow_priority_override(void)
+NvBool NV_API_CALL os_check_access(RsAccessRight accessRight)
 {
-    return capable(CAP_SYS_NICE);
-}
-
-NvU64 NV_API_CALL os_get_num_phys_pages(void)
-{
-    return (NvU64)NV_NUM_PHYSPAGES;
+    switch (accessRight)
+    {
+        case RS_ACCESS_PERFMON:
+        {
+#if defined(CAP_PERFMON)
+            return capable(CAP_PERFMON);
+#else
+            return os_is_administrator();
+#endif
+        }
+        case RS_ACCESS_NICE:
+        {
+            return capable(CAP_SYS_NICE);
+        }
+        default:
+        {
+            return NV_FALSE;
+        }
+    }
 }
 
 char* NV_API_CALL os_string_copy(
@@ -373,7 +421,7 @@ NvS32 NV_API_CALL os_string_compare(const char *str1, const char *str2)
     return strcmp(str1, str2);
 }
 
-void *os_mem_copy_custom(
+static void *os_mem_copy_custom(
     void       *dstPtr,
     const void *srcPtr,
     NvU32       length
@@ -794,7 +842,7 @@ int NV_API_CALL nv_printf(NvU32 debuglevel, const char *printf_format, ...)
     if (debuglevel >= ((cur_debuglevel >> 4) & 0x3))
     {
         size_t length;
-        char *temp;
+        unsigned long flags;
 
         // When printk is called to extend the output of the previous line
         // (i.e. when the previous line did not end in \n), the printk call
@@ -818,20 +866,18 @@ int NV_API_CALL nv_printf(NvU32 debuglevel, const char *printf_format, ...)
         if (length < 1)
             return 0;
 
-        temp = kmalloc(length + sizeof(KERN_CONT), GFP_ATOMIC);
-        if (!temp)
-            return 0;
+        NV_SPIN_LOCK_IRQSAVE(&nv_error_string_lock, flags);
 
         // KERN_CONT changed in the 3.6 kernel, so we can't assume its
         // composition or size.
-        memcpy(temp, KERN_CONT, sizeof(KERN_CONT) - 1);
-        memcpy(temp + sizeof(KERN_CONT) - 1, printf_format, length + 1);
+        memcpy(nv_error_string, KERN_CONT, sizeof(KERN_CONT) - 1);
+        memcpy(nv_error_string + sizeof(KERN_CONT) - 1, printf_format, length + 1);
 
         va_start(arglist, printf_format);
-        chars_written = vprintk(temp, arglist);
+        chars_written = vprintk(nv_error_string, arglist);
         va_end(arglist);
 
-        kfree(temp);
+        NV_SPIN_UNLOCK_IRQRESTORE(&nv_error_string_lock, flags);
     }
 
     return chars_written;
@@ -983,26 +1029,29 @@ void NV_API_CALL os_unmap_kernel_space(
     nv_iounmap(addr, size_bytes);
 }
 
-// flush the cpu's cache, uni-processor version
-NV_STATUS NV_API_CALL os_flush_cpu_cache(void)
+#if NVCPU_IS_AARCH64
+
+static inline void nv_flush_cache_cpu(void *info)
 {
-    CACHE_FLUSH();
-    return NV_OK;
+    if (!nvos_is_chipset_io_coherent())
+    {
+#if defined(NV_FLUSH_CACHE_ALL_PRESENT)
+        flush_cache_all();
+#else
+        WARN_ONCE(0, "kernel does not provide flush_cache_all()\n");
+#endif
+    }
 }
 
 // flush the cache of all cpus
 NV_STATUS NV_API_CALL os_flush_cpu_cache_all(void)
 {
-#if defined(NVCPU_AARCH64)
-    CACHE_FLUSH_ALL();
+    on_each_cpu(nv_flush_cache_cpu, NULL, 1);
     return NV_OK;
-#endif
-    return NV_ERR_NOT_SUPPORTED;
 }
 
 NV_STATUS NV_API_CALL os_flush_user_cache(void)
 {
-#if defined(NVCPU_AARCH64)
     if (!NV_MAY_SLEEP())
     {
         return NV_ERR_NOT_SUPPORTED;
@@ -1013,16 +1062,35 @@ NV_STATUS NV_API_CALL os_flush_user_cache(void)
     // although it is possible. For now, just flush the entire cache to be
     // safe.
     //
-    CACHE_FLUSH_ALL();
+    on_each_cpu(nv_flush_cache_cpu, NULL, 1);
     return NV_OK;
-#else
-    return NV_ERR_NOT_SUPPORTED;
-#endif
 }
 
+#else // NVCPU_IS_AARCH64
+
+NV_STATUS NV_API_CALL os_flush_cpu_cache_all(void)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
+
+NV_STATUS NV_API_CALL os_flush_user_cache(void)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
+
+#endif
+
 void NV_API_CALL os_flush_cpu_write_combine_buffer(void)
 {
-    WRITE_COMBINE_FLUSH();
+#if defined(NVCPU_X86_64)
+    asm volatile("sfence" ::: "memory");
+#elif defined(NVCPU_PPC64LE)
+    __asm__ __volatile__ ("sync" : : : "memory");
+#elif defined(NVCPU_AARCH64)
+    asm volatile("dsb st" : : : "memory");
+#else
+    mb();
+#endif
 }
 
 // override initial debug level from registry
@@ -1031,8 +1099,6 @@ void NV_API_CALL os_dbg_init(void)
     NvU32 new_debuglevel;
     nvidia_stack_t *sp = NULL;
 
-    NV_SPIN_LOCK_INIT(&nv_error_string_lock);
-
     if (nv_kmem_cache_alloc_stack(&sp) != 0)
     {
         return;
@@ -1058,7 +1124,7 @@ void NV_API_CALL os_dbg_set_level(NvU32 new_debuglevel)
 
 NvU64 NV_API_CALL os_get_max_user_va(void)
 {
-	return TASK_SIZE;
+    return TASK_SIZE;
 }
 
 NV_STATUS NV_API_CALL os_schedule(void)
@@ -1128,7 +1194,7 @@ NV_STATUS NV_API_CALL os_queue_work_item(struct os_work_queue *queue, void *data
     return NV_OK;
 }
 
-NV_STATUS NV_API_CALL os_flush_work_queue(struct os_work_queue *queue)
+NV_STATUS NV_API_CALL os_flush_work_queue(struct os_work_queue *queue, NvBool is_unload)
 {
     nv_kthread_q_t *kthread;
 
@@ -1137,9 +1203,13 @@ NV_STATUS NV_API_CALL os_flush_work_queue(struct os_work_queue *queue)
 
     if (NV_MAY_SLEEP())
     {
+        kthread->is_unload_flush_ongoing = is_unload;
+
         if (kthread->q_kthread)
             nv_kthread_q_flush(kthread);
 
+        kthread->is_unload_flush_ongoing = NV_FALSE;
+
         return NV_OK;
     }
     else
@@ -1151,6 +1221,13 @@ NV_STATUS NV_API_CALL os_flush_work_queue(struct os_work_queue *queue)
     }
 }
 
+NvBool NV_API_CALL os_is_queue_flush_ongoing(struct os_work_queue *queue)
+{
+    nv_kthread_q_t *kthread = queue ? &queue->nvk : &nv_kthread_q;
+
+    return kthread->is_unload_flush_ongoing;
+}
+
 extern NvU32 NVreg_EnableDbgBreakpoint;
 
 void NV_API_CALL os_dbg_breakpoint(void)
@@ -1197,90 +1274,6 @@ NvBool NV_API_CALL os_is_efi_enabled(void)
     return efi_enabled(EFI_BOOT);
 }
 
-void NV_API_CALL os_get_screen_info(
-    NvU64 *pPhysicalAddress,
-    NvU16 *pFbWidth,
-    NvU16 *pFbHeight,
-    NvU16 *pFbDepth,
-    NvU16 *pFbPitch,
-    NvU64 consoleBar1Address,
-    NvU64 consoleBar2Address
-)
-{
-    *pPhysicalAddress = 0;
-    *pFbWidth = *pFbHeight = *pFbDepth = *pFbPitch = 0;
-
-#if defined(CONFIG_FB) && defined(NV_NUM_REGISTERED_FB_PRESENT)
-    if (num_registered_fb > 0)
-    {
-        int i;
-
-        for (i = 0; i < num_registered_fb; i++)
-        {
-            if (!registered_fb[i])
-                continue;
-
-            /* Make sure base address is mapped to GPU BAR */
-            if ((registered_fb[i]->fix.smem_start == consoleBar1Address) ||
-                (registered_fb[i]->fix.smem_start == consoleBar2Address))
-            {
-                *pPhysicalAddress = registered_fb[i]->fix.smem_start;
-                *pFbWidth = registered_fb[i]->var.xres;
-                *pFbHeight = registered_fb[i]->var.yres;
-                *pFbDepth = registered_fb[i]->var.bits_per_pixel;
-                *pFbPitch = registered_fb[i]->fix.line_length;
-                return;
-            }
-        }
-    }
-#endif
-
-    /*
-     * If the screen info is not found in the registered FBs then fallback
-     * to the screen_info structure.
-     *
-     * The SYSFB_SIMPLEFB option, if enabled, marks VGA/VBE/EFI framebuffers as
-     * generic framebuffers so the new generic system-framebuffer drivers can
-     * be used instead. DRM_SIMPLEDRM drives the generic system-framebuffers
-     * device created by SYSFB_SIMPLEFB.
-     *
-     * SYSFB_SIMPLEFB registers a dummy framebuffer which does not contain the
-     * information required by os_get_screen_info(), therefore you need to
-     * fall back onto the screen_info structure.
-     *
-     * After commit b8466fe82b79 ("efi: move screen_info into efi init code")
-     * in v6.7, 'screen_info' is exported as GPL licensed symbol for ARM64.
-     */
-
-#if NV_CHECK_EXPORT_SYMBOL(screen_info)
-    /*
-     * If there is not a framebuffer console, return 0 size.
-     *
-     * orig_video_isVGA is set to 1 during early Linux kernel
-     * initialization, and then will be set to a value, such as
-     * VIDEO_TYPE_VLFB or VIDEO_TYPE_EFI if an fbdev console is used.
-     */
-    if (screen_info.orig_video_isVGA > 1)
-    {
-        NvU64 physAddr = screen_info.lfb_base;
-#if defined(VIDEO_CAPABILITY_64BIT_BASE)
-        physAddr |= (NvU64)screen_info.ext_lfb_base << 32;
-#endif
-
-        /* Make sure base address is mapped to GPU BAR */
-        if ((physAddr == consoleBar1Address) ||
-            (physAddr == consoleBar2Address))
-        {
-            *pPhysicalAddress = physAddr;
-            *pFbWidth = screen_info.lfb_width;
-            *pFbHeight = screen_info.lfb_height;
-            *pFbDepth = screen_info.lfb_depth;
-            *pFbPitch = screen_info.lfb_linelength;
-        }
-    }
-#endif
-}
-
 void NV_API_CALL os_dump_stack(void)
 {
     dump_stack();
@@ -1364,6 +1357,16 @@ NV_STATUS NV_API_CALL os_get_version_info(os_version_info * pOsVersionInfo)
     return status;
 }
 
+NV_STATUS NV_API_CALL os_get_is_openrm(NvBool *bIsOpenRm)
+{
+#if defined(NVCPU_X86_64) || defined(NVCPU_AARCH64)
+    *bIsOpenRm = NV_TRUE;
+    return NV_OK;
+#else // defined(NVCPU_X86_64) || defined(NVCPU_AARCH64)
+    return NV_ERR_NOT_SUPPORTED;
+#endif // defined(NVCPU_X86_64) || defined(NVCPU_AARCH64)
+}
+
 NvBool NV_API_CALL os_is_xen_dom0(void)
 {
 #if defined(NV_DOM0_KERNEL_PRESENT)
@@ -1669,7 +1672,7 @@ NV_STATUS NV_API_CALL os_alloc_pages_node
     NV_STATUS status = NV_ERR_NOT_SUPPORTED;
 
 #if defined(__GFP_THISNODE) && defined(GFP_HIGHUSER_MOVABLE) && \
-    defined(__GFP_COMP) && defined(__GFP_NORETRY) && defined(__GFP_NOWARN)
+    defined(__GFP_COMP) && defined(__GFP_NOWARN)
     gfp_t gfp_mask;
     struct page *alloc_addr;
     unsigned int order = get_order(size);
@@ -1686,14 +1689,12 @@ NV_STATUS NV_API_CALL os_alloc_pages_node
      *                              pages, which is needed in order to use
      *                              vm_insert_page API.
      *
-     * 4. __GFP_NORETRY:            Used to avoid the Linux kernel OOM killer.
-     *
-     * 5. __GFP_NOWARN:             Used to avoid a WARN_ON in the slowpath if
+     * 4. __GFP_NOWARN:             Used to avoid a WARN_ON in the slowpath if
      *                              the requested order is too large (just fail
      *                              instead).
      *
-     * 6. (Optional) __GFP_RECLAIM: Used to allow/forbid reclaim.
-     *                              This is part of GFP_USER and consequently 
+     * 5. (Optional) __GFP_RECLAIM: Used to allow/forbid reclaim.
+     *                              This is part of GFP_USER and consequently
      *                              GFP_HIGHUSER_MOVABLE.
      *
      * Some of these flags are relatively more recent, with the last of them
@@ -1706,7 +1707,30 @@ NV_STATUS NV_API_CALL os_alloc_pages_node
      */
 
     gfp_mask = __GFP_THISNODE | GFP_HIGHUSER_MOVABLE | __GFP_COMP |
-               __GFP_NORETRY | __GFP_NOWARN;
+               __GFP_NOWARN;
+    
+#if defined(__GFP_RETRY_MAYFAIL)
+
+    /*
+     * __GFP_RETRY_MAYFAIL :  Used to avoid the Linux kernel OOM killer.
+     *                        To help PMA on paths where UVM might be
+     *                        in memory over subscription. This gives UVM 
+     *                        a chance to free memory before invoking any 
+     *                        action from the OOM killer.
+     *                        Freeing non-essential memory will also benefit 
+     *                        the system as a whole.
+     */
+
+    gfp_mask |= __GFP_RETRY_MAYFAIL;
+#elif defined(__GFP_NORETRY)
+
+    /*
+     *  __GFP_NORETRY :       Use __GFP_NORETRY on older kernels where
+     *                        __GFP_RETRY_MAYFAIL is not present.
+     */
+
+    gfp_mask |= __GFP_NORETRY;
+#endif
 
 #if defined(__GFP_RECLAIM)
     if (flag & NV_ALLOC_PAGES_NODE_SKIP_RECLAIM)
@@ -1810,6 +1834,7 @@ NV_STATUS NV_API_CALL os_open_temporary_file
     void **ppFile
 )
 {
+#if NV_FILESYSTEM_ACCESS_AVAILABLE
 #if defined(O_TMPFILE)
     struct file *file;
     const char *default_path = "/tmp";
@@ -1855,6 +1880,9 @@ NV_STATUS NV_API_CALL os_open_temporary_file
 #else
     return NV_ERR_NOT_SUPPORTED;
 #endif
+#else
+    return NV_ERR_NOT_SUPPORTED;
+#endif
 }
 
 void NV_API_CALL os_close_file
@@ -1862,7 +1890,9 @@ void NV_API_CALL os_close_file
     void *pFile
 )
 {
+#if NV_FILESYSTEM_ACCESS_AVAILABLE
     filp_close(pFile, NULL);
+#endif
 }
 
 #define NV_MAX_NUM_FILE_IO_RETRIES 10
@@ -1875,6 +1905,7 @@ NV_STATUS NV_API_CALL os_write_file
     NvU64 offset
 )
 {
+#if NV_FILESYSTEM_ACCESS_AVAILABLE
     loff_t f_pos = offset;
     ssize_t num_written;
     int num_retries = NV_MAX_NUM_FILE_IO_RETRIES;
@@ -1905,6 +1936,9 @@ retry:
     }
 
     return NV_OK;
+#else
+    return NV_ERR_NOT_SUPPORTED;
+#endif
 }
 
 NV_STATUS NV_API_CALL os_read_file
@@ -1915,6 +1949,7 @@ NV_STATUS NV_API_CALL os_read_file
     NvU64 offset
 )
 {
+#if NV_FILESYSTEM_ACCESS_AVAILABLE
     loff_t f_pos = offset;
     ssize_t num_read;
     int num_retries = NV_MAX_NUM_FILE_IO_RETRIES;
@@ -1945,6 +1980,9 @@ retry:
     }
 
     return NV_OK;
+#else
+    return NV_ERR_NOT_SUPPORTED;
+#endif
 }
 
 NV_STATUS NV_API_CALL os_open_readonly_file
@@ -1953,6 +1991,7 @@ NV_STATUS NV_API_CALL os_open_readonly_file
     void       **ppFile
 )
 {
+#if NV_FILESYSTEM_ACCESS_AVAILABLE
     struct file *file;
 
     /*
@@ -1974,6 +2013,9 @@ NV_STATUS NV_API_CALL os_open_readonly_file
     *ppFile = (void *)file;
 
     return NV_OK;
+#else
+    return NV_ERR_NOT_SUPPORTED;
+#endif
 }
 
 NV_STATUS NV_API_CALL os_open_and_read_file
@@ -2130,6 +2172,22 @@ void NV_API_CALL os_nv_cap_close_fd
     nv_cap_close_fd(fd);
 }
 
+NvS32 NV_API_CALL os_imex_channel_count
+(
+    void
+)
+{
+    return nv_caps_imex_channel_count();
+}
+
+NvS32 NV_API_CALL os_imex_channel_get
+(
+    NvU64 descriptor
+)
+{
+    return nv_caps_imex_channel_get((int)descriptor);
+}
+
 /*
  * Reads the total memory and free memory of a NUMA node from the kernel.
  */
@@ -2214,6 +2272,8 @@ static int os_numa_verify_gpu_memory_zone(struct notifier_block *nb,
     return NOTIFY_OK;
 }
 
+#define ADD_REMOVE_GPU_MEMORY_NUM_SEGMENTS 4
+
 NV_STATUS NV_API_CALL os_numa_add_gpu_memory
 (
     void *handle,
@@ -2227,7 +2287,12 @@ NV_STATUS NV_API_CALL os_numa_add_gpu_memory
     nv_linux_state_t *nvl = pci_get_drvdata(handle);
     nv_state_t *nv = NV_STATE_PTR(nvl);
     NvU64 base = offset + nvl->coherent_link_info.gpu_mem_pa;
-    int ret;
+    int ret = 0;
+    NvU64 memblock_size;
+    NvU64 size_remaining;
+    NvU64 calculated_segment_size;
+    NvU64 segment_size;
+    NvU64 segment_base;
     os_numa_gpu_mem_hotplug_notifier_t notifier =
     {
         .start_pa = base,
@@ -2260,11 +2325,49 @@ NV_STATUS NV_API_CALL os_numa_add_gpu_memory
         goto failed;
     }
 
+    //
+    // Adding all memory at once can take a long time. Split up memory into segments
+    // with schedule() in between to prevent soft lockups. Memory segments for
+    // add_memory_driver_managed() need to be aligned to memblock size.
+    //
+    // If there are any issues splitting into segments, then add all memory at once.
+    //
+    if (os_numa_memblock_size(&memblock_size) == NV_OK)
+    {
+        calculated_segment_size = NV_ALIGN_UP(size / ADD_REMOVE_GPU_MEMORY_NUM_SEGMENTS, memblock_size);
+    }
+    else
+    {
+        // Don't split into segments, add all memory at once
+        calculated_segment_size = size;
+    }
+
+    segment_size = calculated_segment_size;
+    segment_base = base;
+    size_remaining = size;
+
+    while ((size_remaining > 0) &&
+           (ret == 0))
+    {
+        if (segment_size > size_remaining)
+        {
+            segment_size = size_remaining;
+        }
+
 #ifdef NV_ADD_MEMORY_DRIVER_MANAGED_HAS_MHP_FLAGS_ARG
-    ret = add_memory_driver_managed(node, base, size, "System RAM (NVIDIA)", MHP_NONE);
+        ret = add_memory_driver_managed(node, segment_base, segment_size, "System RAM (NVIDIA)", MHP_NONE);
 #else
-    ret = add_memory_driver_managed(node, base, size, "System RAM (NVIDIA)");
+        ret = add_memory_driver_managed(node, segment_base, segment_size, "System RAM (NVIDIA)");
 #endif
+        nv_printf(NV_DBG_SETUP, "NVRM: add_memory_driver_managed() returns: %d for segment_base: 0x%llx, segment_size: 0x%llx\n",
+                  ret, segment_base, segment_size);
+
+        segment_base += segment_size;
+        size_remaining -= segment_size;
+
+        // Yield CPU to prevent soft lockups
+        schedule();
+    }
     unregister_memory_notifier(&notifier.memory_notifier);
 
     if (ret == 0)
@@ -2278,19 +2381,69 @@ NV_STATUS NV_API_CALL os_numa_add_gpu_memory
             zone_end_pfn(zone) != end_pfn)
         {
             nv_printf(NV_DBG_ERRORS, "NVRM: GPU memory zone movable auto onlining failed!\n");
+
 #ifdef NV_OFFLINE_AND_REMOVE_MEMORY_PRESENT
-#ifdef NV_REMOVE_MEMORY_HAS_NID_ARG
-            if (offline_and_remove_memory(node, base, size) != 0)
-#else
-            if (offline_and_remove_memory(base, size) != 0)
-#endif
+            // Since zone movable auto onlining failed, need to remove the added memory.
+            segment_size = calculated_segment_size;
+            segment_base = base;
+            size_remaining = size;
+
+            while (size_remaining > 0)
             {
-                nv_printf(NV_DBG_ERRORS, "NVRM: offline_and_remove_memory failed\n");
+                if (segment_size > size_remaining)
+                {
+                    segment_size = size_remaining;
+                }
+
+#ifdef NV_REMOVE_MEMORY_HAS_NID_ARG
+                ret = offline_and_remove_memory(node, segment_base, segment_size);
+#else
+                ret = offline_and_remove_memory(segment_base, segment_size);
+#endif
+                nv_printf(NV_DBG_SETUP, "NVRM: offline_and_remove_memory() returns: %d for segment_base: 0x%llx, segment_size: 0x%llx\n",
+                          ret, segment_base, segment_size);
+
+                segment_base += segment_size;
+                size_remaining -= segment_size;
+
+                // Yield CPU to prevent soft lockups
+                schedule();
             }
 #endif
             goto failed;
         }
 
+        /*
+         * On systems with cpuset cgroup controller enabled, memory alloc on
+         * this just hotplugged GPU memory node can fail if the
+         * cpuset_hotplug_work is not scheduled yet. cpuset_hotplug_work is
+         * where the current->mems_allowed is updated in the path
+         * cpuset_hotplug_workfn->update_tasks_nodemask. When cpuset is
+         * enabled and current->mems_allowed is not updated, memory allocation
+         * with __GFP_THISNODE and this node id fails. cpuset_wait_for_hotplug
+         * kernel function can be used to wait for the work to finish but that
+         * is not exported. Adding a time loop to wait for
+         * current->mems_allowed to be updated as a WAR while an upstream
+         * kernel fix is being explored. Bug 4385903
+         */
+        if (!node_isset(node, cpuset_current_mems_allowed))
+        {
+            unsigned long delay;
+
+            delay = jiffies + (HZ / 10); // 100ms
+            while(time_before(jiffies, delay) &&
+                  !node_isset(node, cpuset_current_mems_allowed))
+            {
+                os_schedule();
+            }
+
+            if (!node_isset(node, cpuset_current_mems_allowed))
+            {
+                nv_printf(NV_DBG_ERRORS, "NVRM: Hotplugged GPU memory NUMA node: %d "
+                          "not set in current->mems_allowed!\n", node);
+            }
+        }
+
         *nodeId = node;
         clear_bit(node, nvl->coherent_link_info.free_node_bitmap);
         NV_ATOMIC_SET(nvl->numa_info.status, NV_IOCTL_NUMA_STATUS_ONLINE);
@@ -2305,6 +2458,77 @@ failed:
     return NV_ERR_NOT_SUPPORTED;
 }
 
+
+typedef struct {
+    NvU64 base;
+    NvU64 size;
+    NvU32 nodeId;
+    int ret;
+} remove_numa_memory_info_t;
+
+static void offline_numa_memory_callback
+(
+    void *args
+)
+{
+#ifdef NV_OFFLINE_AND_REMOVE_MEMORY_PRESENT
+    remove_numa_memory_info_t *pNumaInfo = (remove_numa_memory_info_t *)args;
+    int ret = 0;
+    NvU64 memblock_size;
+    NvU64 size_remaining;
+    NvU64 calculated_segment_size;
+    NvU64 segment_size;
+    NvU64 segment_base;
+
+    //
+    // Removing all memory at once can take a long time. Split up memory into segments
+    // with schedule() in between to prevent soft lockups. Memory segments for
+    // offline_and_remove_memory() need to be aligned to memblock size.
+    //
+    // If there are any issues splitting into segments, then remove all memory at once.
+    //
+    if (os_numa_memblock_size(&memblock_size) == NV_OK)
+    {
+        calculated_segment_size = NV_ALIGN_UP(pNumaInfo->size / ADD_REMOVE_GPU_MEMORY_NUM_SEGMENTS, memblock_size);
+    }
+    else
+    {
+        // Don't split into segments, remove all memory at once
+        calculated_segment_size = pNumaInfo->size;
+    }
+
+    segment_size = calculated_segment_size;
+    segment_base = pNumaInfo->base;
+    size_remaining = pNumaInfo->size;
+
+    while (size_remaining > 0)
+    {
+        if (segment_size > size_remaining)
+        {
+            segment_size = size_remaining;
+        }
+
+#ifdef NV_REMOVE_MEMORY_HAS_NID_ARG
+        ret = offline_and_remove_memory(pNumaInfo->nodeId,
+                                        segment_base,
+                                        segment_size);
+#else
+        ret = offline_and_remove_memory(segment_base,
+                                        segment_size);
+#endif
+        nv_printf(NV_DBG_SETUP, "NVRM: offline_and_remove_memory() returns: %d for segment_base: 0x%llx, segment_size: 0x%llx\n",
+                  ret, segment_base, segment_size);
+        pNumaInfo->ret |= ret;
+
+        segment_base += segment_size;
+        size_remaining -= segment_size;
+
+        // Yield CPU to prevent soft lockups
+        schedule();
+    }
+#endif
+}
+
 NV_STATUS NV_API_CALL os_numa_remove_gpu_memory
 (
     void *handle,
@@ -2443,3 +2667,16 @@ NV_STATUS NV_API_CALL os_find_ns_pid(void *pid_info, NvU32 *ns_pid)
     return NV_OK;
 }
 
+NvBool NV_API_CALL os_is_init_ns(void)
+{
+    return (task_active_pid_ns(current) == &init_pid_ns);
+}
+
+NV_STATUS NV_API_CALL os_device_vm_present(void)
+{
+#if defined(NV_DEVICE_VM_BUILD)
+    return NV_OK;
+#else
+    return NV_ERR_NOT_SUPPORTED;
+#endif
+}
