@@ -26,8 +26,10 @@
 #include "nvkms-flip.h"
 #include "nvkms-hw-flip.h"
 #include "nvkms-utils-flip.h"
+#include "nvkms-lut.h"
 #include "nvkms-prealloc.h"
 #include "nvkms-private.h"
+#include "nvkms-utils.h"
 #include "nvkms-vrr.h"
 #include "nvkms-dpy.h"
 #include "nvkms-rm.h"
@@ -67,9 +69,12 @@ NvBool nvCheckFlipPermissions(
         layerMask = allLayersMask;
     }
 
-    /* Changing viewPortIn requires permission to alter all layers. */
+    /* Changing viewPortIn or output LUT requires permission to alter all layers. */
 
-    if (pParams->viewPortIn.specified && (layerMask != allLayersMask)) {
+    if ((layerMask != allLayersMask) && ((pParams->viewPortIn.specified) ||
+                                         (pParams->olut.specified) ||
+                                         (pParams->lut.input.specified)  ||
+                                         (pParams->lut.output.specified))) {
         return FALSE;
     }
 
@@ -123,52 +128,82 @@ static NvBool UpdateProposedFlipStateOneApiHead(
     const struct NvKmsFlipCommonParams *pParams,
     NVProposedFlipStateOneApiHead *pProposedApiHead)
 {
+    const NVDevEvoRec *pDevEvo = pDispEvo->pDevEvo;
     const NVDispApiHeadStateEvoRec *pApiHeadState =
         &pDispEvo->apiHeadState[apiHead];
+    const NVDpyEvoRec *pDpyEvo =
+        nvGetOneArbitraryDpyEvo(pApiHeadState->activeDpys, pDispEvo);
+    NvU32 layer;
 
     if (pParams->tf.specified) {
-        const NVDpyEvoRec *pDpyEvo =
-            nvGetOneArbitraryDpyEvo(pApiHeadState->activeDpys, pDispEvo);
-
         pProposedApiHead->dirty.hdr = TRUE;
         pProposedApiHead->hdr.tf = pParams->tf.val;
+    }
 
-        // If enabling HDR...
-        // XXX HDR TODO: Handle other transfer functions
-        if (pParams->tf.val == NVKMS_OUTPUT_TF_PQ) {
-            // Cannot be an SLI configuration.
-            // XXX HDR TODO: Test SLI Mosaic + HDR and remove this check
-            if (pDispEvo->pDevEvo->numSubDevices > 1) {
-                return FALSE;
+    if (pParams->colorimetry.specified) {
+        pProposedApiHead->dirty.hdr = TRUE;
+        pProposedApiHead->hdr.dpyColor.colorimetry = pParams->colorimetry.val;
+    }
+
+    if (pParams->hdrInfoFrame.specified) {
+        pProposedApiHead->dirty.hdr = TRUE;
+        pProposedApiHead->hdr.infoFrameOverride =
+            pParams->hdrInfoFrame.enabled;
+    }
+
+    for (layer = 0; layer < pDevEvo->apiHead[apiHead].numLayers; layer++) {
+        if (pParams->layer[layer].hdr.specified) {
+            pProposedApiHead->dirty.hdr = TRUE;
+            if (pParams->layer[layer].hdr.enabled) {
+                pProposedApiHead->hdr.staticMetadataLayerMask |=
+                    1 << layer;
+            } else {
+                pProposedApiHead->hdr.staticMetadataLayerMask &=
+                    ~(1 << layer);
             }
-
-            // Sink must support HDR.
-            if (!nvIsHDRCapableHead(pDispEvo, apiHead)) {
-                return FALSE;
-            }
-        }
-
-        if (!nvChooseCurrentColorSpaceAndRangeEvo(pDpyEvo,
-                                                  pApiHeadState->timings.yuv420Mode,
-                                                  pParams->tf.val,
-                                                  pDpyEvo->requestedColorSpace,
-                                                  pDpyEvo->requestedColorRange,
-                                                  &pProposedApiHead->hdr.colorSpace,
-                                                  &pProposedApiHead->hdr.colorBpc,
-                                                  &pProposedApiHead->hdr.colorRange)) {
-            return FALSE;
         }
     }
 
-    /*
-     * Change in colorSpace and colorBpc is not handled. For DisplayPort,
-     * colorSpace and colorBpc  can not be changed without a modeset.
-     */
-    if ((pProposedApiHead->hdr.colorSpace !=
-            pApiHeadState->attributes.colorSpace) ||
-        (pProposedApiHead->hdr.colorBpc !=
-             pApiHeadState->attributes.colorBpc)) {
-        return FALSE;
+    if (pProposedApiHead->dirty.hdr) {
+        // If enabling HDR output TF...
+        if (pProposedApiHead->hdr.tf == NVKMS_OUTPUT_TF_PQ) {
+            // Cannot be an SLI configuration.
+            // XXX HDR TODO: Test SLI Mosaic + HDR and remove this check
+            if (pDevEvo->numSubDevices > 1) {
+                return FALSE;
+            }
+
+            /* NVKMS_OUTPUT_TF_PQ requires the RGB color space */
+            if (pProposedApiHead->hdr.dpyColor.format !=
+                    NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_SPACE_RGB) {
+                return FALSE;
+            }
+        }
+
+        // If enabling HDR signaling...
+        // XXX HDR TODO: Handle other colorimetries
+        if (pProposedApiHead->hdr.infoFrameOverride ||
+            (pProposedApiHead->hdr.staticMetadataLayerMask != 0) ||
+            (pProposedApiHead->hdr.dpyColor.colorimetry ==
+                NVKMS_OUTPUT_COLORIMETRY_BT2100)) {
+            const NVDpyEvoRec *pDpyEvoIter;
+
+            // All dpys on apiHead must support HDR.
+            FOR_ALL_EVO_DPYS(pDpyEvoIter,
+                             pApiHeadState->activeDpys,
+                             pDispEvo) {
+                if (!nvDpyIsHDRCapable(pDpyEvoIter)) {
+                    return FALSE;
+                }
+            }
+        }
+
+        if (!nvChooseColorRangeEvo(pDpyEvo->requestedColorRange,
+                                   pProposedApiHead->hdr.dpyColor.format,
+                                   pProposedApiHead->hdr.dpyColor.bpc,
+                                   &pProposedApiHead->hdr.dpyColor.range)) {
+            return FALSE;
+        }
     }
 
     if (pParams->viewPortIn.specified) {
@@ -176,109 +211,62 @@ static NvBool UpdateProposedFlipStateOneApiHead(
         pProposedApiHead->viewPortPointIn = pParams->viewPortIn.point;
     }
 
+    if (!nvValidateSetLutCommonParams(pDispEvo->pDevEvo, &pParams->lut)) {
+        return FALSE;
+    }
+    pProposedApiHead->lut = pParams->lut;
+
     return TRUE;
 }
 
-void nvCancelSDRTransitionTimer(NVDispApiHeadStateEvoRec *pApiHeadState)
-{
-    nvkms_free_timer(pApiHeadState->hdrToSdrTransitionTimer);
-    pApiHeadState->hdrToSdrTransitionTimer = NULL;
-}
-
-static void SDRTransition(void *dataPtr, NvU32 apiHead)
-{
-    NvU32 head;
-    NVDispEvoRec *pDispEvo = dataPtr;
-    NVDispApiHeadStateEvoRec *pApiHeadState =
-        &pDispEvo->apiHeadState[apiHead];
-    NVDpyEvoPtr pDpyEvo =
-        nvGetOneArbitraryDpyEvo(pApiHeadState->activeDpys, pDispEvo);
-
-    nvCancelSDRTransitionTimer(pApiHeadState);
-
-    if (pDpyEvo == NULL) {
-        return;
-    }
-
-    nvAssert(pApiHeadState->hwHeadsMask != 0);
-
-    FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, head) {
-        NVDispHeadStateEvoRec *pHeadState =
-            &pDispEvo->headState[head];
-        nvAssert(pHeadState->hdr.outputState ==
-                 NVKMS_HDR_OUTPUT_STATE_TRANSITIONING_TO_SDR);
-        pHeadState->hdr.outputState = NVKMS_HDR_OUTPUT_STATE_SDR;
-    }
-
-    nvUpdateInfoFrames(pDpyEvo);
-}
-
-static
-void ScheduleSDRTransitionTimer(NVDispEvoRec *pDispEvo, const NvU32 apiHead)
-{
-    NVDispApiHeadStateEvoRec *pApiHeadState =
-        &pDispEvo->apiHeadState[apiHead];
-
-    nvAssert(!pApiHeadState->hdrToSdrTransitionTimer);
-
-    pApiHeadState->hdrToSdrTransitionTimer =
-        nvkms_alloc_timer(SDRTransition,
-                          pDispEvo,
-                          apiHead,
-                          2000000 /* 2 seconds */);
-}
-
 static NvBool GetAllowVrr(const NVDevEvoRec *pDevEvo,
                           const struct NvKmsFlipRequestOneHead *pFlipHead,
-                          NvU32 numFlipHeads,
-                          NvBool allowVrr,
-                          NvBool *pApplyAllowVrr)
+                          NvU32 numFlipHeads)
 {
     NvU32 sd, i;
     const NVDispEvoRec *pDispEvo;
     const NvU32 requestedApiHeadCount = numFlipHeads;
     NvU32 activeApiHeadCount, dirtyMainLayerCount;
+    NvBool allowVrr = TRUE;
 
-    *pApplyAllowVrr = FALSE;
+    if (!pDevEvo->hal->caps.supportsDisplayRate)
+    {
+        /*!
+         * Count active heads so we can make a decision about VRR
+         * and register syncpts if specified.
+         */
+        activeApiHeadCount = dirtyMainLayerCount = 0;
 
-    /*!
-     * Count active heads so we can make a decision about VRR
-     * and register syncpts if specified.
-     */
-    activeApiHeadCount = dirtyMainLayerCount = 0;
-
-    FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
-        NvU32 apiHead;
-        for (apiHead = 0; apiHead < pDevEvo->numApiHeads; apiHead++) {
-            if (nvApiHeadIsActive(pDispEvo, apiHead)) {
-                activeApiHeadCount++;
+        FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
+            NvU32 apiHead;
+            for (apiHead = 0; apiHead < pDevEvo->numApiHeads; apiHead++) {
+                if (nvApiHeadIsActive(pDispEvo, apiHead)) {
+                    activeApiHeadCount++;
+                }
             }
         }
-    }
 
-    for (i = 0; i < numFlipHeads; i++) {
-        if (nvIsLayerDirty(&pFlipHead[i].flip, NVKMS_MAIN_LAYER)) {
-            dirtyMainLayerCount++;
+        for (i = 0; i < numFlipHeads; i++) {
+            if (nvIsLayerDirty(&pFlipHead[i].flip, NVKMS_MAIN_LAYER)) {
+                dirtyMainLayerCount++;
+            }
+        }
+
+        /*
+         * Deactivate VRR if only a subset of the heads are requested,
+         * only a subset of the heads are being flipped, or only a subset
+         * of the heads are allowed a VRR flip.
+         */
+        if ((activeApiHeadCount != requestedApiHeadCount) ||
+                (activeApiHeadCount != dirtyMainLayerCount)) {
+            allowVrr = FALSE;
         }
     }
 
-    /*
-     * Deactivate VRR if only a subset of the heads are requested or
-     * only a subset of the heads are being flipped.
-     */
-    if ((activeApiHeadCount != requestedApiHeadCount) ||
-            (activeApiHeadCount != dirtyMainLayerCount)) {
+    if (!pDevEvo->vrr.enabled) {
         allowVrr = FALSE;
     }
 
-    /*
-     * Apply NvKmsFlipRequest::allowVrr
-     * only if at least one main layer is became dirty.
-     */
-    if (dirtyMainLayerCount > 0) {
-        *pApplyAllowVrr = TRUE;
-    }
-
     return allowVrr;
 }
 
@@ -347,19 +335,146 @@ static void InitNvKmsFlipWorkArea(const NVDevEvoRec *pDevEvo,
                 &pDispEvo->apiHeadState[apiHead];
 
             pProposedApiHead->hdr.tf = pApiHeadState->tf;
-            pProposedApiHead->hdr.colorSpace =
-                pApiHeadState->attributes.colorSpace;
-            pProposedApiHead->hdr.colorBpc =
-                pApiHeadState->attributes.colorBpc;
-            pProposedApiHead->hdr.colorRange =
-                pApiHeadState->attributes.colorRange;
+            pProposedApiHead->hdr.dpyColor = pApiHeadState->attributes.color;
+            pProposedApiHead->hdr.infoFrameOverride =
+                pApiHeadState->hdrInfoFrameOverride;
+            pProposedApiHead->hdr.staticMetadataLayerMask =
+                pApiHeadState->hdrStaticMetadataLayerMask;
 
             pProposedApiHead->viewPortPointIn =
                 pApiHeadState->viewPortPointIn;
+
+            pProposedApiHead->lut.input.specified =
+                FALSE;
+            pProposedApiHead->lut.output.specified =
+                FALSE;
         }
     }
 }
 
+static void CleanupNvKmsFlipWorkArea(NVDevEvoPtr pDevEvo,
+                                     struct NvKmsFlipWorkArea *pWorkArea)
+{
+    const NVDispEvoRec *pDispEvo;
+    NvU32 sd, head;
+
+    FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
+        for (head = 0; head < ARRAY_LEN(pWorkArea->sd[sd].head); head++) {
+            /*
+             * If the flip failed or wasn't committed, any TMO surfaces
+             * allocated by nvSetTmoLutSurfaceEvo will be left in newState with
+             * 1 refcnt, so free them now.
+             */
+            nvFreeUnrefedTmoLutSurfacesEvo(pDevEvo,
+                                           &pWorkArea->sd[sd].head[head].newState,
+                                           head);
+        }
+    }
+}
+
+static void FlipEvoOneApiHead(NVDispEvoRec *pDispEvo,
+                              const NvU32 apiHead,
+                              const struct NvKmsFlipWorkArea *pWorkArea,
+                              const NvBool allowFlipLock,
+                              NVEvoUpdateState *pUpdateState)
+{
+    NVDevEvoRec *pDevEvo = pDispEvo->pDevEvo;
+    const NvU32 sd = pDispEvo->displayOwner;
+    NvU32 head;
+    const NVProposedFlipStateOneApiHead *pProposedApiHead =
+        &pWorkArea->disp[sd].apiHead[apiHead].proposedFlipState;
+    NVDispApiHeadStateEvoRec *pApiHeadState =
+        &pDispEvo->apiHeadState[apiHead];
+    NVDpyEvoRec *pDpyEvo =
+        nvGetOneArbitraryDpyEvo(pApiHeadState->activeDpys, pDispEvo);
+    const NVT_EDID_INFO *pInfo = &pDpyEvo->parsedEdid.info;
+    const NVT_HDR_STATIC_METADATA *pHdrInfo =
+        &pInfo->hdr_static_metadata_info;
+
+    nvAssert(nvApiHeadIsActive(pDispEvo, apiHead));
+
+    if (pProposedApiHead->lut.input.specified ||
+        pProposedApiHead->lut.output.specified) {
+        /* Set LUT settings */
+        nvEvoSetLut(pDispEvo, apiHead, FALSE /* kickoff */,
+                    &pProposedApiHead->lut);
+        nvEvoStageLUTNotifier(pDispEvo, apiHead);
+    }
+
+    FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, head) {
+        nvFlipEvoOneHead(pDevEvo, sd, head, pHdrInfo,
+                         &pWorkArea->sd[sd].head[head].newState,
+                         allowFlipLock,
+                         pUpdateState);
+
+        if (pProposedApiHead->dirty.hdr) {
+            /* Update hardware's current colorSpace and colorRange */
+            nvUpdateCurrentHardwareColorSpaceAndRangeEvo(
+                pDispEvo,
+                head,
+                &pProposedApiHead->hdr.dpyColor,
+                pUpdateState);
+        }
+    }
+
+    if (pProposedApiHead->dirty.hdr) {
+        pApiHeadState->attributes.color = pProposedApiHead->hdr.dpyColor;
+        pApiHeadState->tf = pProposedApiHead->hdr.tf;
+
+        pApiHeadState->hdrInfoFrameOverride =
+            pProposedApiHead->hdr.infoFrameOverride;
+        pApiHeadState->hdrStaticMetadataLayerMask =
+            pProposedApiHead->hdr.staticMetadataLayerMask;
+
+        nvUpdateInfoFrames(pDpyEvo);
+    }
+
+    if (pProposedApiHead->dirty.viewPortPointIn) {
+        pApiHeadState->viewPortPointIn =
+            pProposedApiHead->viewPortPointIn;
+    }
+}
+
+static NvU32 FlipEvo2Head1OrOneDisp(NVDispEvoRec *pDispEvo,
+                                    struct NvKmsFlipWorkArea *pWorkArea,
+                                    const NvBool skipUpdate)
+{
+    NVDevEvoRec *pDevEvo = pDispEvo->pDevEvo;
+    NvU32 flip2Heads1OrApiHeadsMask = 0x0;
+
+    for (NvU32 apiHead = 0; apiHead < NVKMS_MAX_HEADS_PER_DISP; apiHead++) {
+        NVDispApiHeadStateEvoRec *pApiHeadState =
+            &pDispEvo->apiHeadState[apiHead];
+        const NvBool b2Heads1Or =
+            (nvPopCount32(pApiHeadState->hwHeadsMask) >= 2);
+
+        if (!nvApiHeadIsActive(pDispEvo, apiHead) || !b2Heads1Or) {
+            continue;
+        }
+
+        nvkms_memset(&pWorkArea->updateState, 0,
+                     sizeof(pWorkArea->updateState));
+
+        FlipEvoOneApiHead(pDispEvo, apiHead, pWorkArea,
+                          TRUE /* allowFlipLock */, &pWorkArea->updateState);
+
+        /*
+         * If api-head is using 2Heads1OR mode then it can not be flip with
+         * other ap-heads in a single update; because each api-head, which is
+         * using 2Heads1OR mode, uses different fliplock group and kicking off
+         * multiple fliplock groups as part of a single update call is not
+         * supported yet.
+         */
+        pDevEvo->hal->Update(pDevEvo, &pWorkArea->updateState,
+                             TRUE /* releaseElv */);
+        nvAssert(!skipUpdate);
+
+        flip2Heads1OrApiHeadsMask |= NVBIT(apiHead);
+    }
+
+    return flip2Heads1OrApiHeadsMask;
+}
+
 /*!
  * Program a flip on all requested layers on all requested heads on
  * all requested disps in NvKmsFlipRequest.
@@ -384,24 +499,23 @@ NvBool nvFlipEvo(NVDevEvoPtr pDevEvo,
                  const struct NvKmsFlipRequestOneHead *pFlipHead,
                  NvU32 numFlipHeads,
                  NvBool commit,
-                 NvBool requestAllowVrr,
                  struct NvKmsFlipReply *reply,
                  NvBool skipUpdate,
                  NvBool allowFlipLock)
 {
     NvS32 vrrSemaphoreIndex = -1;
     NvU32 apiHead, sd;
-    NvBool applyAllowVrr = FALSE;
     NvBool ret = FALSE;
+    enum NvKmsFlipResult result = NV_KMS_FLIP_RESULT_INVALID_PARAMS;
     NvBool changed = FALSE;
+    NvBool vrrOnSubsetOfHeads; 
     NVDispEvoPtr pDispEvo;
-    const NvBool allowVrr =
-        GetAllowVrr(pDevEvo, pFlipHead, numFlipHeads,
-                    requestAllowVrr, &applyAllowVrr);
-    struct NvKmsFlipWorkArea *pWorkArea =
-        nvPreallocGet(pDevEvo, PREALLOC_TYPE_FLIP_WORK_AREA,
-                      sizeof(*pWorkArea));
+    const NvBool allowVrrDev =
+        GetAllowVrr(pDevEvo, pFlipHead, numFlipHeads);
+    struct NvKmsFlipWorkArea *pWorkArea;
     NvU32 i;
+    
+    vrrOnSubsetOfHeads = pDevEvo->hal->caps.supportsDisplayRate;
 
     /*
      * Do not execute NVKMS_IOCTL_FLIP if the display channel yet has not
@@ -412,15 +526,51 @@ NvBool nvFlipEvo(NVDevEvoPtr pDevEvo,
      * NVKMS_IOCTL_FLIP requests.
      */
     if (pDevEvo->coreInitMethodsPending) {
-        goto done;
+        if (reply) {
+            reply->flipResult = result;
+        }
+        return ret;
     }
 
+    pWorkArea = nvPreallocGet(pDevEvo, PREALLOC_TYPE_FLIP_WORK_AREA,
+                              sizeof(*pWorkArea));
     InitNvKmsFlipWorkArea(pDevEvo, pWorkArea);
 
+    NvU32 allowVrrApiHeadMasks[NVKMS_MAX_SUBDEVICES];
+    NvU32 applyAllowVrrApiHeadMasks[NVKMS_MAX_SUBDEVICES];
+
+    nvkms_memset(allowVrrApiHeadMasks, 0, sizeof(allowVrrApiHeadMasks));
+    nvkms_memset(applyAllowVrrApiHeadMasks, 0, sizeof(applyAllowVrrApiHeadMasks));
+
     /* Validate the flip parameters and update the work area. */
+    
+    /*
+     * In the pre-displayrate case, any head requesting a VRR flip means
+     * all heads within pDevEvo request a VRR flip. Similarly, any head
+     * with a dirty main layer means allowVrr must be applied to all heads.
+     */
+    NvBool anyAllowVrrHead = FALSE;
+    NvBool anyDirtyMainLayer = FALSE;
+    for (i = 0; i < numFlipHeads; i++) {
+        if (pFlipHead[i].flip.allowVrr) {
+            anyAllowVrrHead = TRUE;
+        }
+        if (nvIsLayerDirty(&pFlipHead[i].flip, NVKMS_MAIN_LAYER)) {
+            anyDirtyMainLayer = TRUE;
+        }
+    }
+
     for (i = 0; i < numFlipHeads; i++) {
         const NvU32 apiHead = pFlipHead[i].head;
         const NvU32 sd = pFlipHead[i].sd;
+        const NvBool allowVrrHead = pFlipHead[i].flip.allowVrr;
+        const NvBool dirtyMainLayer = nvIsLayerDirty(&pFlipHead[i].flip, NVKMS_MAIN_LAYER);
+        const NvBool allowVrr = allowVrrDev && 
+                                (vrrOnSubsetOfHeads ? allowVrrHead : anyAllowVrrHead);
+
+        const NvBool applyAllowVrr = (vrrOnSubsetOfHeads ? 
+                                      dirtyMainLayer : anyDirtyMainLayer);
+
         NVDispEvoPtr pDispEvo = pDevEvo->pDispEvo[sd];
         NvU32 head;
         const NVDispApiHeadStateEvoRec *pApiHeadState =
@@ -454,7 +604,12 @@ NvBool nvFlipEvo(NVDevEvoPtr pDevEvo,
                 goto done;
             }
         }
-
+        if (applyAllowVrr) {
+            applyAllowVrrApiHeadMasks[sd] |= (1 << apiHead);
+            if (allowVrr) {
+                allowVrrApiHeadMasks[sd] |= (1 << apiHead);
+            }
+        }
         pWorkArea->sd[sd].changed = TRUE;
         changed = TRUE;
     }
@@ -470,8 +625,46 @@ NvBool nvFlipEvo(NVDevEvoPtr pDevEvo,
         goto done;
     }
 
+    /* XXX: Fail flip if LUT update in progress.
+     *
+     * Really, we should have a more robust system for this, but currently, the
+     * only user of the LUT parameter to the flip IOCTL is nvidia-drm, which
+     * waits for flips to be complete anyways. We should actually find a way to
+     * properly queue as many LUT-changing flips as we support queued flips in
+     * general.
+     *
+     * This failure returns NV_KMS_FLIP_RESULT_IN_PROGRESS rather than
+     * NV_KMS_FLIP_RESULT_INVALID_PARAMS.
+     *
+     * See bug 4054546 for efforts to update this system.
+     */
+    for (sd = 0; sd < pDevEvo->numSubDevices; sd++) {
+        pDispEvo = pDevEvo->gpus[sd].pDispEvo;
+        for (apiHead = 0; apiHead < pDevEvo->numApiHeads; apiHead++) {
+            const NVProposedFlipStateOneApiHead *pProposedApiHead =
+                &pWorkArea->disp[sd].apiHead[apiHead].proposedFlipState;
+
+            if ((pProposedApiHead->lut.input.specified ||
+                pProposedApiHead->lut.output.specified) &&
+                !nvEvoIsLUTNotifierComplete(pDispEvo, apiHead)) {
+
+                if (commit) {
+                    nvEvoLogDispDebug(
+                        pDispEvo,
+                        EVO_LOG_ERROR,
+                        "Flip request with LUT parameter on API Head %d while LUT update outstanding",
+                        apiHead);
+                }
+
+                result = NV_KMS_FLIP_RESULT_IN_PROGRESS;
+                goto done;
+            }
+        }
+    }
+
     if (!commit) {
         ret = NV_TRUE;
+        result = NV_KMS_FLIP_RESULT_SUCCESS;
         goto done;
     }
 
@@ -487,11 +680,13 @@ NvBool nvFlipEvo(NVDevEvoPtr pDevEvo,
      */
 
     ret = TRUE;
+    result = NV_KMS_FLIP_RESULT_SUCCESS;
 
-    nvPreFlip(pDevEvo, pWorkArea, applyAllowVrr, allowVrr, skipUpdate);
+    nvPreFlip(pDevEvo, pWorkArea, applyAllowVrrApiHeadMasks, 
+              allowVrrApiHeadMasks, skipUpdate);
 
     for (sd = 0; sd < pDevEvo->numSubDevices; sd++) {
-        NVEvoUpdateState updateState = { };
+        NvU32 flip2Heads1OrApiHeadsMask = 0x0;
 
         if (!pWorkArea->sd[sd].changed) {
             continue;
@@ -499,77 +694,56 @@ NvBool nvFlipEvo(NVDevEvoPtr pDevEvo,
 
         pDispEvo = pDevEvo->gpus[sd].pDispEvo;
 
+        flip2Heads1OrApiHeadsMask =
+            FlipEvo2Head1OrOneDisp(pDispEvo, pWorkArea, skipUpdate);
+
+        nvkms_memset(&pWorkArea->updateState, 0,
+                     sizeof(pWorkArea->updateState));
+
+        /*
+         * Ensure that we only commit the LUT notifiers staged in this
+         * nvFlipEvo call.
+         */
+        nvEvoClearStagedLUTNotifiers(pDispEvo);
+
         for (apiHead = 0; apiHead < NVKMS_MAX_HEADS_PER_DISP; apiHead++) {
-            NvU32 head;
-            const NVProposedFlipStateOneApiHead *pProposedApiHead =
-                &pWorkArea->disp[sd].apiHead[apiHead].proposedFlipState;
-            NVDispApiHeadStateEvoRec *pApiHeadState =
-                &pDispEvo->apiHeadState[apiHead];
-            NVDpyEvoRec *pDpyEvo =
-                nvGetOneArbitraryDpyEvo(pApiHeadState->activeDpys, pDispEvo);
-            const NVT_EDID_INFO *pInfo = &pDpyEvo->parsedEdid.info;
-            const NVT_HDR_STATIC_METADATA *pHdrInfo =
-                &pInfo->hdr_static_metadata_info;
-
-
-
-            FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, head) {
-                nvFlipEvoOneHead(pDevEvo, sd, head, pHdrInfo,
-                                 &pWorkArea->sd[sd].head[head].newState,
-                                 allowFlipLock,
-                                 &updateState);
-
-                if (pProposedApiHead->dirty.hdr) {
-                    /* Update hardware's current colorSpace and colorRange */
-                    nvUpdateCurrentHardwareColorSpaceAndRangeEvo(
-                        pDispEvo,
-                        head,
-                        pProposedApiHead->hdr.colorSpace,
-                        pProposedApiHead->hdr.colorRange,
-                        &updateState);
-                }
+            if (!nvApiHeadIsActive(pDispEvo, apiHead) ||
+                    ((NVBIT(apiHead) & flip2Heads1OrApiHeadsMask) != 0x0)) {
+                continue;
             }
 
-            if (pProposedApiHead->dirty.hdr) {
-                pApiHeadState->attributes.colorSpace =
-                    pProposedApiHead->hdr.colorSpace;
-                pApiHeadState->attributes.colorBpc =
-                    pProposedApiHead->hdr.colorBpc;
-                pApiHeadState->attributes.colorRange =
-                    pProposedApiHead->hdr.colorRange;
-
-                if (pProposedApiHead->hdr.tf == NVKMS_OUTPUT_TF_PQ) {
-                    nvCancelSDRTransitionTimer(pApiHeadState);
-                } else if (pApiHeadState->tf == NVKMS_OUTPUT_TF_PQ) {
-                    ScheduleSDRTransitionTimer(pDispEvo, apiHead);
-                }
-
-                pApiHeadState->tf = pProposedApiHead->hdr.tf;
-
-                nvUpdateInfoFrames(pDpyEvo);
-            }
-
-            if (pProposedApiHead->dirty.viewPortPointIn) {
-                pApiHeadState->viewPortPointIn =
-                    pProposedApiHead->viewPortPointIn;
-            }
+            FlipEvoOneApiHead(pDispEvo, apiHead, pWorkArea, allowFlipLock,
+                              &pWorkArea->updateState);
         }
 
         if (!skipUpdate) {
-            pDevEvo->hal->Update(pDevEvo, &updateState, TRUE /* releaseElv */);
+            nvEvoFlipUpdate(pDispEvo, &pWorkArea->updateState);
         }
     }
 
-    nvPostFlip(pDevEvo, pWorkArea, skipUpdate, applyAllowVrr, &vrrSemaphoreIndex);
+    nvPostFlip(pDevEvo, pWorkArea, skipUpdate, applyAllowVrrApiHeadMasks,
+               &vrrSemaphoreIndex);
 
-    FillNvKmsFlipReply(pDevEvo, pWorkArea, applyAllowVrr, vrrSemaphoreIndex,
+    NvBool replyApplyVrr = NV_FALSE;
+
+    for (sd = 0; sd < pDevEvo->numSubDevices; sd++) {
+        if (applyAllowVrrApiHeadMasks[sd] > 0) {
+            replyApplyVrr = NV_TRUE;
+            break;
+        }
+    }
+
+    FillNvKmsFlipReply(pDevEvo, pWorkArea, replyApplyVrr, vrrSemaphoreIndex,
                        pFlipHead, numFlipHeads, reply);
 
     /* fall through */
 
 done:
-
+    CleanupNvKmsFlipWorkArea(pDevEvo, pWorkArea);
     nvPreallocRelease(pDevEvo, PREALLOC_TYPE_FLIP_WORK_AREA);
+    if (reply) {
+        reply->flipResult = result;
+    }
 
     return ret;
 }
@@ -583,17 +757,18 @@ void nvApiHeadGetLayerSurfaceArray(const NVDispEvoRec *pDispEvo,
     const NvU32 sd = pDispEvo->displayOwner;
     const NVDispApiHeadStateEvoRec *pApiHeadState =
         &pDispEvo->apiHeadState[apiHead];
-    NvU32 head, headCount;
+    NvU32 head;
+    NvBool firstHead;
 
     nvAssert(apiHead != NV_INVALID_HEAD);
 
-    headCount = 0;
+    firstHead = NV_TRUE;
     FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, head) {
         const NVEvoSubDevHeadStateRec *pSdHeadState =
             &pDispEvo->pDevEvo->gpus[sd].headState[head];
         NvU8 eye;
 
-        if (headCount == 0) {
+        if (firstHead) {
             for (eye = NVKMS_LEFT; eye < NVKMS_MAX_EYES; eye++) {
                 pSurfaceEvos[eye] =
                     pSdHeadState->layer[layer].pSurfaceEvo[eye];
@@ -605,7 +780,7 @@ void nvApiHeadGetLayerSurfaceArray(const NVDispEvoRec *pDispEvo,
             }
         }
 
-        headCount++;
+        firstHead = NV_FALSE;
     }
 }
 
@@ -618,16 +793,17 @@ void nvApiHeadGetCursorInfo(const NVDispEvoRec *pDispEvo,
     const NvU32 sd = pDispEvo->displayOwner;
     const NVDispApiHeadStateEvoRec *pApiHeadState =
         &pDispEvo->apiHeadState[apiHead];
-    NvU32 head, headCount;
+    NvU32 head;
+    NvBool firstHead;
 
     nvAssert(apiHead != NV_INVALID_HEAD);
 
-    headCount = 0;
+    firstHead = NV_TRUE;
     FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, head) {
         const NVEvoSubDevHeadStateRec *pSdHeadState =
             &pDispEvo->pDevEvo->gpus[sd].headState[head];
 
-        if (headCount == 0) {
+        if (firstHead) {
             *ppSurfaceEvo = pSdHeadState->cursor.pSurfaceEvo;
             *x = pSdHeadState->cursor.x;
             *y = pSdHeadState->cursor.y;
@@ -637,7 +813,7 @@ void nvApiHeadGetCursorInfo(const NVDispEvoRec *pDispEvo,
             nvAssert(*y == pSdHeadState->cursor.y);
         }
 
-        headCount++;
+        firstHead = NV_FALSE;
     }
 }
 
@@ -651,18 +827,19 @@ void nvApiHeadSetViewportPointIn(const NVDispEvoRec *pDispEvo,
     const NVDispApiHeadStateEvoRec *pApiHeadState =
         &pDispEvo->apiHeadState[apiHead];
     NvU16 hwViewportInWidth;
-    NvU32 head, headCount;
+    NvU32 head;
+    NvBool firstHead;
 
     nvAssert(apiHead != NV_INVALID_HEAD);
 
-    headCount = 0;
+    firstHead = NV_TRUE;
     FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, head) {
         const NVDispHeadStateEvoRec *pHeadState =
             &pDispEvo->headState[head];
         const NVHwModeTimingsEvo *pTimings =
             &pHeadState->timings;
 
-        if (headCount == 0) {
+        if (firstHead) {
             hwViewportInWidth = pTimings->viewPort.in.width;
         } else {
             nvAssert(hwViewportInWidth == pTimings->viewPort.in.width);
@@ -670,14 +847,14 @@ void nvApiHeadSetViewportPointIn(const NVDispEvoRec *pDispEvo,
 
         nvPushEvoSubDevMaskDisp(pDispEvo);
         pDevEvo->hal->SetViewportPointIn(pDevEvo, head,
-            x + (hwViewportInWidth * pHeadState->tilePosition), y,
+            x + (hwViewportInWidth * pHeadState->mergeHeadSection), y,
             &updateState);
         nvPopEvoSubDevMask(pDevEvo);
 
-        headCount++;
+        firstHead = NV_FALSE;
     }
 
-    if (headCount != 0) {
+    if (!firstHead) {
         nvEvoUpdateAndKickOff(pDispEvo, FALSE /* sync */, &updateState,
                               TRUE /* releaseElv */);
     }
@@ -689,19 +866,20 @@ NvU32 nvApiHeadGetActiveViewportOffset(NVDispEvoRec *pDispEvo,
     NVDevEvoRec *pDevEvo = pDispEvo->pDevEvo;
     const NVDispApiHeadStateEvoRec *pApiHeadState =
         &pDispEvo->apiHeadState[apiHead];
-    NvU32 head, headCount;
+    NvU32 head;
     NvU32 offset = 0;
+    NvBool firstHead;
 
     nvAssert(apiHead != NV_INVALID_HEAD);
 
-    headCount = 0;
+    firstHead = NV_TRUE;
     FOR_EACH_EVO_HW_HEAD_IN_MASK(pApiHeadState->hwHeadsMask, head) {
-        if (headCount == 0) {
+        if (firstHead) {
             offset = pDevEvo->hal->GetActiveViewportOffset(pDispEvo, head);
         } else {
             nvAssert(offset == pDevEvo->hal->GetActiveViewportOffset(pDispEvo, head));
         }
-        headCount++;
+        firstHead = NV_FALSE;
     }
 
     return offset;
@@ -967,7 +1145,7 @@ void nvIdleLayerChannels(NVDevEvoRec *pDevEvo,
         }
 
         if (!allIdle) {
-            if (nvExceedsTimeoutUSec(&startTime, timeout)) {
+            if (nvExceedsTimeoutUSec(pDevEvo, &startTime, timeout)) {
                 break;
             }
             nvkms_yield();
