@@ -33,10 +33,71 @@
 
 // NV0000_CTRL_CMD_OS_UNIX_IMPORT_OBJECT_FROM_FD
 #include "ctrl/ctrl0000/ctrl0000unix.h"
+// NV0000_CTRL_CMD_CLIENT_GET_ADDR_SPACE_TYPE_VIDMEM
+#include "ctrl/ctrl0000/ctrl0000client.h"
+
+/* NV0041_CTRL_SURFACE_INFO */
+#include "ctrl/ctrl0041.h"
 
 /* NV01_MEMORY_SYSTEM_OS_DESCRIPTOR */
 #include "class/cl0071.h"
 
+static void CpuUnmapSurface(
+    NVDevEvoPtr pDevEvo,
+    NVSurfaceEvoPtr pSurfaceEvo)
+{
+    const NvU32 planeIndex = 0;
+    NvU32 sd;
+
+    if (pSurfaceEvo->planes[planeIndex].rmHandle == 0) {
+        return;
+    }
+
+    for (sd = 0; sd < pDevEvo->numSubDevices; sd++) {
+
+        if (pSurfaceEvo->cpuAddress[sd] != NULL) {
+            nvRmApiUnmapMemory(nvEvoGlobal.clientHandle,
+                               pDevEvo->pSubDevices[sd]->handle,
+                               pSurfaceEvo->planes[planeIndex].rmHandle,
+                               pSurfaceEvo->cpuAddress[sd],
+                               0);
+            pSurfaceEvo->cpuAddress[sd] = NULL;
+        }
+    }
+}
+
+NvBool nvEvoCpuMapSurface(
+    NVDevEvoPtr pDevEvo,
+    NVSurfaceEvoPtr pSurfaceEvo)
+{
+    const NvU32 planeIndex = 0;
+    NvU32 sd;
+
+    /*
+     * We should only be called here with surfaces that contain a single plane.
+     */
+    nvAssert(nvKmsGetSurfaceMemoryFormatInfo(pSurfaceEvo->format)->numPlanes == 1);
+
+    for (sd = 0; sd < pDevEvo->numSubDevices; sd++) {
+
+        NvU32 result = nvRmApiMapMemory(
+                        nvEvoGlobal.clientHandle,
+                        pDevEvo->pSubDevices[sd]->handle,
+                        pSurfaceEvo->planes[planeIndex].rmHandle,
+                        0,
+                        pSurfaceEvo->planes[planeIndex].rmObjectSizeInBytes,
+                        (void **) &pSurfaceEvo->cpuAddress[sd],
+                        0);
+
+        if (result != NVOS_STATUS_SUCCESS) {
+            CpuUnmapSurface(pDevEvo, pSurfaceEvo);
+            return FALSE;
+        }
+    }
+
+    return TRUE;
+}
+
 static void FreeSurfaceEvoStruct(NVSurfaceEvoPtr pSurfaceEvo)
 {
     if (pSurfaceEvo == NULL) {
@@ -54,7 +115,6 @@ static void FreeSurfaceEvoStruct(NVSurfaceEvoPtr pSurfaceEvo)
 static void FreeSurfaceEvoRm(NVDevEvoPtr pDevEvo, NVSurfaceEvoPtr pSurfaceEvo)
 {
     NvU64 structRefCnt;
-    NvU32 firstPlaneRmHandle;
     NvU8 planeIndex;
 
     if ((pDevEvo == NULL) || (pSurfaceEvo == NULL)) {
@@ -64,30 +124,16 @@ static void FreeSurfaceEvoRm(NVDevEvoPtr pDevEvo, NVSurfaceEvoPtr pSurfaceEvo)
     nvAssert(pSurfaceEvo->rmRefCnt == 0);
 
     FOR_ALL_VALID_PLANES(planeIndex, pSurfaceEvo) {
-        nvRmEvoFreeDispContextDMA(pDevEvo,
-                                  &pSurfaceEvo->planes[planeIndex].ctxDma);
+        pDevEvo->hal->FreeSurfaceDescriptor(pDevEvo,
+                                            nvEvoGlobal.clientHandle,
+                                            &pSurfaceEvo->planes[planeIndex].surfaceDesc);
     }
 
-    firstPlaneRmHandle = pSurfaceEvo->planes[0].rmHandle;
-
-    if (firstPlaneRmHandle != 0) {
-
-        NvU32 sd;
-
-        for (sd = 0; sd < pDevEvo->numSubDevices; sd++) {
-
-            if (pSurfaceEvo->cpuAddress[sd] != NULL) {
-                nvRmApiUnmapMemory(nvEvoGlobal.clientHandle,
-                                   pDevEvo->pSubDevices[sd]->handle,
-                                   firstPlaneRmHandle,
-                                   pSurfaceEvo->cpuAddress[sd],
-                                   0);
-                pSurfaceEvo->cpuAddress[sd] = NULL;
-            }
-        }
+    CpuUnmapSurface(pDevEvo, pSurfaceEvo);
 
+    if (pSurfaceEvo->planes[0].rmHandle != 0) {
         nvHsUnmapSurfaceFromDevice(pDevEvo,
-                                   firstPlaneRmHandle,
+                                   pSurfaceEvo->planes[0].rmHandle,
                                    pSurfaceEvo->gpuAddress);
     }
 
@@ -328,6 +374,72 @@ static NvBool ValidateRegisterSurfaceRequest(
     return TRUE;
 }
 
+static NvBool ValidateSurfaceAddressSpace(
+    NVDevEvoPtr pDevEvo,
+    const struct NvKmsRegisterSurfaceRequest *pRequest,
+    NvU32 rmHandle)
+{
+    NV0041_CTRL_GET_SURFACE_INFO_PARAMS surfaceInfoParams = {};
+    NV0041_CTRL_SURFACE_INFO surfaceInfo = {};
+    NV_STATUS status;
+
+    /*
+     * Don't do these checks on tegra. Tegra has different capabilities.
+     * Here we always say display is possible so we never fail framebuffer
+     * creation.
+     */
+    if (pDevEvo->isSOCDisplay) {
+        return TRUE;
+    }
+
+    /*
+     * Don't do these checks for surfaces that do not need access to display
+     * hardware.
+     */
+    if (pRequest->noDisplayHardwareAccess) {
+        return TRUE;
+    }
+
+    /*
+     * If the memory is not isochronous, the memory will not be scanned out to a
+     * display. The checks are not needed for such memory types.
+     */
+    if (pRequest->isoType != NVKMS_MEMORY_ISO) {
+        return TRUE;
+    }
+
+    /*
+     * Check if the memory we are registering this surface with is valid. We
+     * cannot scan out sysmem or compressed buffers.
+     *
+     * If we cannot use this memory for display it may be resident in sysmem
+     * or may belong to another GPU.
+     */
+    surfaceInfo.index = NV0041_CTRL_SURFACE_INFO_INDEX_ADDR_SPACE_TYPE;
+
+    surfaceInfoParams.surfaceInfoListSize = 1;
+    surfaceInfoParams.surfaceInfoList = (NvP64)&surfaceInfo;
+
+    status = nvRmApiControl(nvEvoGlobal.clientHandle,
+                            rmHandle,
+                            NV0041_CTRL_CMD_GET_SURFACE_INFO,
+                            &surfaceInfoParams,
+                            sizeof(surfaceInfoParams));
+    if (status != NV_OK) {
+        nvEvoLogDevDebug(pDevEvo, EVO_LOG_ERROR,
+                          "Failed to get memory location of RM memory object 0x%x",
+                          rmHandle);
+        return FALSE;
+    }
+
+    if (surfaceInfo.data != NV0000_CTRL_CMD_CLIENT_GET_ADDR_SPACE_TYPE_VIDMEM) {
+        nvEvoLogDevDebug(pDevEvo, EVO_LOG_ERROR,
+                         "Memory used for surface not appropriate for scanout");
+        return FALSE;
+    }
+
+    return TRUE;
+}
 
 void nvEvoRegisterSurface(NVDevEvoPtr pDevEvo,
                           struct NvKmsPerOpenDev *pOpenDev,
@@ -483,21 +595,23 @@ void nvEvoRegisterSurface(NVDevEvoPtr pDevEvo,
             goto fail;
         }
 
+        if (!ValidateSurfaceAddressSpace(pDevEvo, pRequest, planeRmHandle)) {
+            goto fail;
+        }
+
         /* XXX Validate sizeInBytes: can we query the surface size from RM? */
 
         if (!pRequest->noDisplayHardwareAccess) {
-
-            const NvU32 planeCtxDma =
-                nvRmEvoAllocateAndBindDispContextDMA(
-                    pDevEvo,
-                    planeRmHandle,
-                    pRequest->layout,
-                    pRequest->planes[planeIndex].rmObjectSizeInBytes - 1);
-            if (!planeCtxDma) {
+            NvU32 ret =
+                nvRmAllocAndBindSurfaceDescriptor(
+                        pDevEvo,
+                        planeRmHandle,
+                        pRequest->layout,
+                        pRequest->planes[planeIndex].rmObjectSizeInBytes - 1,
+                        &pSurfaceEvo->planes[planeIndex].surfaceDesc);
+            if (ret != NVOS_STATUS_SUCCESS) {
                 goto fail;
             }
-
-            pSurfaceEvo->planes[planeIndex].ctxDma = planeCtxDma;
         }
 
         pSurfaceEvo->planes[planeIndex].pitch =
@@ -508,7 +622,7 @@ void nvEvoRegisterSurface(NVDevEvoPtr pDevEvo,
                             pRequest->planes[planeIndex].rmObjectSizeInBytes;
     }
 
-    pSurfaceEvo->requireCtxDma = !pRequest->noDisplayHardwareAccess;
+    pSurfaceEvo->requireDisplayHardwareAccess = !pRequest->noDisplayHardwareAccess;
     pSurfaceEvo->noDisplayCaching = pRequest->noDisplayCaching;
 
     /*
@@ -537,26 +651,14 @@ void nvEvoRegisterSurface(NVDevEvoPtr pDevEvo,
     /*
      * Map the first plane of the surface only into the CPU's address space.
      * This is the only valid plane since we would have already rejected
-     * multi-planar semaphore requests earlier.
+     * multi-planar NISO surface requests earlier in
+     *
+     * nvEvoRegisterSurface() => ValidateRegisterSurfaceRequest() =>
+     * ValidatePlaneProperties().
      */
     if (needCpuMapping) {
-
-        NvU32 sd;
-
-        for (sd = 0; sd < pDevEvo->numSubDevices; sd++) {
-
-            result = nvRmApiMapMemory(
-                    nvEvoGlobal.clientHandle,
-                    pDevEvo->pSubDevices[sd]->handle,
-                    pSurfaceEvo->planes[0].rmHandle,
-                    0,
-                    pRequest->planes[0].rmObjectSizeInBytes,
-                    (void **) &pSurfaceEvo->cpuAddress[sd],
-                    0);
-
-            if (result != NVOS_STATUS_SUCCESS) {
-                goto fail;
-            }
+        if (!nvEvoCpuMapSurface(pDevEvo, pSurfaceEvo)) {
+            goto fail;
         }
     }
 
@@ -594,7 +696,7 @@ struct ClearSurfaceUsageCache {
         } layer[NVKMS_MAX_LAYERS_PER_HEAD];
 
         NvBool flipCursorToNull         : 1;
-    } head[NVKMS_MAX_SUBDEVICES][NVKMS_MAX_HEADS_PER_DISP];
+    } apiHead[NVKMS_MAX_SUBDEVICES][NVKMS_MAX_HEADS_PER_DISP];
 };
 
 /*
@@ -607,69 +709,67 @@ ClearSurfaceUsageCollect(NVDevEvoPtr pDevEvo,
                          struct ClearSurfaceUsageCache *pCache)
 {
     NVDispEvoPtr pDispEvo;
-    NvU32 head, sd;
+    NvU32 apiHead, sd;
 
     FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
 
-        for (head = 0; head < pDevEvo->numHeads; head++) {
-
-            const NVEvoSubDevHeadStateRec *pSdHeadState =
-                &pDevEvo->gpus[sd].headState[head];
-            const NVFlipChannelEvoHwState *pMainFlipState =
-                &pSdHeadState->layer[NVKMS_MAIN_LAYER];
+        for (apiHead = 0; apiHead < pDevEvo->numApiHeads; apiHead++) {
+            NvU32 usageMaskOneHead = nvCollectSurfaceUsageMaskOneApiHead(pDispEvo,
+                apiHead, pSurfaceEvo);
+            NvU32 usageMaskMainLayer = DRF_IDX_VAL(_SURFACE,
+                _USAGE_MASK, _LAYER, NVKMS_MAIN_LAYER, usageMaskOneHead);
             NvU32 layer;
 
-            if (!nvHeadIsActive(pDispEvo, head)) {
-                continue;
-            }
-
             /*
              * XXX NVKMS TODO: flip across heads/subdevices for all scenarios
              * that are flip locked.
              */
 
-            if (!pMainFlipState->syncObject.usingSyncpt &&
-                (pSurfaceEvo == pMainFlipState->syncObject.u.semaphores.acquireSurface.pSurfaceEvo ||
-                 pSurfaceEvo == pMainFlipState->syncObject.u.semaphores.releaseSurface.pSurfaceEvo)) {
-                pCache->head[sd][head].layer[NVKMS_MAIN_LAYER].flipSemaphoreToNull = TRUE;
+            if (FLD_TEST_DRF(_SURFACE, _USAGE_MASK_LAYER, _SEMAPHORE,
+                    _ENABLE, usageMaskMainLayer)) {
+                pCache->apiHead[sd][apiHead].layer[NVKMS_MAIN_LAYER].
+                    flipSemaphoreToNull = TRUE;
             }
 
-            if (pSurfaceEvo == pMainFlipState->pSurfaceEvo[NVKMS_LEFT] ||
-                pSurfaceEvo == pMainFlipState->pSurfaceEvo[NVKMS_RIGHT] ||
-                pSurfaceEvo == pMainFlipState->completionNotifier.surface.pSurfaceEvo) {
-                pCache->head[sd][head].layer[NVKMS_MAIN_LAYER].flipToNull = TRUE;
+            if (FLD_TEST_DRF(_SURFACE, _USAGE_MASK_LAYER, _NOTIFIER,
+                    _ENABLE, usageMaskMainLayer) ||
+                    FLD_TEST_DRF(_SURFACE, _USAGE_MASK_LAYER, _SCANOUT,
+                        _ENABLE, usageMaskMainLayer)) {
+                pCache->apiHead[sd][apiHead].layer[NVKMS_MAIN_LAYER].
+                    flipToNull = TRUE;
             }
 
-            for (layer = 0; layer < pDevEvo->head[head].numLayers; layer++) {
-                const NVFlipChannelEvoHwState *pLayerFlipState =
-                    &pSdHeadState->layer[layer];
+            for (layer = 0;
+                 layer < pDevEvo->apiHead[apiHead].numLayers; layer++) {
+                NvU32 usageMaskOneLayer = DRF_IDX_VAL(_SURFACE,
+                    _USAGE_MASK, _LAYER, layer, usageMaskOneHead);
 
                 if (layer == NVKMS_MAIN_LAYER) {
                     continue;
                 }
 
-                if (pSurfaceEvo == pLayerFlipState->pSurfaceEvo[NVKMS_LEFT] ||
-                    pSurfaceEvo == pLayerFlipState->pSurfaceEvo[NVKMS_RIGHT] ||
-                    pSurfaceEvo == pLayerFlipState->completionNotifier.surface.pSurfaceEvo ||
-                    (!pLayerFlipState->syncObject.usingSyncpt &&
-                     (pSurfaceEvo == pLayerFlipState->syncObject.u.semaphores.acquireSurface.pSurfaceEvo ||
-                      pSurfaceEvo == pLayerFlipState->syncObject.u.semaphores.releaseSurface.pSurfaceEvo))) {
-                    pCache->head[sd][head].layer[layer].flipToNull = TRUE;
-                }
-
-                /*
-                 * EVO requires that, when flipping the base channel (aka main layer) to
-                 * NULL, overlay channel is also flipped to NULL.
-                 */
-                if (pCache->head[sd][head].layer[NVKMS_MAIN_LAYER].flipToNull &&
-                    (pLayerFlipState->pSurfaceEvo[NVKMS_LEFT] != NULL ||
-                     pLayerFlipState->pSurfaceEvo[NVKMS_RIGHT] != NULL)) {
-                    pCache->head[sd][head].layer[layer].flipToNull = TRUE;
+                if (usageMaskOneLayer != 0x0) {
+                    pCache->apiHead[sd][apiHead].layer[layer].
+                        flipToNull = TRUE;
+                } if (pCache->apiHead[sd][apiHead].layer[NVKMS_MAIN_LAYER].
+                        flipToNull) {
+                    NVSurfaceEvoPtr pSurfaceEvos[NVKMS_MAX_EYES] = { };
+                    /*
+                     * EVO requires that, when flipping the base channel
+                     * (aka main layer) to NULL, overlay channel is also
+                     * flipped to NULL.
+                     */
+                    if ((pSurfaceEvos[NVKMS_LEFT] != NULL) ||
+                            (pSurfaceEvos[NVKMS_RIGHT] != NULL)) {
+                        pCache->apiHead[sd][apiHead].layer[layer].
+                            flipToNull = TRUE;
+                     }
                 }
             }
 
-            if (pSurfaceEvo == pSdHeadState->cursor.pSurfaceEvo) {
-                pCache->head[sd][head].flipCursorToNull = TRUE;
+            if (FLD_TEST_DRF(_SURFACE, _USAGE_MASK, _CURSOR,
+                    _ENABLE, usageMaskOneHead) != 0x0) {
+                pCache->apiHead[sd][apiHead].flipCursorToNull = TRUE;
             }
         }
     }
@@ -700,11 +800,13 @@ ClearSurfaceUsageApply(NVDevEvoPtr pDevEvo,
                        NvBool skipUpdate)
 {
     NVDispEvoPtr pDispEvo;
-    NvU32 head, sd;
-    NvBool found = FALSE;
-    struct NvKmsFlipRequest *request = nvCalloc(1, sizeof(*request));
+    NvU32 apiHead, sd;
+    const NvU32 maxApiHeads = pDevEvo->numApiHeads * pDevEvo->numSubDevices;
+    struct NvKmsFlipRequestOneHead *pFlipApiHead =
+        nvCalloc(1, sizeof(*pFlipApiHead) * maxApiHeads);
+    NvU32 numFlipApiHeads = 0;
 
-    if (request == NULL) {
+    if (pFlipApiHead == NULL) {
         nvAssert(!"Failed to allocate memory");
         return;
     }
@@ -712,46 +814,58 @@ ClearSurfaceUsageApply(NVDevEvoPtr pDevEvo,
     /* 1. Issue a flip of any overlay layer to NULL */
     FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
 
-        for (head = 0; head < pDevEvo->numHeads; head++) {
+        for (apiHead = 0; apiHead < pDevEvo->numApiHeads; apiHead++) {
 
-            struct NvKmsFlipCommonParams *pRequestOneHead =
-                &request->sd[sd].head[head];
+            struct NvKmsFlipCommonParams *pRequestOneApiHead =
+                &pFlipApiHead[numFlipApiHeads].flip;
             NvU32 layer;
+            NvBool found = FALSE;
 
-            if (!nvHeadIsActive(pDispEvo, head)) {
+            if (!nvApiHeadIsActive(pDispEvo, apiHead)) {
                 continue;
             }
 
-            for (layer = 0; layer < pDevEvo->head[head].numLayers; layer++) {
+            for (layer = 0;
+                 layer < pDevEvo->apiHead[apiHead].numLayers; layer++) {
 
                 if (layer == NVKMS_MAIN_LAYER) {
                     continue;
                 }
 
-                if (pCache->head[sd][head].layer[layer].flipToNull) {
-                    pRequestOneHead->layer[layer].surface.specified = TRUE;
+                if (pCache->apiHead[sd][apiHead].layer[layer].flipToNull) {
+                    pRequestOneApiHead->layer[layer].surface.specified = TRUE;
                     // No need to specify sizeIn/sizeOut as we are flipping NULL surface.
-                    pRequestOneHead->layer[layer].compositionParams.specified = TRUE;
-                    pRequestOneHead->layer[layer].syncObjects.specified = TRUE;
-                    pRequestOneHead->layer[layer].completionNotifier.specified = TRUE;
+                    pRequestOneApiHead->layer[layer].compositionParams.specified = TRUE;
+                    pRequestOneApiHead->layer[layer].syncObjects.specified = TRUE;
+                    pRequestOneApiHead->layer[layer].completionNotifier.specified = TRUE;
 
-                    request->sd[sd].requestedHeadsBitMask |= NVBIT(head);
                     found = TRUE;
 
-                    pCache->head[sd][head].layer[layer].needToIdle = TRUE;
+                    pCache->apiHead[sd][apiHead].layer[layer].needToIdle = TRUE;
                 }
             }
+
+            if (found) {
+                pFlipApiHead[numFlipApiHeads].sd = sd;
+                pFlipApiHead[numFlipApiHeads].head = apiHead;
+                numFlipApiHeads++;
+                nvAssert(numFlipApiHeads <= maxApiHeads);
+            }
         }
     }
 
-    if (found) {
-        request->commit = NV_TRUE;
-
-        nvFlipEvo(pDevEvo, pDevEvo->pNvKmsOpenDev, request, NULL, skipUpdate,
+    if (numFlipApiHeads > 0) {
+        nvFlipEvo(pDevEvo, pDevEvo->pNvKmsOpenDev,
+                  pFlipApiHead,
+                  numFlipApiHeads,
+                  TRUE  /* commit */,
+                  NULL  /* pReply */,
+                  skipUpdate,
                   FALSE /* allowFlipLock */);
 
-        nvkms_memset(request, 0, sizeof(*request));
-        found = FALSE;
+        nvkms_memset(pFlipApiHead, 0,
+            sizeof(pFlipApiHead[0]) * numFlipApiHeads);
+        numFlipApiHeads = 0;
     }
 
     /*
@@ -766,46 +880,57 @@ ClearSurfaceUsageApply(NVDevEvoPtr pDevEvo,
     /* 2. Issue a flip of any main layer to NULL */
     FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
 
-        for (head = 0; head < pDevEvo->numHeads; head++) {
+        for (apiHead = 0; apiHead < pDevEvo->numApiHeads; apiHead++) {
 
-            struct NvKmsFlipCommonParams *pRequestOneHead =
-                &request->sd[sd].head[head];
+            struct NvKmsFlipCommonParams *pRequestOneApiHead =
+                &pFlipApiHead[numFlipApiHeads].flip;
+            NvBool found = FALSE;
 
-            if (!nvHeadIsActive(pDispEvo, head)) {
+            if (!nvApiHeadIsActive(pDispEvo, apiHead)) {
                 continue;
             }
 
-            if (pCache->head[sd][head].layer[NVKMS_MAIN_LAYER].flipToNull ||
-                pCache->head[sd][head].layer[NVKMS_MAIN_LAYER].flipSemaphoreToNull) {
+            if (pCache->apiHead[sd][apiHead].layer[NVKMS_MAIN_LAYER].flipToNull ||
+                pCache->apiHead[sd][apiHead].layer[NVKMS_MAIN_LAYER].flipSemaphoreToNull) {
 
-                if (pCache->head[sd][head].layer[NVKMS_MAIN_LAYER].flipToNull) {
-                    pRequestOneHead->layer[NVKMS_MAIN_LAYER].surface.specified = TRUE;
+                if (pCache->apiHead[sd][apiHead].layer[NVKMS_MAIN_LAYER].flipToNull) {
+                    pRequestOneApiHead->layer[NVKMS_MAIN_LAYER].surface.specified = TRUE;
                     // No need to specify sizeIn/sizeOut as we are flipping NULL surface.
-                    pRequestOneHead->layer[NVKMS_MAIN_LAYER].completionNotifier.specified = TRUE;
+                    pRequestOneApiHead->layer[NVKMS_MAIN_LAYER].completionNotifier.specified = TRUE;
 
-                    pCache->head[sd][head].layer[NVKMS_MAIN_LAYER].needToIdle = TRUE;
+                    pCache->apiHead[sd][apiHead].layer[NVKMS_MAIN_LAYER].needToIdle = TRUE;
                 }
 
                 /* XXX arguably we should also idle for this case, but we
                  * don't currently have a way to do so without also
                  * clearing the ISO surface */
-                pRequestOneHead->layer[NVKMS_MAIN_LAYER].syncObjects.val.useSyncpt = FALSE;
-                pRequestOneHead->layer[NVKMS_MAIN_LAYER].syncObjects.specified = TRUE;
+                pRequestOneApiHead->layer[NVKMS_MAIN_LAYER].syncObjects.val.useSyncpt = FALSE;
+                pRequestOneApiHead->layer[NVKMS_MAIN_LAYER].syncObjects.specified = TRUE;
 
-                request->sd[sd].requestedHeadsBitMask |= NVBIT(head);
                 found = TRUE;
             }
+
+            if (found) {
+                pFlipApiHead[numFlipApiHeads].sd = sd;
+                pFlipApiHead[numFlipApiHeads].head = apiHead;
+                numFlipApiHeads++;
+                nvAssert(numFlipApiHeads <= maxApiHeads);
+            }
         }
     }
 
-    if (found) {
-        request->commit = NV_TRUE;
-
-        nvFlipEvo(pDevEvo, pDevEvo->pNvKmsOpenDev, request, NULL, skipUpdate,
+    if (numFlipApiHeads > 0) {
+        nvFlipEvo(pDevEvo, pDevEvo->pNvKmsOpenDev,
+                  pFlipApiHead,
+                  numFlipApiHeads,
+                  TRUE  /* commit */,
+                  NULL  /* pReply */,
+                  skipUpdate,
                   FALSE /* allowFlipLock */);
 
-        nvkms_memset(request, 0, sizeof(*request));
-        found = FALSE;
+        nvkms_memset(pFlipApiHead, 0,
+            sizeof(pFlipApiHead[0]) * numFlipApiHeads);
+        numFlipApiHeads = 0;
     }
 
     /*
@@ -814,111 +939,51 @@ ClearSurfaceUsageApply(NVDevEvoPtr pDevEvo,
      *    forcibly idle any problematic channels.
      */
     if (!skipUpdate) {
-        NvU64 startTime = 0;
-        const NvU32 timeout = 500000; // .5 seconds
-        NvBool allIdle;
-
-        do {
-            allIdle = TRUE;
-            FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
-
-                for (head = 0; head < pDevEvo->numHeads; head++) {
-                    NvU32 layer;
-
-                    if (!nvHeadIsActive(pDispEvo, head)) {
-                        continue;
+        NvU32 layerMaskPerSdApiHead[NVKMS_MAX_SUBDEVICES]
+            [NVKMS_MAX_HEADS_PER_DISP] = { };
+        FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
+            for (apiHead = 0; apiHead < pDevEvo->numApiHeads; apiHead++) {
+                for (NvU32 layer = 0;
+                     layer < pDevEvo->apiHead[apiHead].numLayers; layer++) {
+                    if (pCache->apiHead[sd][apiHead].layer[layer].needToIdle) {
+                        layerMaskPerSdApiHead[sd][apiHead] |= NVBIT(layer);
                     }
-
-                    for (layer = 0; layer < pDevEvo->head[head].numLayers; layer++) {
-                        NvBool isMethodPending;
-
-                        if (!pCache->head[sd][head].layer[layer].needToIdle) {
-                            continue;
-                        }
-
-                        if (pDevEvo->hal->IsChannelMethodPending(
-                                pDevEvo, pDevEvo->head[head].layer[layer], sd,
-                                &isMethodPending) &&
-                            isMethodPending) {
-
-                            allIdle = FALSE;
-                        } else {
-                            /* This has been completed, no need to keep trying */
-                            pCache->head[sd][head].layer[layer].needToIdle = FALSE;
-                        }
-                    }
-                }
-            }
-
-            if (!allIdle) {
-                if (nvExceedsTimeoutUSec(&startTime, timeout)) {
-                    break;
-                }
-                nvkms_yield();
-            }
-        } while (!allIdle);
-
-        /* If we timed out above, force things to be idle. */
-        if (!allIdle) {
-            NVEvoIdleChannelState idleChannelState = { };
-            NvBool tryToForceIdle = FALSE;
-
-            FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
-
-                for (head = 0; head < pDevEvo->numHeads; head++) {
-                    NvU32 layer;
-
-                    if (!nvHeadIsActive(pDispEvo, head)) {
-                        continue;
-                    }
-
-                    for (layer = 0; layer < pDevEvo->head[head].numLayers; layer++) {
-                        if (pCache->head[sd][head].layer[layer].needToIdle) {
-                            idleChannelState.subdev[sd].channelMask |=
-                                pDevEvo->head[head].layer[layer]->channelMask;
-                            tryToForceIdle = TRUE;
-                        }
-                    }
-                }
-            }
-
-            if (tryToForceIdle) {
-                NvBool ret = pDevEvo->hal->ForceIdleSatelliteChannel(pDevEvo, &idleChannelState);
-                if (!ret) {
-                    nvAssert(ret);
                 }
             }
         }
+        nvIdleLayerChannels(pDevEvo, layerMaskPerSdApiHead);
     }
 
     /* 4. Issue a flip of any core channels to NULL */
     FOR_ALL_EVO_DISPLAYS(pDispEvo, sd, pDevEvo) {
 
-        for (head = 0; head < pDevEvo->numHeads; head++) {
+        for (apiHead = 0; apiHead < pDevEvo->numApiHeads; apiHead++) {
 
-            struct NvKmsFlipCommonParams *pRequestOneHead =
-                &request->sd[sd].head[head];
-
-            if (!nvHeadIsActive(pDispEvo, head)) {
+            if (!nvApiHeadIsActive(pDispEvo, apiHead)) {
                 continue;
             }
 
-            if (pCache->head[sd][head].flipCursorToNull) {
-                pRequestOneHead->cursor.imageSpecified = TRUE;
-                request->sd[sd].requestedHeadsBitMask |= NVBIT(head);
-                found = TRUE;
+            if (pCache->apiHead[sd][apiHead].flipCursorToNull) {
+                pFlipApiHead[numFlipApiHeads].flip.cursor.imageSpecified = TRUE;
+                pFlipApiHead[numFlipApiHeads].sd = sd;
+                pFlipApiHead[numFlipApiHeads].head = apiHead;
+                numFlipApiHeads++;
+                nvAssert(numFlipApiHeads <= maxApiHeads);
             }
         }
     }
 
-    if (found) {
-        request->commit = NV_TRUE;
-
-        nvFlipEvo(pDevEvo, pDevEvo->pNvKmsOpenDev, request, NULL, skipUpdate,
+    if (numFlipApiHeads > 0) {
+        nvFlipEvo(pDevEvo, pDevEvo->pNvKmsOpenDev,
+                  pFlipApiHead,
+                  numFlipApiHeads,
+                  TRUE  /* commit */,
+                  NULL  /* pReply */,
+                  skipUpdate,
                   FALSE /* allowFlipLock */);
     }
 
-    nvFree(request);
+    nvFree(pFlipApiHead);
 }
 
 /*
@@ -1003,7 +1068,8 @@ void nvEvoFreeClientSurfaces(NVDevEvoPtr pDevEvo,
 void nvEvoUnregisterSurface(NVDevEvoPtr pDevEvo,
                             struct NvKmsPerOpenDev *pOpenDev,
                             NvKmsSurfaceHandle surfaceHandle,
-                            NvBool skipUpdate)
+                            NvBool skipUpdate,
+                            NvBool skipSync)
 {
     NVEvoApiHandlesRec *pOpenDevSurfaceHandles =
         nvGetSurfaceHandlesFromOpenDev(pOpenDev);
@@ -1037,7 +1103,7 @@ void nvEvoUnregisterSurface(NVDevEvoPtr pDevEvo,
     /* Remove the handle from the calling client's namespace. */
     nvEvoDestroyApiHandle(pOpenDevSurfaceHandles, surfaceHandle);
 
-    nvEvoDecrementSurfaceRefCnts(pDevEvo, pSurfaceEvo);
+    nvEvoDecrementSurfaceRefCntsWithSync(pDevEvo, pSurfaceEvo, skipSync);
 }
 
 void nvEvoReleaseSurface(NVDevEvoPtr pDevEvo,
@@ -1077,6 +1143,13 @@ void nvEvoIncrementSurfaceRefCnts(NVSurfaceEvoPtr pSurfaceEvo)
 
 void nvEvoDecrementSurfaceRefCnts(NVDevEvoPtr pDevEvo,
                                   NVSurfaceEvoPtr pSurfaceEvo)
+{
+    nvEvoDecrementSurfaceRefCntsWithSync(pDevEvo, pSurfaceEvo, NV_FALSE);
+}
+
+void nvEvoDecrementSurfaceRefCntsWithSync(NVDevEvoPtr pDevEvo,
+                                          NVSurfaceEvoPtr pSurfaceEvo,
+                                          NvBool skipSync)
 {
     nvAssert(pSurfaceEvo->rmRefCnt >= 1);
     pSurfaceEvo->rmRefCnt--;
@@ -1089,38 +1162,8 @@ void nvEvoDecrementSurfaceRefCnts(NVDevEvoPtr pDevEvo,
          * GLS hasn't had the opportunity to release semaphores with pending
          * flips. (Bug 2050970)
          */
-        if (pSurfaceEvo->requireCtxDma) {
-            /*
-             * XXX NVKMS TODO
-             * Make the sync more efficient: we only need to sync if the
-             * in-flight methods flip away from this surface.
-             */
-            NvU32 head;
-
-            /*
-             * If the core channel is no longer allocated, we don't need to
-             * sync.  This assumes the channels are allocated/deallocated
-             * together.
-             */
-            if (pDevEvo->core) {
-
-                if (pDevEvo->hal->ClearSurfaceUsage != NULL) {
-                    pDevEvo->hal->ClearSurfaceUsage(pDevEvo, pSurfaceEvo);
-                }
-
-                nvRMSyncEvoChannel(pDevEvo, pDevEvo->core, __LINE__);
-
-                for (head = 0; head < pDevEvo->numHeads; head++) {
-                    NvU32 layer;
-
-                    for (layer = 0; layer < pDevEvo->head[head].numLayers; layer++) {
-                        NVEvoChannelPtr pChannel =
-                            pDevEvo->head[head].layer[layer];
-
-                        nvRMSyncEvoChannel(pDevEvo, pChannel, __LINE__);
-                    }
-                }
-            }
+        if (!skipSync && pSurfaceEvo->requireDisplayHardwareAccess) {
+            nvEvoClearSurfaceUsage(pDevEvo, pSurfaceEvo);
         }
 
         FreeSurfaceEvoRm(pDevEvo, pSurfaceEvo);
@@ -1141,12 +1184,13 @@ static NVSurfaceEvoPtr GetSurfaceFromHandle(
     const NvKmsSurfaceHandle surfaceHandle,
     const NvBool isUsedByCursorChannel,
     const NvBool isUsedByLayerChannel,
-    const NvBool requireCtxDma)
+    const NvBool requireDisplayHardwareAccess,
+    const NvBool maybeUsedBy3d)
 {
     NVSurfaceEvoPtr pSurfaceEvo =
         nvEvoGetPointerFromApiHandle(pOpenDevSurfaceHandles, surfaceHandle);
 
-    nvAssert(requireCtxDma || (!isUsedByCursorChannel && !isUsedByLayerChannel));
+    nvAssert(requireDisplayHardwareAccess || (!isUsedByCursorChannel && !isUsedByLayerChannel));
 
     if (pSurfaceEvo == NULL) {
         return NULL;
@@ -1156,7 +1200,7 @@ static NVSurfaceEvoPtr GetSurfaceFromHandle(
         return NULL;
     }
 
-    if (requireCtxDma && !pSurfaceEvo->requireCtxDma) {
+    if (requireDisplayHardwareAccess && !pSurfaceEvo->requireDisplayHardwareAccess) {
         return NULL;
     }
 
@@ -1167,15 +1211,13 @@ static NVSurfaceEvoPtr GetSurfaceFromHandle(
     }
 
     /*
-     * XXX If !requireCtxDma, fetched surfaces aren't going to be accessed by
-     * the display hardware, so they shouldn't need to be checked by
-     * nvEvoGetHeadSetStoragePitchValue(). These surfaces will be used as a
-     * texture by the 3d engine. But previously all surfaces were checked by
+     * XXX If maybeUsedBy3d, the fetched surface may be used as a texture by the
+     * 3d engine.  Previously, all surfaces were checked by
      * nvEvoGetHeadSetStoragePitchValue() at registration time, and we don't
      * know if nvEvoGetHeadSetStoragePitchValue() was protecting us from any
      * surface dimensions that could cause trouble for the 3d engine.
      */
-    if (isUsedByLayerChannel || !requireCtxDma) {
+    if (isUsedByLayerChannel || maybeUsedBy3d) {
         NvU8 planeIndex;
 
         FOR_ALL_VALID_PLANES(planeIndex, pSurfaceEvo) {
@@ -1203,10 +1245,11 @@ NVSurfaceEvoPtr nvEvoGetSurfaceFromHandle(
                                 surfaceHandle,
                                 isUsedByCursorChannel,
                                 isUsedByLayerChannel,
-                                TRUE /* requireCtxDma */);
+                                TRUE /* requireDisplayHardwareAccess */,
+                                TRUE /* maybeUsedBy3d */);
 }
 
-NVSurfaceEvoPtr nvEvoGetSurfaceFromHandleNoCtxDmaOk(
+NVSurfaceEvoPtr nvEvoGetSurfaceFromHandleNoDispHWAccessOk(
     const NVDevEvoRec *pDevEvo,
     const NVEvoApiHandlesRec *pOpenDevSurfaceHandles,
     NvKmsSurfaceHandle surfaceHandle)
@@ -1216,7 +1259,22 @@ NVSurfaceEvoPtr nvEvoGetSurfaceFromHandleNoCtxDmaOk(
                                 surfaceHandle,
                                 FALSE /* isUsedByCursorChannel */,
                                 FALSE /* isUsedByLayerChannel */,
-                                FALSE /* requireCtxDma */);
+                                FALSE /* requireDisplayHardwareAccess */,
+                                TRUE /* maybeUsedBy3d */);
+}
+
+NVSurfaceEvoPtr nvEvoGetSurfaceFromHandleNoHWAccess(
+    const NVDevEvoRec *pDevEvo,
+    const NVEvoApiHandlesRec *pOpenDevSurfaceHandles,
+    NvKmsSurfaceHandle surfaceHandle)
+{
+    return GetSurfaceFromHandle(pDevEvo,
+                                pOpenDevSurfaceHandles,
+                                surfaceHandle,
+                                FALSE /* isUsedByCursorChannel */,
+                                FALSE /* isUsedByLayerChannel */,
+                                FALSE /* requireDisplayHardwareAccess */,
+                                FALSE /* maybeUsedBy3d */);
 }
 
 /*!
