@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -22,14 +22,23 @@
  */
 
 #include "gpu/gpu.h"
+#include "gpu/gpu_child_class_defs.h"
 #include "os/os.h"
 #include "nverror.h"
+#include "vgpu/rpc.h"
+#include "nvrm_registry.h"
 
 #include "published/hopper/gh100/hwproject.h"
 #include "published/hopper/gh100/dev_gc6_island.h"
+#include "published/hopper/gh100/dev_gc6_island_addendum.h"
 #include "published/hopper/gh100/dev_pmc.h"
 #include "published/hopper/gh100/dev_xtl_ep_pcfg_gpu.h"
 #include "published/hopper/gh100/pri_nv_xal_ep.h"
+#include "published/hopper/gh100/dev_vm.h"
+#include "published/hopper/gh100/dev_xtl_ep_pri.h"
+
+#include "published/hopper/gh100/dev_therm.h"
+#include "published/hopper/gh100/dev_therm_addendum.h"
 
 #include "ctrl/ctrl2080/ctrl2080mc.h"
 
@@ -52,7 +61,7 @@ gpuWriteBusConfigReg_GH100
     NvU32      value
 )
 {
-    return gpuWriteBusConfigCycle(pGpu, index, value);
+    return gpuWriteBusConfigCycle_HAL(pGpu, index, value);
 }
 
 /*!
@@ -72,7 +81,29 @@ gpuReadBusConfigReg_GH100
     NvU32     *pData
 )
 {
-    return gpuReadBusConfigCycle(pGpu, index, pData);
+    return gpuReadBusConfigCycle_HAL(pGpu, index, pData);
+}
+
+/*!
+ * @brief Read the non-private registers on vGPU through mirror space
+ *
+ * @param[in]  pGpu   GPU object pointer
+ * @param[in]  index  Register offset in PCIe config space
+ * @param[out] pData  Value of the register
+ *
+ * @returns    NV_OK on success
+ */
+NV_STATUS
+gpuReadPassThruConfigReg_GH100
+(
+    OBJGPU    *pGpu,
+    NvU32      index,
+    NvU32     *pData
+)
+{
+    *pData = GPU_REG_RD32(pGpu, DEVICE_BASE(NV_EP_PCFGM) + index);
+
+    return NV_OK;
 }
 
 /*!
@@ -148,6 +179,15 @@ NvU32 gpuGetPhysAddrWidth_GH100
  */
 NvBool gpuIsAtsSupportedWithSmcMemPartitioning_GH100(OBJGPU *pGpu)
 {
+    //
+    // ATS is supported only on self-hosted Hopper.
+    // Also enable it for simulation platforms for fmodel/RTL verification
+    // with CPU model.
+    //
+    if (gpuIsSelfHosted(pGpu))
+    {
+        return NV_TRUE;
+    }
 
     if (IS_SIMULATION(pGpu))
     {
@@ -157,47 +197,6 @@ NvBool gpuIsAtsSupportedWithSmcMemPartitioning_GH100(OBJGPU *pGpu)
     return NV_FALSE;
 }
 
-/*!
- * @brief Read back device ID and subsystem ID
- *
- * @note This function is designed to avoid OBJBIF dependence for the case that
- *       RM need to get device id and subsystem id early in the init process,
- *       before OBJBIF is initialized.
- *
- * @param[in]  pGpu   GPU object pointer
- * @param[out] pDevId Return value of device ID
- * @param[out] pSsId  Return value of subsystem ID
- */
-void
-gpuReadDeviceId_GH100
-(
-    OBJGPU *pGpu,
-    NvU32   *pDevId,
-    NvU32   *pSsId
-)
-{
-    NvU32   deviceId;
-    NvU32   ssId;
-
-    if (pDevId == NULL || pSsId == NULL) return;
-
-    if (GPU_BUS_CFG_CYCLE_RD32(pGpu, NV_EP_PCFG_GPU_ID, &deviceId) != NV_OK)
-    {
-        NV_PRINTF(LEVEL_ERROR,
-                  "Unable to read NV_EP_PCFG_GPU_ID\n");
-        return;
-    }
-    *pDevId = deviceId;
-
-    if (GPU_BUS_CFG_CYCLE_RD32(pGpu, NV_EP_PCFG_GPU_SUBSYSTEM_ID, &ssId) != NV_OK)
-    {
-        NV_PRINTF(LEVEL_ERROR,
-                  "Unable to read NV_EP_PCFG_GPU_SUBSYSTEM_ID\n");
-        return;
-    }
-    *pSsId = ssId;
-}
-
 /*!
  * @brief        Handle SEC_FAULT
  *
@@ -209,7 +208,8 @@ gpuHandleSecFault_GH100
     OBJGPU *pGpu
 )
 {
-    NvU32   secDebug = 0;
+    NvU32 secDebug = 0;
+    NvU32 iffPos;
 
     //
     // Read SEC_FAULT config space to determine what went wrong.
@@ -221,9 +221,7 @@ gpuHandleSecFault_GH100
     NV_PRINTF(LEVEL_FATAL, "SEC_FAULT lockdown detected. This is fatal. "
                             "RM will now shut down. NV_EP_PCFG_GPU_VSEC_DEBUG_SEC: 0x%x\n", secDebug);
 
-// (void)REF_NUM grants us mcheck protection
 #define LOG_SEC_FAULT(field) \
-    (void)REF_NUM(NV_EP_PCFG_GPU_VSEC_DEBUG_SEC##field, 0); \
     if (DRF_VAL(_EP_PCFG_GPU, _VSEC_DEBUG_SEC, field, secDebug) != 0) \
     { \
         MODS_ARCH_ERROR_PRINTF("NV_EP_PCFG_GPU_VSEC_DEBUG_SEC" #field "\n"); \
@@ -248,10 +246,19 @@ gpuHandleSecFault_GH100
     LOG_SEC_FAULT(_FAULT_PRI);
     LOG_SEC_FAULT(_FAULT_WDG);
     LOG_SEC_FAULT(_FAULT_BOOTFSM);
-    LOG_SEC_FAULT(_IFF_POS);
 
 #undef LOG_SEC_FAULT
 
+    // IFF_POS has a multi-bit value
+    iffPos = DRF_VAL(_EP_PCFG_GPU, _VSEC_DEBUG_SEC, _IFF_POS, secDebug);
+    if (iffPos != 0)
+    {
+        MODS_ARCH_ERROR_PRINTF("NV_EP_PCFG_GPU_VSEC_DEBUG_SEC_IFF_POS value: 0x%x\n", iffPos);
+        NV_PRINTF(LEVEL_FATAL, "SEC_FAULT type: _IFF_POS value: 0x%x\n", iffPos);
+        nvErrorLog_va((void *)(pGpu), SEC_FAULT_ERROR,
+                      "SEC_FAULT: _IFF_POS value: 0x%x", iffPos);
+    }
+
     //
     // After SEC_FAULT occurs, the GPU will only return SCPM dummy values until properly reset.
     // Only cold reset will clear SEC_FAULT, not hot reset. This GPU is as good as lost.
@@ -359,91 +366,205 @@ gpuGetFlaVasSize_GH100
 
 static const GPUCHILDPRESENT gpuChildrenPresent_GH100[] =
 {
-    { classId(OBJFUSE), 1, },
-    { classId(OBJSEQ), 1, },
-    { classId(GpuMutexMgr), 1, },
-    { classId(OBJTMR), 1, },
-    { classId(GraphicsManager), 1, },
-    { classId(MIGManager), 1, },
-    { classId(KernelMIGManager), 1, },
-    { classId(KernelGraphicsManager), 1, },
-    { classId(OBJVBIOS), 1, },
-    { classId(OBJDCB), 1, },
-    { classId(OBJGPIO), 1, },
-    { classId(I2c), 1, },
-    { classId(Spi), 1, },
-    { classId(KernelRc), 1, },
-    { classId(OBJRC), 1, },
-    { classId(OBJSTEREO), 1, },
-    { classId(Intr), 1, },
-    { classId(OBJINFOROM), 1, },
-    { classId(NvDebugDump), 1, },
-    { classId(SMDebugger), 1, },
-    { classId(OBJGPULOG), 1, },
-    { classId(OBJGPUMON), 1, },
-    { classId(OBJSWENG), 1 },
-    { classId(OBJUVM), 1 },
-    { classId(OBJACR), 1 },
-    { classId(OBJBIF), 1 },
-    { classId(KernelBif), 1 },
-    { classId(OBJBSP), 8 },
-    { classId(OBJBUS), 1 },
-    { classId(KernelBus), 1 },
-    { classId(OBJCE), 10 },
-    { classId(KernelCE), 10 },
-    { classId(OBJCIPHER), 1 },
-    { classId(ClockManager), 1 },
-    { classId(OBJDISP), 1 },
-    { classId(KernelDisplay), 1 },
-    { classId(VirtMemAllocator), 1 },
-    { classId(OBJDPAUX), 1 },
-    { classId(Fan), 1 },
-    { classId(OBJHSHUBMANAGER), 1 },
-    { classId(Hshub), 5 },
-    { classId(MemorySystem), 1 },
-    { classId(KernelMemorySystem), 1 },
-    { classId(MemoryManager), 1 },
-    { classId(OBJFBFLCN), 1 },
-    {classId(KernelFifo), 1},
-    { classId(OBJFIFO), 1 },
-    { classId(OBJGMMU), 1 },
-    { classId(KernelGmmu), 1},
-    { classId(Graphics), 8 },
-    { classId(KernelGraphics), 8 },
-    { classId(OBJHDACODEC), 1 },
-    { classId(OBJHWPM), 1 },
-    { classId(Lpwr   ), 1 },
-    { classId(OBJLSFM), 1 },
-    { classId(OBJMC), 1 },
-    { classId(KernelMc), 1 },
-    { classId(PrivRing), 1 },
-    { classId(SwIntr), 1 },
-    { classId(OBJNVJPG), 8 },
-    { classId(KernelNvlink), 1 },
-    { classId(Nvlink), 1 },
-    { classId(Perf), 1 },
-    { classId(KernelPerf), 1 },
-    { classId(Pmgr), 1 },
-    { classId(Pmu), 1 },
-    { classId(KernelPmu), 1 },
-    { classId(OBJSEC2), 1 },
-    { classId(Gsp), 1 },
-    { classId(OBJFSP), 1 },
-    { classId(KernelFsp), 1 },
-    { classId(Therm), 1 },
-    { classId(OBJVOLT), 1 },
-    { classId(OBJGRIDDISPLAYLESS), 1 },
-    { classId(OBJVMMU), 1 },
-    { classId(OBJOFA), 1 },
-    { classId(KernelGsp), 1},
-    { classId(KernelSec2), 1},
-    { classId(KernelCcu), 1 },
+    GPU_CHILD_PRESENT(OBJTMR, 1),
+    GPU_CHILD_PRESENT(KernelMIGManager, 1),
+    GPU_CHILD_PRESENT(KernelGraphicsManager, 1),
+    GPU_CHILD_PRESENT(KernelRc, 1),
+    GPU_CHILD_PRESENT(Intr, 1),
+    GPU_CHILD_PRESENT(NvDebugDump, 1),
+    GPU_CHILD_PRESENT(OBJSWENG, 1),
+    GPU_CHILD_PRESENT(OBJUVM, 1),
+    GPU_CHILD_PRESENT(KernelBif, 1),
+    GPU_CHILD_PRESENT(KernelBus, 1),
+    GPU_CHILD_PRESENT(KernelCE, 10),
+    GPU_CHILD_PRESENT(KernelDisplay, 1),
+    GPU_CHILD_PRESENT(VirtMemAllocator, 1),
+    GPU_CHILD_PRESENT(KernelMemorySystem, 1),
+    GPU_CHILD_PRESENT(MemoryManager, 1),
+    GPU_CHILD_PRESENT(KernelFifo, 1),
+    GPU_CHILD_PRESENT(KernelGmmu, 1),
+    GPU_CHILD_PRESENT(KernelGraphics, 8),
+    GPU_CHILD_PRESENT(KernelHwpm, 1),
+    GPU_CHILD_PRESENT(KernelMc, 1),
+    GPU_CHILD_PRESENT(SwIntr, 1),
+    GPU_CHILD_PRESENT(KernelNvlink, 1),
+    GPU_CHILD_PRESENT(KernelPerf, 1),
+    GPU_CHILD_PRESENT(KernelPmu, 1),
+    GPU_CHILD_PRESENT(KernelFsp, 1),
+    GPU_CHILD_PRESENT(KernelGsp, 1),
+    GPU_CHILD_PRESENT(KernelSec2, 1),
+    GPU_CHILD_PRESENT(ConfidentialCompute, 1),
+    GPU_CHILD_PRESENT(KernelCcu, 1),
 };
 
 const GPUCHILDPRESENT *
 gpuGetChildrenPresent_GH100(OBJGPU *pGpu, NvU32 *pNumEntries)
 {
-    *pNumEntries = NV_ARRAY_ELEMENTS32(gpuChildrenPresent_GH100);
+    *pNumEntries = NV_ARRAY_ELEMENTS(gpuChildrenPresent_GH100);
     return gpuChildrenPresent_GH100;
 }
 
+/*!
+ * @brief Determine if GPU is configured in Self Hosted mode.
+ * In Hopper+, GPU can be configured to work in Self Hosted,
+ * Externally Hosted and Endpoint mode.
+ *
+ * @param[in]      pGpu           OBJGPU pointer
+ *
+ * @return void
+ */
+void
+gpuDetermineSelfHostedMode_KERNEL_GH100
+(
+    OBJGPU *pGpu
+)
+{
+    if (IS_VIRTUAL(pGpu))
+    {
+        VGPU_STATIC_INFO *pVSI = GPU_GET_STATIC_INFO(pGpu);
+        if (pVSI->bSelfHostedMode)
+        {
+            pGpu->bIsSelfHosted = NV_TRUE;
+            NV_PRINTF(LEVEL_INFO, "SELF HOSTED mode detected after reading VGPU static info.\n");
+        }
+    }
+    if (IS_GSP_CLIENT(pGpu))
+    {
+        GspStaticConfigInfo *pGSCI = GPU_GET_GSP_STATIC_INFO(pGpu);
+        if (pGSCI->bSelfHostedMode)
+        {
+            pGpu->bIsSelfHosted = NV_TRUE;
+            NV_PRINTF(LEVEL_INFO, "SELF HOSTED mode detected after reading GSP static info.\n");
+        }
+    }
+}
+
+/*!
+ * @brief Determine SoC type if GPU is configured in Self Hosted mode.
+ *
+ * @param[in]      pGpu           OBJGPU pointer
+ *
+ * @return SoC type
+ */
+NvU32
+gpuDetermineSelfHostedSocType_GH100
+(
+    OBJGPU *pGpu
+)
+{
+    if (!gpuIsSelfHosted(pGpu))
+    {
+        return NV0000_CTRL_SYSTEM_SH_SOC_TYPE_NA;
+    }
+    return NV0000_CTRL_SYSTEM_SH_SOC_TYPE_NV_GRACE;
+}
+
+/*!
+ * Check if CC bit has been set in the scratch register
+ *
+ * @param[in]  pGpu  GPU object pointer
+ */
+NvBool
+gpuIsCCEnabledInHw_GH100
+(
+    OBJGPU *pGpu
+)
+{
+    NvU32 val = GPU_REG_RD32(pGpu, NV_PGC6_AON_SECURE_SCRATCH_GROUP_20_CC);
+    return FLD_TEST_DRF(_PGC6, _AON_SECURE_SCRATCH_GROUP_20_CC, _MODE_ENABLED, _TRUE, val);
+}
+
+/*!
+ * Check if dev mode bit has been set in the scratch register
+ *
+ * @param[in]  pGpu  GPU object pointer
+ */
+NvBool
+gpuIsDevModeEnabledInHw_GH100
+(
+    OBJGPU *pGpu
+)
+{
+    NvU32 val = GPU_REG_RD32(pGpu, NV_PGC6_AON_SECURE_SCRATCH_GROUP_20_CC);
+    return FLD_TEST_DRF(_PGC6, _AON_SECURE_SCRATCH_GROUP_20_CC, _DEV_ENABLED, _TRUE, val);
+}
+
+/*!
+ * Check if protected pcie has been set in the scratch register
+ *
+ * @param[in]  pGpu  GPU object pointer
+ */
+NvBool
+gpuIsProtectedPcieEnabledInHw_GH100
+(
+    OBJGPU *pGpu
+)
+{
+    NvU32 data;
+    NvU32 scratchVal = GPU_REG_RD32(pGpu, NV_PGC6_AON_SECURE_SCRATCH_GROUP_20_CC);
+    NvBool bIsScratchSet = FLD_TEST_DRF(_PGC6, _AON_SECURE_SCRATCH_GROUP_20_CC, _MULTI_GPU_MODE,
+                                        _PROTECTED_PCIE, scratchVal);
+
+    if (!bIsScratchSet)
+        return NV_FALSE;
+
+    if ((osReadRegistryDword(pGpu, NV_REG_STR_RM_PPCIE_ENABLED, &data) == NV_OK) &&
+        (data == NV_REG_STR_RM_PPCIE_ENABLED_YES))
+    {
+        return NV_TRUE;
+    }
+
+    return NV_FALSE;
+}
+
+/*!
+ * @brief Check if register being accessed is within guest BAR0 space.
+ *
+ * @param[in] pGpu   OBJGPU pointer
+ * @param[in] addr   Address being validated
+ */
+NV_STATUS
+gpuSanityCheckVirtRegAccess_GH100
+(
+    OBJGPU *pGpu,
+    NvU32   addr
+)
+{
+    // Not applicable in PV mode
+    if (IS_VIRTUAL_WITHOUT_SRIOV(pGpu))
+    {
+        return NV_OK;
+    }
+
+    if ((addr >= DEVICE_BASE(NV_EP_PCFGM)) &&
+        (addr < DEVICE_EXTENT(NV_EP_PCFGM)))
+    {
+        return NV_OK;
+    }
+
+    // Check if address in NV_VIRTUAL_FUNCTION range, if not error out.
+    if ((addr < DRF_EXTENT(NV_VIRTUAL_FUNCTION_PRIV)) ||
+        ((addr >= DRF_BASE(NV_VIRTUAL_FUNCTION)) && (addr < DRF_EXTENT(NV_VIRTUAL_FUNCTION))))
+    {
+        return NV_OK;
+    }
+
+    return NV_ERR_INVALID_ADDRESS;
+}
+
+/*!
+ * @brief Check if GSP-FMC Inst_in_sys ucode needs to be booted.
+ *
+ * @param[in]  pGpu          OBJGPU pointer
+ *
+ * @return NV_TRUE if GSP Inst_in_sys FMC needs to be booted, or NV_FALSE otherwise
+ */
+NvBool
+gpuIsGspToBootInInstInSysMode_GH100
+(
+    OBJGPU    *pGpu
+)
+{
+
+    return NV_FALSE;
+}
