@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -21,6 +21,7 @@
  * DEALINGS IN THE SOFTWARE.
  */
 
+#include "gpu_mgr/gpu_mgr.h"
 #include "gpu/gpu.h"
 #include "gpu/gpu_child_class_defs.h"
 #include "gpu_mgr/gpu_mgr_sli.h"
@@ -30,6 +31,7 @@
 #include "published/maxwell/gm107/dev_nv_xve.h"
 #include "published/maxwell/gm107/dev_nv_xve1.h"
 #include "published/maxwell/gm107/dev_fuse.h"
+#include "published/maxwell/gm107/dev_pri_ringstation_sys.h"
 
 /*!
  * @brief Returns SR-IOV capabilities
@@ -63,6 +65,95 @@ gpuFuseSupportsDisplay_GM107
     return GPU_FLD_TEST_DRF_DEF(pGpu, _FUSE, _STATUS_OPT_DISPLAY, _DATA, _ENABLE);
 }
 
+/*!
+ * @brief Read the pcie spec registers using config cycles
+ *
+ * @param[in]  pGpu      GPU object pointer
+ * @param[in]  hwDefAddr HW defined register address
+ * @param[out] pData     Value of the register
+ *
+ * @returns    NV_OK on success
+ */
+NV_STATUS
+gpuReadBusConfigCycle_GM107
+(
+    OBJGPU *pGpu,
+    NvU32   hwDefAddr,
+    NvU32  *pData
+)
+{
+    NvU32     domain              = gpuGetDomain(pGpu);
+    NvU8      bus                 = gpuGetBus(pGpu);
+    NvU8      device              = gpuGetDevice(pGpu);
+    NvU8      function            = 0;
+    NvBool    bIsCCFeatureEnabled = NV_FALSE;
+    NV_STATUS status              = NV_OK;
+
+    status = gpuConfigAccessSanityCheck_HAL(pGpu);
+    if (status != NV_OK)
+    {
+        return status;
+    }
+
+    // Todo: Reorganize this sanity check and handle for later chips
+    bIsCCFeatureEnabled = gpuIsCCFeatureEnabled(pGpu);
+
+    if (IS_PASSTHRU(pGpu) && !bIsCCFeatureEnabled)
+    {
+        gpuReadPassThruConfigReg_HAL(pGpu, hwDefAddr, pData);
+    }
+    else
+    {
+        if (pGpu->hPci == NULL)
+        {
+            pGpu->hPci = osPciInitHandle(domain, bus, device, function, NULL, NULL);
+        }
+
+        *pData = osPciReadDword(pGpu->hPci, hwDefAddr);
+    }
+
+    return NV_OK;
+}
+
+/*!
+ * @brief Write to pcie spec registers using config cycles
+ *
+ * @param[in] pGpu      GPU object pointer
+ * @param[in] hwDefAddr HW defined register address
+ * @param[in] value     Write this value to the register
+ *
+ * @returns    NV_OK on success
+ */
+NV_STATUS
+gpuWriteBusConfigCycle_GM107
+(
+    OBJGPU *pGpu,
+    NvU32   hwDefAddr,
+    NvU32   value
+)
+{
+    NvU32     domain   = gpuGetDomain(pGpu);
+    NvU8      bus      = gpuGetBus(pGpu);
+    NvU8      device   = gpuGetDevice(pGpu);
+    NvU8      function = 0;
+    NV_STATUS status   = NV_OK;
+
+    status = gpuConfigAccessSanityCheck_HAL(pGpu);
+    if (status != NV_OK)
+    {
+        return status;
+    }
+
+    if (pGpu->hPci == NULL)
+    {
+        pGpu->hPci = osPciInitHandle(domain, bus, device, function, NULL, NULL);
+    }
+
+    osPciWriteDword(pGpu->hPci, hwDefAddr, value);
+
+    return NV_OK;
+}
+
 /*!
  * @brief gpuReadBusConfigRegEx_GM107
  *
@@ -258,20 +349,6 @@ gpuWriteFunctionConfigRegEx_GM107
     return NV_OK;
 }
 
-void
-gpuReadDeviceId_GM107
-(
-    OBJGPU *pGpu,
-    NvU32  *devId,
-    NvU32  *ssId
-)
-{
-    if (devId == NULL || ssId == NULL) return;
-
-    *devId = GPU_REG_RD32(pGpu, DEVICE_BASE(NV_PCFG) + NV_XVE_ID);
-    *ssId  = GPU_REG_RD32(pGpu, DEVICE_BASE(NV_PCFG) + NV_XVE_SUBSYSTEM);
-}
-
 /*!
  * @brief Perform gpu-dependent error handling for error during register read sanity check
  *
@@ -377,6 +454,7 @@ gpuGetIdInfo_GM107(OBJGPU *pGpu)
 static const GPUCHILDORDER
 gpuChildOrderList_GM200[] =
 {
+    {classId(OBJVBIOS),           GCO_ALL},
     {classId(ConfidentialCompute),      GCO_ALL},
     {classId(Pxuc),               GCO_ALL},
     {classId(OBJBIF),             GCO_ALL},
@@ -439,6 +517,7 @@ gpuChildOrderList_GM200[] =
     {classId(KernelGraphics),     GCO_ALL},
     {classId(OBJLSFM),            GCO_LIST_INIT},
     {classId(OBJACR),             GCO_LIST_INIT},
+    {classId(OBJINFOROM),         GCO_ALL},
     {classId(Pmu),                GCO_LIST_INIT},
     {classId(KernelPmu),          GCO_LIST_INIT},
     {classId(Lpwr   ),            GCO_LIST_INIT},
@@ -459,15 +538,16 @@ gpuChildOrderList_GM200[] =
     {classId(KernelFifo),              GCO_ALL},
     {classId(OBJFIFO),            GCO_ALL},
     {classId(OBJDPAUX),           GCO_ALL},
-    {classId(OBJINFOROM),         GCO_ALL},
     {classId(OBJUVM),             GCO_ALL},
     {classId(OBJGPULOG),          GCO_LIST_INIT | GCO_LIST_LOAD},
     {classId(OBJGPUMON),          GCO_ALL},
     {classId(OBJGPULOG),          GCO_LIST_UNLOAD | GCO_LIST_DESTROY},
+    {classId(KernelHwpm),         GCO_ALL},
     {classId(OBJHWPM),            GCO_ALL},
     {classId(OBJSWENG),           GCO_ALL},
     {classId(OBJGRIDDISPLAYLESS), GCO_ALL},
-    {classId(KernelCcu),      GCO_ALL},
+    {classId(KernelCcu),          GCO_ALL},
+    {classId(KernelGsplite),      GCO_ALL},
 };
 
 const GPUCHILDORDER *
@@ -497,7 +577,6 @@ static const GPUCHILDPRESENT gpuChildrenPresent_GM200[] =
     GPU_CHILD_PRESENT(KernelRc, 1),
     GPU_CHILD_PRESENT(Intr, 1),
     GPU_CHILD_PRESENT(NvDebugDump, 1),
-    GPU_CHILD_PRESENT(OBJGPUMON, 1),
     GPU_CHILD_PRESENT(OBJSWENG, 1),
     GPU_CHILD_PRESENT(KernelBif, 1),
     GPU_CHILD_PRESENT(KernelBus, 1),
@@ -509,6 +588,7 @@ static const GPUCHILDPRESENT gpuChildrenPresent_GM200[] =
     GPU_CHILD_PRESENT(KernelFifo, 1),
     GPU_CHILD_PRESENT(KernelGmmu, 1),
     GPU_CHILD_PRESENT(KernelGraphics, 1),
+    GPU_CHILD_PRESENT(KernelHwpm, 1),
     GPU_CHILD_PRESENT(KernelMc, 1),
     GPU_CHILD_PRESENT(SwIntr, 1),
     GPU_CHILD_PRESENT(KernelPerf, 1),
@@ -634,3 +714,42 @@ gpuDetectSliLinkFromGpus_GK104
         }
     }
 }
+
+/*!
+ * @brief Get error that arises during sanity check on a register read value
+ *
+ * @param[in]       pGpu             GPU object pointer
+ * @param[in]       value            Register value
+ * @param[out]      pErrorString     Error string pointer
+ *
+ * @return void
+ */
+void
+gpuGetSanityCheckRegReadError_GM107
+(
+    OBJGPU *pGpu,
+    NvU32 value,
+    const char **pErrorString
+)
+{
+#define PRI_ERROR(err)                                    \
+if (DRF_VAL(_PPRIV, _SYS_PRI_ERROR, _CODE, value) == err) \
+{                                                         \
+    *pErrorString = #err;                                 \
+}
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_HOST_FECS_ERR);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_HOST_PRI_TIMEOUT);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_HOST_FB_ACK_TIMEOUT);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_TIMEOUT);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_DECODE);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_RESET);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_FLOORSWEEP);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_STUCK_ACK);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_0_EXPECTED_ACK);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_FENCE_ERROR);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_SUBID_ERROR);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_ORPHAN);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_DEAD_RING);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_TRAP);
+    PRI_ERROR(NV_PPRIV_SYS_PRI_ERROR_CODE_FECS_PRI_CLIENT_ERR);
+}
