@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -21,14 +21,19 @@
  * DEALINGS IN THE SOFTWARE.
  */
 
+#define NVOC_KERNEL_GRAPHICS_MANAGER_H_PRIVATE_ACCESS_ALLOWED
+
 #include "kernel/gpu/gr/kernel_graphics_manager.h"
 #include "kernel/gpu/gr/kernel_graphics.h"
+#include "kernel/gpu/gr/fecs_event_list.h"
 
+#include "kernel/gpu/device/device.h"
 #include "kernel/gpu/fifo/kernel_channel_group_api.h"
 #include "kernel/gpu/fifo/kernel_channel_group.h"
 
 #include "kernel/rmapi/client.h"
 #include "kernel/rmapi/client_resource.h"
+#include "nvRmReg.h"
 
 // COMPUTE
 #include "class/clb0c0.h"
@@ -41,6 +46,8 @@
 #include "class/clc7c0.h"
 #include "class/clc9c0.h"
 #include "class/clcbc0.h"
+#include "class/clcdc0.h"
+#include "class/clcec0.h"
 // GFX
 #include "class/clb097.h"
 #include "class/clb197.h"
@@ -52,11 +59,14 @@
 #include "class/clc797.h"
 #include "class/clc997.h"
 #include "class/clcb97.h"
+#include "class/clcd97.h"
+#include "class/clce97.h"
 // TWOD
 #include "class/cl902d.h"
 
 // MEM2MEM
 #include "class/cla140.h"
+#include "class/clcd40.h"
 
 static NvBool
 _kgrmgrGPUInstanceHasComputeInstances
@@ -105,6 +115,8 @@ kgrmgrGetGrObjectType_IMPL
         case AMPERE_COMPUTE_B:
         case ADA_COMPUTE_A:
         case HOPPER_COMPUTE_A:
+        case BLACKWELL_COMPUTE_A:
+        case BLACKWELL_COMPUTE_B:
             *pObjectType = GR_OBJECT_TYPE_COMPUTE;
             break;
         case MAXWELL_A:
@@ -117,12 +129,15 @@ kgrmgrGetGrObjectType_IMPL
         case AMPERE_B:
         case ADA_A:
         case HOPPER_A:
+        case BLACKWELL_A:
+        case BLACKWELL_B:
             *pObjectType = GR_OBJECT_TYPE_3D;
             break;
         case FERMI_TWOD_A:
             *pObjectType = GR_OBJECT_TYPE_2D;
             break;
         case KEPLER_INLINE_TO_MEMORY_B:
+        case BLACKWELL_INLINE_TO_MEMORY_A:
             *pObjectType = GR_OBJECT_TYPE_MEM;
             break;
         default:
@@ -167,6 +182,8 @@ kgrmgrIsCtxBufSupported_IMPL
         case GR_CTX_BUFFER_PAGEPOOL:
             // fall-through
         case GR_CTX_BUFFER_RTV_CB:
+            // fall-through
+        case GR_CTX_BUFFER_SETUP:
             bSupported = NV_FALSE;
             break;
 
@@ -288,6 +305,15 @@ kgrmgrCtrlSetChannelHandle_IMPL
     }
 }
 
+static void
+_kgrmgrInitRegistryOverrides
+(
+    OBJGPU *pGpu,
+    KernelGraphicsManager *pKernelGraphicsManager
+)
+{
+}
+
 NV_STATUS
 kgrmgrConstructEngine_IMPL
 (
@@ -296,6 +322,10 @@ kgrmgrConstructEngine_IMPL
     ENGDESCRIPTOR          engDesc
 )
 {
+    _kgrmgrInitRegistryOverrides(pGpu, pKernelGraphicsManager);
+
+    fecsGlobalLoggingInit(pGpu, pKernelGraphicsManager, &pKernelGraphicsManager->pFecsGlobalTraceInfo);
+
     return NV_OK;
 }
 
@@ -305,8 +335,12 @@ kgrmgrDestruct_IMPL
     KernelGraphicsManager *pKernelGraphicsManager
 )
 {
+    OBJGPU *pGpu = ENG_GET_GPU(pKernelGraphicsManager);
+
+    fecsGlobalLoggingTeardown(pGpu, pKernelGraphicsManager);
+
     portMemSet(&pKernelGraphicsManager->legacyKgraphicsStaticInfo.floorsweepingMasks, 0,
-               sizeof(pKernelGraphicsManager->legacyKgraphicsStaticInfo.floorsweepingMasks));
+           sizeof(pKernelGraphicsManager->legacyKgraphicsStaticInfo.floorsweepingMasks));
     portMemFree(pKernelGraphicsManager->legacyKgraphicsStaticInfo.pPpcMasks);
     pKernelGraphicsManager->legacyKgraphicsStaticInfo.pPpcMasks = NULL;
     portMemFree(pKernelGraphicsManager->legacyKgraphicsStaticInfo.pGrInfo);
@@ -314,6 +348,16 @@ kgrmgrDestruct_IMPL
     pKernelGraphicsManager->legacyKgraphicsStaticInfo.bInitialized = NV_FALSE;
 }
 
+void
+kgrmgrStateDestroy_IMPL
+(
+    OBJGPU *pGpu,
+    KernelGraphicsManager *pKernelGraphicsManager
+)
+{
+    fecsRemoveAllBindpointsForGpu(pGpu);
+}
+
 /*!
  * @brief Set legacy Kgraphics Static Info (i.e. state of GR0)
  */
@@ -362,20 +406,20 @@ cleanup:
 }
 
 /*!
- * @brief Retrieves associated KernelGraphics engine for given client / route info
+ * @brief Retrieves associated KernelGraphics engine for given device / route info
  *
  * @param[in]  pGpu
  * @param[in]  pKernelGraphicsManager
- * @param[in]  hClient                       Client handle
+ * @param[in]  pDevice
  * @param[in]  grRouteInfo                   Client-provided info to direct GR accesses
  * @param[out] ppKernelGraphics (Optional)   Ptr to store appropriate KernelGraphics *, if desired.
  */
 NV_STATUS
-kgrmgrCtrlRouteKGR_IMPL
+kgrmgrCtrlRouteKGRWithDevice_IMPL
 (
     OBJGPU *pGpu,
     KernelGraphicsManager *pKernelGraphicsManager,
-    NvHandle hClient,
+    Device *pDevice,
     const NV2080_CTRL_GR_ROUTE_INFO *pGrRouteInfo,
     KernelGraphics **ppKernelGraphics
 )
@@ -387,6 +431,7 @@ kgrmgrCtrlRouteKGR_IMPL
     NvU32 grIdx;
     NV2080_CTRL_GR_ROUTE_INFO grRouteInfo = *pGrRouteInfo;
     KernelMIGManager *pKernelMIGManager = GPU_GET_KERNEL_MIG_MANAGER(pGpu);
+    RsClient *pClient;
 
     if (!IS_MIG_IN_USE(pGpu))
     {
@@ -394,8 +439,10 @@ kgrmgrCtrlRouteKGR_IMPL
         goto done;
     }
 
+    pClient = RES_GET_CLIENT(pDevice);
+
     NV_CHECK_OK_OR_RETURN(LEVEL_ERROR,
-        kmigmgrGetInstanceRefFromClient(pGpu, pKernelMIGManager, hClient, &ref));
+        kmigmgrGetInstanceRefFromDevice(pGpu, pKernelMIGManager, pDevice, &ref));
 
     //
     // Compute instances always have 1 GR engine, so automatically fill in
@@ -408,8 +455,13 @@ kgrmgrCtrlRouteKGR_IMPL
     }
     else
     {
-        RS_PRIV_LEVEL privLevel = rmclientGetCachedPrivilegeByHandle(hClient);
-        if (!rmclientIsAdminByHandle(hClient, privLevel) &&
+        RmClient *pRmClient = dynamicCast(pClient, RmClient);
+
+        if (pRmClient == NULL)
+            return NV_ERR_INVALID_OBJECT_HANDLE;
+
+        RS_PRIV_LEVEL privLevel = rmclientGetCachedPrivilege(pRmClient);
+        if (!rmclientIsAdmin(pRmClient, privLevel) &&
              _kgrmgrGPUInstanceHasComputeInstances(pGpu, pKernelGraphicsManager, ref.pKernelMIGGpuInstance))
         {
             return NV_ERR_INSUFFICIENT_PERMISSIONS;
@@ -448,7 +500,7 @@ kgrmgrCtrlRouteKGR_IMPL
             NvHandle hChannel = DRF_VAL64(2080_CTRL_GR, _ROUTE_INFO_DATA,
                                           _CHANNEL_HANDLE, grRouteInfo.route);
 
-            status = CliGetKernelChannel(hClient, hChannel, &pKernelChannel);
+            status = CliGetKernelChannel(pClient, hChannel, &pKernelChannel);
             if (status != NV_OK)
             {
                 RsResourceRef         *pChanGrpRef;
@@ -459,12 +511,13 @@ kgrmgrCtrlRouteKGR_IMPL
                 // If retrieving a channel with the given hChannel doesn't work,
                 // try interpreting it as a handle to a channel group instead.
                 //
-                status = CliGetChannelGroup(hClient, hChannel, &pChanGrpRef, NULL);
+                status = CliGetChannelGroup(pClient->hClient, hChannel,
+                    &pChanGrpRef, NULL);
                 if (NV_OK != status)
                 {
                     NV_PRINTF(LEVEL_ERROR,
                               "Failed to find a channel or TSG with given handle 0x%08x associated with hClient=0x%08x\n",
-                              hChannel, hClient);
+                              hChannel, pClient->hClient);
                     return NV_ERR_INVALID_ARGUMENT;
                 }
 
@@ -532,6 +585,21 @@ kgrmgrGetLegacyGpcMask_IMPL
     return pKernelGraphicsManager->legacyKgraphicsStaticInfo.floorsweepingMasks.gpcMask;
 }
 
+/*!
+ * @return legacy physical GFX GPC mask enumerated by this chip
+ */
+NvU32
+kgrmgrGetLegacyPhysGfxGpcMask_IMPL
+(
+    OBJGPU *pGpu,
+    KernelGraphicsManager *pKernelGraphicsManager
+)
+{
+    NV_ASSERT_OR_RETURN(pKernelGraphicsManager->legacyKgraphicsStaticInfo.bInitialized, 0);
+
+    return pKernelGraphicsManager->legacyKgraphicsStaticInfo.floorsweepingMasks.physGfxGpcMask;
+}
+
 /*!
  * @return legacy TPC mask for certain GPC
  *
@@ -627,7 +695,7 @@ kgrmgrAllocVeidsForGrIdx_IMPL
     KERNEL_MIG_GPU_INSTANCE *pKernelMIGGPUInstance
 )
 {
-    NvU32 maxVeidsPerGpc;
+    NvU32 veidSizePerSpan;
     NvU32 veidStart = 0;
     NvU32 veidEnd = 0;
     NvU32 GPUInstanceVeidEnd;
@@ -639,20 +707,20 @@ kgrmgrAllocVeidsForGrIdx_IMPL
     NV_ASSERT_OR_RETURN(pKernelGraphicsManager->grIdxVeidMask[grIdx] == 0, NV_ERR_INVALID_STATE);
 
     NV_ASSERT_OK_OR_RETURN(
-        kgrmgrGetMaxVeidsPerGpc(pGpu, pKernelGraphicsManager, &maxVeidsPerGpc));
+        kgrmgrGetVeidSizePerSpan(pGpu, pKernelGraphicsManager, &veidSizePerSpan));
 
     // We statically assign VEIDs to a GR based on the number of GPCs connected to it
-    if (veidCount % maxVeidsPerGpc != 0)
+    if (veidCount % veidSizePerSpan != 0)
     {
-        NV_PRINTF(LEVEL_ERROR, "veidCount %d is not aligned to maxVeidsPerGpc=%d\n", veidCount, maxVeidsPerGpc);
+        NV_PRINTF(LEVEL_ERROR, "veidCount %d is not aligned to veidSizePerSpan=%d\n", veidCount, veidSizePerSpan);
         return NV_ERR_INVALID_ARGUMENT;
     }
 
     NV_ASSERT_OR_RETURN(veidSpanOffset != KMIGMGR_SPAN_OFFSET_INVALID, NV_ERR_INVALID_ARGUMENT);
 
-    veidStart = (veidSpanOffset * maxVeidsPerGpc) + pKernelMIGGPUInstance->resourceAllocation.veidOffset;
+    veidStart = (veidSpanOffset * veidSizePerSpan) + pKernelMIGGPUInstance->resourceAllocation.veidOffset;
     veidEnd = veidStart + veidCount - 1;
-    
+
     NV_ASSERT_OR_RETURN(veidStart < veidEnd, NV_ERR_INVALID_STATE);
     NV_ASSERT_OR_RETURN(veidStart < 64, NV_ERR_INVALID_ARGUMENT);
     NV_ASSERT_OR_RETURN(veidEnd < 64, NV_ERR_INVALID_ARGUMENT);
@@ -671,7 +739,14 @@ kgrmgrAllocVeidsForGrIdx_IMPL
 
     // VEID range should not overlap with existing VEIDs in use
     NV_ASSERT_OR_RETURN((pKernelGraphicsManager->veidInUseMask & reqVeidMask) == 0, NV_ERR_STATE_IN_USE);
-    NV_ASSERT_OR_RETURN((GPUInstanceFreeVeidMask & reqVeidMask) == reqVeidMask, NV_ERR_INSUFFICIENT_RESOURCES);
+
+    //
+    // For swizzId 0, client can use any range of VEIDs if not used, so we don't
+    // need to do the range check
+    //
+    NV_ASSERT_OR_RETURN(((pKernelMIGGPUInstance->swizzId == 0) ||
+                         ((GPUInstanceFreeVeidMask & reqVeidMask) == reqVeidMask)),
+                        NV_ERR_INSUFFICIENT_RESOURCES);
 
     // mark each VEID in the range as "in use"
     pKernelGraphicsManager->veidInUseMask |= reqVeidMask;
@@ -699,30 +774,42 @@ kgrmgrClearVeidsForGrIdx_IMPL
     pKernelGraphicsManager->grIdxVeidMask[grIdx] = 0;
 }
 
-/*!
- * @brief   Function to get max VEID count per GPC
+/*! 
+ * @brief Get VEID step size
  */
 NV_STATUS
-kgrmgrGetMaxVeidsPerGpc_IMPL
+kgrmgrGetVeidStepSize_IMPL
 (
     OBJGPU *pGpu,
     KernelGraphicsManager *pKernelGraphicsManager,
-    NvU32 *pMaxVeidsPerGpc
+    NvU32 *pVeidStepSize
 )
 {
-    NvU32 maxVeids;
-    NvU32 maxGpcCount;
-
-    NV_ASSERT_OR_RETURN(pMaxVeidsPerGpc != NULL, NV_ERR_INVALID_ARGUMENT);
+    NV_ASSERT_OR_RETURN(pVeidStepSize != NULL, NV_ERR_INVALID_ARGUMENT);
     NV_ASSERT_OR_RETURN(pKernelGraphicsManager->legacyKgraphicsStaticInfo.bInitialized, NV_ERR_INVALID_STATE);
     NV_ASSERT_OR_RETURN(pKernelGraphicsManager->legacyKgraphicsStaticInfo.pGrInfo != NULL, NV_ERR_INVALID_STATE);
 
-    maxVeids = pKernelGraphicsManager->legacyKgraphicsStaticInfo.pGrInfo->infoList[NV0080_CTRL_GR_INFO_INDEX_MAX_SUBCONTEXT_COUNT].data;
+    *pVeidStepSize = pKernelGraphicsManager->legacyKgraphicsStaticInfo.pGrInfo->infoList[NV0080_CTRL_GR_INFO_INDEX_LITTER_MIN_SUBCTX_PER_SMC_ENG].data;
 
-    maxGpcCount = gpuGetLitterValues_HAL(pGpu, NV2080_CTRL_GR_INFO_INDEX_LITTER_NUM_GPCS);
-    NV_ASSERT_OR_RETURN(maxGpcCount != 0, NV_ERR_INSUFFICIENT_RESOURCES);
+    return NV_OK;
+}
 
-    *pMaxVeidsPerGpc = (maxVeids / maxGpcCount);
+/*!
+ * @brief Get VEID size for each span
+ */
+NV_STATUS
+kgrmgrGetVeidSizePerSpan_IMPL
+(
+    OBJGPU *pGpu,
+    KernelGraphicsManager *pKernelGraphicsManager,
+    NvU32 *pVeidSizePerSpan
+)
+{
+    NV_ASSERT_OR_RETURN(pVeidSizePerSpan != NULL, NV_ERR_INVALID_ARGUMENT);
+
+    // VEIDs for each span should be the VEID size we assign to the smallest GPC count
+    NV_CHECK_OK_OR_RETURN(LEVEL_ERROR,
+        kgrmgrGetVeidsFromGpcCount_HAL(pGpu, pKernelGraphicsManager, 1, pVeidSizePerSpan));
 
     return NV_OK;
 }
@@ -968,7 +1055,7 @@ kgrmgrCheckVeidsRequest_IMPL
     KERNEL_MIG_GPU_INSTANCE *pKernelMIGGPUInstance
 )
 {
-    NvU32 maxVeidsPerGpc;
+    NvU32 veidSizePerSpan;
     NvU32 veidStart = 0;
     NvU32 veidEnd = 0;
     NvU32 GPUInstanceVeidEnd;
@@ -977,18 +1064,18 @@ kgrmgrCheckVeidsRequest_IMPL
     NvU64 veidMask;
 
     NV_ASSERT_OK_OR_RETURN(
-        kgrmgrGetMaxVeidsPerGpc(pGpu, pKernelGraphicsManager, &maxVeidsPerGpc));
+        kgrmgrGetVeidSizePerSpan(pGpu, pKernelGraphicsManager, &veidSizePerSpan));
 
     NV_ASSERT_OR_RETURN(pSpanStart != NULL, NV_ERR_INVALID_ARGUMENT);
     NV_ASSERT_OR_RETURN(pInUseMask != NULL, NV_ERR_INVALID_ARGUMENT);
 
     // We statically assign VEIDs to a GR based on the number of GPCs connected to it
-    if (veidCount % maxVeidsPerGpc != 0)
+    if (veidCount % veidSizePerSpan != 0)
     {
-        NV_PRINTF(LEVEL_ERROR, "veidCount %d is not aligned to maxVeidsPerGpc=%d\n", veidCount, maxVeidsPerGpc);
+        NV_PRINTF(LEVEL_ERROR, "veidCount %d is not aligned to veidSizePerSpan=%d\n", veidCount, veidSizePerSpan);
         return NV_ERR_INVALID_ARGUMENT;
     }
-    
+
     // Create a mask for VEIDs associated with this GPU instance
     GPUInstanceVeidEnd = pKernelMIGGPUInstance->resourceAllocation.veidOffset + pKernelMIGGPUInstance->resourceAllocation.veidCount - 1;
     GPUInstanceVeidMask = DRF_SHIFTMASK64(GPUInstanceVeidEnd:pKernelMIGGPUInstance->resourceAllocation.veidOffset);
@@ -999,7 +1086,7 @@ kgrmgrCheckVeidsRequest_IMPL
 
     if (*pSpanStart != KMIGMGR_SPAN_OFFSET_INVALID)
     {
-        veidStart = (*pSpanStart * maxVeidsPerGpc) + pKernelMIGGPUInstance->resourceAllocation.veidOffset;
+        veidStart = (*pSpanStart * veidSizePerSpan) + pKernelMIGGPUInstance->resourceAllocation.veidOffset;
         veidEnd = veidStart + veidCount - 1;
 
         NV_ASSERT_OR_RETURN(veidStart < veidEnd, NV_ERR_INVALID_STATE);
@@ -1007,11 +1094,11 @@ kgrmgrCheckVeidsRequest_IMPL
         NV_ASSERT_OR_RETURN(veidEnd < 64, NV_ERR_INVALID_ARGUMENT);
     }
     else
-    {        
+    {
         NvU64 reqVeidMask = DRF_SHIFTMASK64(veidCount - 1:0);
         NvU32 i;
 
-        for (i = pKernelMIGGPUInstance->resourceAllocation.veidOffset; i <= GPUInstanceVeidEnd; i += maxVeidsPerGpc)
+        for (i = pKernelMIGGPUInstance->resourceAllocation.veidOffset; i <= GPUInstanceVeidEnd; i += veidSizePerSpan)
         {
             // See if requested slots are available within this range
             if (((GPUInstanceFreeVeidMask >> i) & reqVeidMask) == reqVeidMask)
@@ -1029,14 +1116,22 @@ kgrmgrCheckVeidsRequest_IMPL
     veidMask = DRF_SHIFTMASK64(veidEnd:veidStart);
     NV_ASSERT_OR_RETURN(veidMask != 0x0, NV_ERR_INVALID_STATE);
 
-    NV_CHECK_OR_RETURN(LEVEL_ERROR, (GPUInstanceVeidMask & veidMask) == veidMask, NV_ERR_INSUFFICIENT_RESOURCES);
+    //
+    // For swizzId 0, client can use any range of VEIDs if not used, so we don't
+    // need to do the range check
+    //
+    NV_CHECK_OR_RETURN(LEVEL_ERROR, 
+                       ((pKernelMIGGPUInstance->swizzId == 0) ||
+                        (GPUInstanceVeidMask & veidMask) == veidMask),
+                       NV_ERR_INSUFFICIENT_RESOURCES);
 
     // VEID range should not overlap with existing VEIDs in use
     NV_ASSERT_OR_RETURN((*pInUseMask & veidMask) == 0, NV_ERR_STATE_IN_USE);
 
     NV_ASSERT(veidStart >= pKernelMIGGPUInstance->resourceAllocation.veidOffset);
 
-    *pSpanStart = (veidStart - pKernelMIGGPUInstance->resourceAllocation.veidOffset) / maxVeidsPerGpc;
+    *pSpanStart = (veidStart - pKernelMIGGPUInstance->resourceAllocation.veidOffset) / veidSizePerSpan;
     *pInUseMask |= veidMask;
     return NV_OK;
 }
+
