@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2015-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2015-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -47,7 +47,7 @@ static NV_STATUS _pmaNumaAvailableEvictableRange(PMA *pPma, NvS32 *validRegionLi
 static NV_STATUS _pmaNumaAllocateRange(PMA *pPma, NvU32 numaNodeId, NvLength actualSize,
     NvU64 pageSize, NvU64 *pPages, NvBool bScrubOnAlloc, NvBool allowEvict, NvS32 *validRegionList,
     NvU64 *allocatedCount);
-static NV_STATUS _pmaNumaAllocatePages (PMA *pPma, NvU32 numaNodeId, NvU32 pageSize,
+static NV_STATUS _pmaNumaAllocatePages (PMA *pPma, NvU32 numaNodeId, NvU64 pageSize,
     NvLength allocationCount, NvU64 *pPages, NvBool bScrubOnAlloc, NvBool allowEvict, NvS32 *validRegionList,
     NvU64 *allocatedPages);
 
@@ -163,10 +163,39 @@ NV_STATUS _pmaNumaAvailableEvictableRange
  */
 static NvBool _pmaCheckFreeFramesToSkipReclaim(PMA *pPma)
 {
-    return (100 * pPma->pmaStats.numFreeFrames < 
+    return (100 * pPma->pmaStats.numFreeFrames <
              (pPma->pmaStats.num2mbPages * (_PMA_2MB >> PMA_PAGE_SHIFT) * pPma->numaReclaimSkipThreshold));
 }
 
+/*!
+ * Translate a page returned by kernel to internal PMA page offset.
+ * @return NV_OK if the translation is successful.
+ *         NV_ERR_INVALID_STATE if the address is out of bound of PMA region
+ */
+static NV_STATUS
+_pmaTranslateKernelPage
+(
+    PMA   *pPma,
+    NvU64  sysPhysAddr,
+    NvU64  pageSize,
+    NvU64 *pGpaPhysAddr
+)
+{
+    NV_ASSERT_OR_RETURN(pGpaPhysAddr != NULL, NV_ERR_INVALID_ARGUMENT);
+
+    // Check returned page against online region
+    if ((sysPhysAddr < pPma->coherentCpuFbBase) ||
+        ((sysPhysAddr + pageSize) > (pPma->coherentCpuFbBase + pPma->coherentCpuFbSize)))
+    {
+        return NV_ERR_INVALID_STATE;
+    }
+
+    *pGpaPhysAddr = sysPhysAddr - pPma->coherentCpuFbBase;
+
+    // Check returned page against internal PMA structures
+    return pmaCheckRangeAgainstRegionDesc(pPma, *pGpaPhysAddr, pageSize);
+}
+
 /*!
  * @brief  Allocate contiguous memory for Numa
  *
@@ -189,6 +218,8 @@ NV_STATUS _pmaNumaAllocateRange
     NvU32 flags = OS_ALLOC_PAGES_NODE_NONE;
     *allocatedCount    = 0;
 
+    NV_ASSERT_OR_RETURN(actualSize >= osGetPageSize(), NV_ERR_INVALID_ARGUMENT);
+
     // check if numFreeFrames(64KB) are below a certain % of PMA managed memory(indicated by num2mbPages).
     if (_pmaCheckFreeFramesToSkipReclaim(pPma))
     {
@@ -202,15 +233,20 @@ NV_STATUS _pmaNumaAllocateRange
 
     if (status == NV_OK)
     {
-        NvU32 j;
-        // j=0 head page is already refcounted  at allocation
-        for (j = 1; j < (actualSize >> PMA_PAGE_SHIFT); j++)
+        NvU8 osPageShift = osGetPageShift();
+
+        // Skip the first page as it is refcounted at allocation.
+        osAllocAcquirePage(sysPhysAddr + (1 << osPageShift), (actualSize >> osPageShift) - 1);
+
+        // GPA needs to be acquired by shifting by the ATS aperture base address
+        status = _pmaTranslateKernelPage(pPma, sysPhysAddr, actualSize, &gpaPhysAddr);
+        if (status != NV_OK)
         {
-            osAllocAcquirePage(sysPhysAddr + (j << PMA_PAGE_SHIFT));
+            NV_PRINTF(LEVEL_ERROR, "Alloc from OS invalid for sysPhysAddr = 0x%llx actualSize = 0x%llx!\n",
+                                   sysPhysAddr, actualSize);
+            goto exit;
         }
 
-        gpaPhysAddr = sysPhysAddr - pPma->coherentCpuFbBase;
-        NV_ASSERT(gpaPhysAddr < pPma->coherentCpuFbBase);
         *allocatedCount = 1;
 
         if (bScrubOnAlloc)
@@ -246,6 +282,7 @@ scrub_exit:
         goto allocated;
     }
 
+exit:
     portSyncSpinlockAcquire(pPma->pPmaLock);
 
     NV_PRINTF(LEVEL_INFO, "Allocate from OS failed for allocation size = %lld!\n",
@@ -319,7 +356,7 @@ static NV_STATUS _pmaNumaAllocatePages
 (
     PMA     *pPma,
     NvU32    numaNodeId,
-    NvU32    pageSize,
+    NvU64    pageSize,
     NvLength allocationCount,
     NvU64   *pPages,
     NvBool   bScrubOnAlloc,
@@ -330,10 +367,12 @@ static NV_STATUS _pmaNumaAllocatePages
 {
     NV_STATUS status = NV_ERR_NO_MEMORY;
     NvU64     sysPhysAddr;
-    NvU64     i = 0, j = 0;
-    NvU32 flags = OS_ALLOC_PAGES_NODE_NONE;
+    NvU64     i = 0;
+    NvU32     flags = OS_ALLOC_PAGES_NODE_NONE;
+    NvU8      osPageShift = osGetPageShift();
 
     NV_ASSERT(allocationCount);
+    NV_ASSERT_OR_RETURN(pageSize >= osGetPageSize(), NV_ERR_INVALID_ARGUMENT);
 
     // check if numFreeFrames are below certain % of PMA managed memory.
     if (_pmaCheckFreeFramesToSkipReclaim(pPma))
@@ -348,23 +387,25 @@ static NV_STATUS _pmaNumaAllocatePages
         status = osAllocPagesNode((int)numaNodeId, (NvLength) pageSize, flags, &sysPhysAddr);
         if (status != NV_OK)
         {
-            NV_PRINTF(LEVEL_INFO, "Alloc from OS failed for i= %lld allocationCount = %lld pageSize = %lld!\n",
+            NV_PRINTF(LEVEL_ERROR, "Alloc from OS failed for i= %lld allocationCount = %lld pageSize = %lld!\n",
                                    i, (NvU64) allocationCount, (NvU64) pageSize);
             break;
         }
 
         // GPA needs to be acquired by shifting by the ATS aperture base address
-        NV_ASSERT(sysPhysAddr >= pPma->coherentCpuFbBase);
-        pPages[i] = sysPhysAddr - pPma->coherentCpuFbBase;
-
-        // Skip the head page at offset 0 (j=0) as it is refcounted at allocation
-        for (j = 1; j < (pageSize >> PMA_PAGE_SHIFT); j++)
+        status = _pmaTranslateKernelPage(pPma, sysPhysAddr, pageSize, &pPages[i]);
+        if (status != NV_OK)
         {
-            osAllocAcquirePage(sysPhysAddr + (j << PMA_PAGE_SHIFT));
+            NV_PRINTF(LEVEL_ERROR, "Alloc from OS invalid for i= %lld allocationCount = %lld pageSize = %lld!\n",
+                                   i, (NvU64) allocationCount, (NvU64) pageSize);
+            break;
         }
+
+        // Skip the first page as it is refcounted at allocation.
+        osAllocAcquirePage(sysPhysAddr + (1 << osPageShift), (pageSize >> osPageShift) - 1);
     }
 
-    if (bScrubOnAlloc)
+    if (bScrubOnAlloc && (i > 0))
     {
         PSCRUB_NODE pPmaScrubList = NULL;
         NvU64 count;
@@ -465,7 +506,7 @@ NV_STATUS pmaNumaAllocate
 (
     PMA                    *pPma,
     NvLength                allocationCount,
-    NvU32                   pageSize,
+    NvU64                   pageSize,
     PMA_ALLOCATION_OPTIONS *allocationOptions,
     NvU64                  *pPages
 )
@@ -493,9 +534,9 @@ NV_STATUS pmaNumaAllocate
         return NV_ERR_INVALID_ARGUMENT;
     }
 
-    if (pageSize > _PMA_2MB)
+    if (pageSize > _PMA_512MB)
     {
-        NV_PRINTF(LEVEL_FATAL, "Cannot allocate with more than 2MB contiguity.\n");
+        NV_PRINTF(LEVEL_FATAL, "Cannot allocate with more than 512MB contiguity.\n");
         return NV_ERR_INVALID_ARGUMENT;
     }
 
@@ -508,7 +549,8 @@ NV_STATUS pmaNumaAllocate
 
     if (contigFlag)
     {
-        if (((NvU64)allocationCount) * ((NvU64) pageSize) > NV_U32_MAX)
+        NvU64 contigTotal;
+        if (!portSafeMulU64(allocationCount, pageSize, &contigTotal) || contigTotal > NV_U32_MAX)
         {
             NV_PRINTF(LEVEL_FATAL, "Cannot allocate more than 4GB contiguous memory in one call.\n");
             return NV_ERR_INVALID_ARGUMENT;
@@ -523,7 +565,7 @@ NV_STATUS pmaNumaAllocate
     // caller does not want to skip scrubber.
     // Caller may want to skip scrubber when it knows the memory is zero'ed or when we are
     // initializing RM structures needed by the scrubber itself.
-    //  
+    //
     if (pPma->bScrubOnFree && !bSkipScrubFlag)
     {
         portSyncMutexAcquire(pPma->pAllocLock);
@@ -609,12 +651,14 @@ NV_STATUS pmaNumaAllocate
                     status = NV_ERR_NO_MEMORY;
                     break;
                 }
-                pPma->pMapInfo->pmaMapChangeStateAttrib(pMap, frameOffset, allocOption, NV_TRUE);
+                pPma->pMapInfo->pmaMapChangeStateAttrib(pMap, frameOffset, allocOption, MAP_MASK);
             }
             if (status != NV_OK)
                 break;
         }
 
+        pPma->pStatsUpdateCb(pPma->pStatsUpdateCtx, pPma->pmaStats.numFreeFrames);
+
         if (status == NV_OK)
         {
             allocationOptions->numPagesAllocated = (NvLength)finalAllocatedCount;
@@ -658,6 +702,9 @@ void pmaNumaFreeInternal
 )
 {
     NvU64 i, j;
+    NvU8 osPageShift = osGetPageShift();
+
+    NV_ASSERT_OR_RETURN_VOID(PMA_PAGE_SHIFT >= osPageShift);
 
     NV_PRINTF(LEVEL_INFO, "Freeing pPage[0] = %llx pageCount %lld\n", pPages[0], pageCount);
 
@@ -698,16 +745,18 @@ void pmaNumaFreeInternal
                 //
                 if (currentStatus & STATE_UNPIN)
                 {
-                    pPma->pMapInfo->pmaMapChangeStateAttribEx(pPma->pRegions[regId], (frameNum + j),
-                                                              ATTRIB_NUMA_REUSE, ATTRIB_NUMA_REUSE);
+                    pPma->pMapInfo->pmaMapChangeStateAttrib(pPma->pRegions[regId], (frameNum + j),
+                                                            ATTRIB_NUMA_REUSE, ATTRIB_NUMA_REUSE);
                 }
                 continue;
             }
             sysPagePhysAddr = sysPhysAddr + (j << PMA_PAGE_SHIFT);
-            osAllocReleasePage(sysPagePhysAddr);
-            pPma->pMapInfo->pmaMapChangeStateAttribEx(pPma->pRegions[regId], (frameNum + j), newStatus, ~ATTRIB_EVICTING);
+            osAllocReleasePage(sysPagePhysAddr, 1 << (PMA_PAGE_SHIFT - osPageShift));
+            pPma->pMapInfo->pmaMapChangeStateAttrib(pPma->pRegions[regId], (frameNum + j), newStatus, ~ATTRIB_EVICTING);
         }
     }
+
+    pPma->pStatsUpdateCb(pPma->pStatsUpdateCtx, pPma->pmaStats.numFreeFrames);
 }
 
 void pmaNumaSetReclaimSkipThreshold(PMA *pPma, NvU32 skipReclaimPercent)
