@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -21,6 +21,8 @@
  * DEALINGS IN THE SOFTWARE.
  */
 
+#define NVOC_KERNEL_MIG_MANAGER_H_PRIVATE_ACCESS_ALLOWED
+
 #include "kernel/gpu/mem_mgr/mem_mgr.h"
 #include "kernel/gpu/mem_mgr/heap.h"
 #include "kernel/gpu/mig_mgr/kernel_mig_manager.h"
@@ -112,16 +114,16 @@ kmigmgrCreateGPUInstanceCheck_GA100
     // Make sure there are no channels alive on any of these engines
     if (kfifoEngineListHasChannel(pGpu, pKernelFifo, engines, engineCount))
         return NV_ERR_STATE_IN_USE;
-    
+
     //
-    // Check for any alive P2P references to this GPU. P2P objects must 
-    // be re-created after disabling MIG. If it is allowed for  MIG to 
-    // continue enablement without all P2P objects torn down, there is 
+    // Check for any alive P2P references to this GPU. P2P objects must
+    // be re-created after disabling MIG. If it is allowed for  MIG to
+    // continue enablement without all P2P objects torn down, there is
     // the possibility that P2P mappings and state will never be updated.
     //
     if (bMemoryPartitioningNeeded || !kmigmgrIsMIGNvlinkP2PSupportOverridden(pGpu, pKernelMIGManager))
     {
-        NV_CHECK_OR_RETURN(LEVEL_ERROR, 
+        NV_CHECK_OR_RETURN(LEVEL_ERROR,
             !kbusIsGpuP2pAlive(pGpu, GPU_GET_KERNEL_BUS(pGpu)),
             NV_ERR_STATE_IN_USE);
     }
@@ -168,6 +170,9 @@ kmigmgrIsGPUInstanceFlagValid_GA100
     NvU32 computeSizeFlag = DRF_VAL(2080_CTRL_GPU, _PARTITION_FLAG,
                                     _COMPUTE_SIZE, gpuInstanceFlag);
 
+    NvU32 gfxSizeFlag = DRF_VAL(2080_CTRL_GPU, _PARTITION_FLAG,
+                                    _GFX_SIZE, gpuInstanceFlag);
+
     switch (memSizeFlag)
     {
         case NV2080_CTRL_GPU_PARTITION_FLAG_MEMORY_SIZE_FULL:
@@ -190,12 +195,33 @@ kmigmgrIsGPUInstanceFlagValid_GA100
         case NV2080_CTRL_GPU_PARTITION_FLAG_COMPUTE_SIZE_MINI_QUARTER:
         case NV2080_CTRL_GPU_PARTITION_FLAG_COMPUTE_SIZE_EIGHTH:
             break;
+        case NV2080_CTRL_GPU_PARTITION_FLAG_COMPUTE_SIZE_RESERVED_INTERNAL_06:
+        case NV2080_CTRL_GPU_PARTITION_FLAG_COMPUTE_SIZE_RESERVED_INTERNAL_07:
+            return NV_FALSE;
         default:
             NV_PRINTF(LEVEL_ERROR, "Unrecognized GPU compute partitioning flag 0x%x\n",
                       computeSizeFlag);
             return NV_FALSE;
     }
 
+    switch (gfxSizeFlag)
+    {
+        case NV2080_CTRL_GPU_PARTITION_FLAG_GFX_SIZE_NONE:
+            break;
+        case NV2080_CTRL_GPU_PARTITION_FLAG_GFX_SIZE_FULL:
+        case NV2080_CTRL_GPU_PARTITION_FLAG_GFX_SIZE_HALF:
+        case NV2080_CTRL_GPU_PARTITION_FLAG_GFX_SIZE_MINI_HALF:
+        case NV2080_CTRL_GPU_PARTITION_FLAG_GFX_SIZE_QUARTER:
+        case NV2080_CTRL_GPU_PARTITION_FLAG_GFX_SIZE_EIGHTH:
+        case NV2080_CTRL_GPU_PARTITION_FLAG_GFX_SIZE_RESERVED_INTERNAL_06:
+        case NV2080_CTRL_GPU_PARTITION_FLAG_GFX_SIZE_RESERVED_INTERNAL_07:
+            return NV_FALSE;
+        default:
+            NV_PRINTF(LEVEL_ERROR, "Unrecognized GPU GFX partitioning flag 0x%x\n",
+                      gfxSizeFlag);
+            return NV_FALSE;
+    }
+
     return NV_TRUE;
 }
 
@@ -365,3 +391,118 @@ kmigmgrIsMemoryPartitioningNeeded_GA100
     // Memory partitioning is needed for non-zero swizzIds
     return (swizzId != 0);
 }
+
+/*!
+ * @brief   Returns the span covered by the swizzId
+ */
+NV_RANGE
+kmigmgrSwizzIdToSpan_GA100
+(
+    OBJGPU *pGpu,
+    KernelMIGManager *pKernelMIGManager,
+    NvU32 swizzId
+)
+{
+    KernelGraphicsManager *pKernelGraphicsManager = GPU_GET_KERNEL_GRAPHICS_MANAGER(pGpu);
+    NV_RANGE ret;
+    NvU8 spanLen;
+    NvU32 maxValidSwizzId;
+
+    NV_ASSERT_OR_RETURN(kgrmgrGetLegacyKGraphicsStaticInfo(pGpu, pKernelGraphicsManager)->bInitialized, NV_RANGE_EMPTY);
+    NV_ASSERT_OR_RETURN(kgrmgrGetLegacyKGraphicsStaticInfo(pGpu, pKernelGraphicsManager)->pGrInfo != NULL, NV_RANGE_EMPTY);
+
+    if (kmigmgrIsA100ReducedConfig(pGpu, pKernelMIGManager))
+        spanLen = 4;
+    else
+        spanLen = kgrmgrGetLegacyKGraphicsStaticInfo(pGpu, pKernelGraphicsManager)->pGrInfo->infoList[NV2080_CTRL_GR_INFO_INDEX_MAX_PARTITIONABLE_GPCS].data;
+
+    switch (swizzId)
+    {
+        case 0:
+            ret = rangeMake(0, spanLen - 1);
+            break;
+        case 1:
+            ret = rangeMake(0, (spanLen/2) - 1);
+            break;
+        case 2:
+            ret = rangeMake(spanLen/2, spanLen - 1);
+            break;
+        case 3:
+            ret = rangeMake(0, (spanLen/4) - 1);
+            break;
+        case 4:
+            ret = rangeMake((spanLen/4), (spanLen/2) - 1);
+            break;
+        case 5:
+            ret = rangeMake((spanLen/2), (3*(spanLen/4)) - 1);
+            break;
+        case 6:
+            ret = rangeMake((3*(spanLen/4)), spanLen - 1);
+            break;
+        case 7:
+            ret = rangeMake(0, 0);
+            break;
+        case 8:
+            ret = rangeMake(1, 1);
+            break;
+        case 9:
+            ret = rangeMake(2, 2);
+            break;
+        case 10:
+            ret = rangeMake(3, 3);
+            break;
+        case 11:
+            ret = rangeMake(4, 4);
+            break;
+        case 12:
+            ret = rangeMake(5, 5);
+            break;
+        case 13:
+            ret = rangeMake(6, 6);
+            break;
+        case 14:
+            ret = rangeMake(7, 7);
+            break;
+        default:
+            NV_PRINTF(LEVEL_ERROR, "Unsupported swizzid 0x%x\n", swizzId);
+            DBG_BREAKPOINT();
+            ret = NV_RANGE_EMPTY;
+            break;
+    }
+
+    maxValidSwizzId = (spanLen * 2) - 2;
+
+    if (swizzId > maxValidSwizzId)
+    {
+        ret = NV_RANGE_EMPTY;
+    }
+
+    return ret;
+}
+
+/*!
+ * @brief Check if we are running on a reduced config GPU then set the corresponding flag
+ */
+void
+kmigmgrDetectReducedConfig_GA100
+(
+    OBJGPU *pGpu,
+    KernelMIGManager *pKernelMIGManager
+)
+{
+    const KERNEL_MIG_MANAGER_STATIC_INFO *pStaticInfo = kmigmgrGetStaticInfo(pGpu, pKernelMIGManager);
+    NvU32 i;
+
+    NV_ASSERT_OR_RETURN_VOID(pStaticInfo != NULL);
+
+    for (i = 0; i < pStaticInfo->pCIProfiles->profileCount; ++i)
+    {
+        // Reduced config A100 does not support 1/8 compute size
+        if (pStaticInfo->pCIProfiles->profiles[i].computeSize == NV2080_CTRL_GPU_PARTITION_FLAG_COMPUTE_SIZE_EIGHTH)
+        {
+            return;
+        }
+    }
+
+    pKernelMIGManager->bIsA100ReducedConfig = NV_TRUE;
+}
