@@ -26,12 +26,15 @@
  * @brief TURING specific HAL MMU routines reside in this file
  */
 
+#include "nvtypes.h"
 #define NVOC_KERN_GMMU_H_PRIVATE_ACCESS_ALLOWED
 
 #include "gpu/mmu/kern_gmmu.h"
 #include "gpu/gpu.h"
 #include "kernel/gpu/intr/intr.h"
+#include "gpu/mem_sys/kern_mem_sys.h"
 
+#include "published/turing/tu102/dev_fault.h"
 #include "published/turing/tu102/dev_fb.h"
 #include "published/turing/tu102/dev_vm.h"
 
@@ -42,26 +45,20 @@
  * @param pGpu
  * @param pKernelGmmu
  * @param pTimeOut      Timeout for the invalidate operation.
- * @param gfid          GFID
  */
 NV_STATUS
 kgmmuCheckPendingInvalidates_TU102
 (
     OBJGPU     *pGpu,
     KernelGmmu *pKernelGmmu,
-    RMTIMEOUT  *pTimeOut,
-    NvU32       gfid
+    RMTIMEOUT  *pTimeOut
 )
 {
     NV_STATUS status = NV_OK;
 
     while (1)
     {
-        NvU32   regVal;
-
-        {
-            regVal = GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE);
-        }
+        NvU32 regVal = GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE);
 
         if (FLD_TEST_DRF(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE, _TRIGGER, _FALSE, regVal))
             break;
@@ -98,23 +95,23 @@ kgmmuCommitTlbInvalidate_TU102
 {
     NV_STATUS status = NV_OK;
 
+    NV_ASSERT_OR_RETURN(IS_GFID_PF(pParams->gfid), NV_ERR_INVALID_ARGUMENT);
+
+    if (IS_VIRTUAL(pGpu))
+    {
+        // Prevent PF from updating INVALIDATE_ALL_PDB from VF requests, bug 3356599
+        pParams->regVal = FLD_SET_DRF(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE, _ALL_PDB, _FALSE, pParams->regVal);
+    }
+
     if (!FLD_TEST_DRF(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE, _ALL_PDB, _TRUE, pParams->regVal))
     {
         kgmmuSetPdbToInvalidate_HAL(pGpu, pKernelGmmu, pParams);
     }
 
-    {
-        if (IS_VIRTUAL(pGpu))
-        {
-            // Prevent VF from updating INVALIDATE_ALL_PDB, bug 3356599
-            pParams->regVal = FLD_SET_DRF(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE, _ALL_PDB, _FALSE, pParams->regVal);
-        }
-
-        GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE, pParams->regVal);
-    }
+    GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE, pParams->regVal);
 
     // Wait for the invalidate command to complete.
-    status = kgmmuCheckPendingInvalidates_HAL(pGpu, pKernelGmmu, &pParams->timeout, pParams->gfid);
+    status = kgmmuCheckPendingInvalidates_HAL(pGpu, pKernelGmmu, &pParams->timeout);
 
     return status;
 }
@@ -135,53 +132,19 @@ kgmmuSetPdbToInvalidate_TU102
     TLB_INVALIDATE_PARAMS *pParams
 )
 {
-    {
-        GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_PDB,
-                DRF_NUM(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE_PDB, _ADDR,
-                        NvU64_LO32(pParams->pdbAddress >>
-                                NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_PDB_ADDR_ALIGNMENT)) |
-                DRF_NUM(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE_PDB, _APERTURE, pParams->pdbAperture));
+    NV_ASSERT_OR_RETURN_VOID(IS_GFID_PF(pParams->gfid));
 
-        GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_UPPER_PDB,
-                DRF_NUM(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE_UPPER_PDB, _ADDR,
-                        NvU64_LO32((pParams->pdbAddress >>
-                                    NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_PDB_ADDR_ALIGNMENT) >>
-                                    DRF_SIZE(NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_PDB_ADDR))));
-    }
-}
+    GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_PDB,
+            DRF_NUM(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE_PDB, _ADDR,
+                    NvU64_LO32(pParams->pdbAddress >>
+                            NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_PDB_ADDR_ALIGNMENT)) |
+            DRF_NUM(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE_PDB, _APERTURE, pParams->pdbAperture));
 
-/**
- * @brief     Initialize the GMMU format families.
- * @details   Turing supports PA based comptagline allocation policy
- *
- * @param[in] pGpu
- * @param[in] pKernelGmmu
- */
-NV_STATUS
-kgmmuFmtFamiliesInit_TU102(OBJGPU *pGpu, KernelGmmu *pKernelGmmu)
-{
-    extern NV_STATUS kgmmuFmtFamiliesInit_GV100(OBJGPU *pGpu, KernelGmmu *pKernelGmmu);
-
-    NvU32            v;
-    NV_STATUS        result;
-    GMMU_FMT_FAMILY *pFam;
-
-    for (v = 0; v < GMMU_FMT_MAX_VERSION_COUNT; ++v)
-    {
-        pFam = pKernelGmmu->pFmtFamilies[v];
-        if (NULL != pFam)
-        {
-            if (kgmmuIsBug2720120WarEnabled(pKernelGmmu))
-            {
-                NV_ASSERT_OK_OR_RETURN(kgmmuSetupWarForBug2720120_HAL(pKernelGmmu, pFam));
-            }
-        }
-    }
-
-    // inherit former FmtFamilies setup procedure
-    result = kgmmuFmtFamiliesInit_GV100(pGpu, pKernelGmmu);
-
-    return result;
+    GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_UPPER_PDB,
+            DRF_NUM(_VIRTUAL_FUNCTION_PRIV, _MMU_INVALIDATE_UPPER_PDB, _ADDR,
+                    NvU64_LO32((pParams->pdbAddress >>
+                                NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_PDB_ADDR_ALIGNMENT) >>
+                                DRF_SIZE(NV_VIRTUAL_FUNCTION_PRIV_MMU_INVALIDATE_PDB_ADDR))));
 }
 
 /*!
@@ -321,6 +284,8 @@ kgmmuReadMmuFaultStatus_TU102
 )
 {
     {
+        NV_ASSERT(IS_GFID_PF(gfid));
+
         return GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_STATUS);
     }
 }
@@ -346,6 +311,8 @@ kgmmuReadMmuFaultBufferSize_TU102
 )
 {
     {
+        NV_ASSERT(IS_GFID_PF(gfid));
+
         return GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_BUFFER_SIZE(index));
     }
 }
@@ -369,6 +336,52 @@ kgmmuReadFaultBufferGetPtr_TU102
     return NV_OK;
 }
 
+NV_STATUS
+kgmmuWriteFaultBufferGetPtr_TU102
+(
+    OBJGPU               *pGpu,
+    KernelGmmu           *pKernelGmmu,
+    NvU32                 index,
+    NvU32                 getValue,
+    THREAD_STATE_NODE    *pThreadState
+)
+{
+    NV_ASSERT_OR_RETURN((index < NUM_FAULT_BUFFERS), NV_ERR_INVALID_ARGUMENT);
+
+    if (index == REPLAYABLE_FAULT_BUFFER)
+    {
+        //
+        // If HW has detected an overflow condition (PUT == GET - 1 and a fault has
+        // arrived, which is dropped due to no more space in the fault buffer), it will
+        // not deliver any more faults into the buffer until the overflow condition has
+        // been cleared. The overflow condition is cleared by updating the GET index to
+        // indicate space in the buffer and writing 1 to the OVERFLOW bit in GET.
+        // Unfortunately, this can not be done in the same write because it can collide
+        // with an arriving fault on the same cycle, resulting in the overflow condition
+        // being instantly reasserted.
+        // However, if the index is updated first and then the OVERFLOW bit is cleared
+        // such a collision will not cause a reassertion of the overflow condition.
+        //
+        GPU_VREG_WR32_EX(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_BUFFER_GET(index), getValue, pThreadState);
+        getValue = FLD_SET_DRF(_VIRTUAL_FUNCTION_PRIV, _MMU_FAULT_BUFFER_GET, _OVERFLOW, _CLEAR, getValue);
+        getValue = FLD_SET_DRF(_VIRTUAL_FUNCTION_PRIV, _MMU_FAULT_BUFFER_GET, _GETPTR_CORRUPTED, _CLEAR, getValue);
+    }
+    else
+    {
+        //
+        // Mask off the _OVERFLOW and _GETPTR_CORRUPTED bits. Writing 0 should
+        // have no effect, but they're write 1 to clear so make sure we're not
+        // trying to clear an overflow or the corrupted bit.
+        //
+        getValue = FLD_SET_DRF(_VIRTUAL_FUNCTION_PRIV, _MMU_FAULT_BUFFER_GET, _OVERFLOW, _NO, getValue);
+        getValue = FLD_SET_DRF(_VIRTUAL_FUNCTION_PRIV, _MMU_FAULT_BUFFER_GET, _GETPTR_CORRUPTED, _NO, getValue);
+    }
+
+    GPU_VREG_WR32_EX(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_BUFFER_GET(index), getValue, pThreadState);
+
+    return NV_OK;
+}
+
 NV_STATUS
 kgmmuReadFaultBufferPutPtr_TU102
 (
@@ -448,6 +461,187 @@ kgmmuClearReplayableFaultIntr_TU102
     intrClearLeafVector_HAL(pGpu, pIntr, NV_PFB_PRI_MMU_INT_VECTOR_FAULT_NOTIFY_REPLAYABLE, pThreadState);
 }
 
+/**
+ * @brief Print information about a MMU fault
+ *
+ * @param[in]  pGpu                  OBJGPU pointer
+ * @param[in]  pKernelGmmu           KernelGmmu pointer
+ * @param[in]  mmuFaultEngineId      Engine ID of the faulted engine
+ * @param[in]  pMmuExceptionData     FIFO exception packet used
+ *                                   for printing fault info
+ * @returns
+ */
+void
+kgmmuPrintFaultInfo_TU102
+(
+    OBJGPU                  *pGpu,
+    KernelGmmu              *pKernelGmmu,
+    NvU32                    mmuFaultEngineId,
+    FIFO_MMU_EXCEPTION_DATA *pMmuExceptionData
+)
+{
+    if (kmemsysCbcIsSafe_HAL(pGpu, GPU_GET_KERNEL_MEMORY_SYSTEM(pGpu)) == NV_FALSE)
+    {
+        NV_PRINTF(LEVEL_ERROR,
+                  "MMU Fault: CBC Backingstore unsafe, this can be reported as "
+                  "UNSUPPORTED_KIND error\n");
+    }
+
+    void kgmmuPrintFaultInfo_GV100(OBJGPU *pGpu, KernelGmmu *pKernelGmmu, NvU32, FIFO_MMU_EXCEPTION_DATA*);
+    kgmmuPrintFaultInfo_GV100(pGpu, pKernelGmmu, mmuFaultEngineId, pMmuExceptionData);
+}
+
+/*!
+ * @brief Check if the given engineID is BAR1
+ *
+ * @param[in] pKernelGmmu  KernelGmmu object
+ * @param[in] engineID     Engine ID
+ *
+ * @return True if BAR1
+ */
+NvBool
+kgmmuIsFaultEngineBar1_TU102
+(
+    KernelGmmu *pKernelGmmu,
+    NvU32       engineID
+)
+{
+    return (engineID == NV_PFAULT_MMU_ENG_ID_BAR1);
+}
+
+/*!
+ * @brief Check if the given engineID is BAR2
+ *
+ * @param[in] pKernelGmmu  KernelGmmu object
+ * @param[in] engineID     Engine ID
+ *
+ * @return True if BAR2
+ */
+
+NvBool
+kgmmuIsFaultEngineBar2_TU102
+(
+    KernelGmmu *pKernelGmmu,
+    NvU32       engineID
+)
+{
+    return (engineID == NV_PFAULT_MMU_ENG_ID_BAR2);
+}
+
+/*!
+ * @brief Returns the HW default size of MMU fault buffers (replayable
+ *        or non replayable buffers).
+ *
+ * @param[in] pGpu         OBJGPU  pointer
+ * @param[in] pKernelGmmu  KernelGmmu pointer
+ * @param[in] index        Fault buffer index
+ *
+ * @returns size of MMU fault buffers
+ */
+NvU32
+kgmmuSetAndGetDefaultFaultBufferSize_TU102
+(
+    OBJGPU           *pGpu,
+    KernelGmmu       *pKernelGmmu,
+    FAULT_BUFFER_TYPE index,
+    NvU32             gfid
+)
+{
+    NvU32 faultBufferSize;
+    {
+        {
+            NV_ASSERT(IS_GFID_PF(gfid));
+
+            GPU_VREG_FLD_IDX_WR_DRF_DEF(pGpu, _VIRTUAL_FUNCTION_PRIV, _MMU_FAULT_BUFFER_SIZE, index, _SET_DEFAULT, _YES);
+            faultBufferSize = GPU_VREG_IDX_RD_DRF(pGpu, _VIRTUAL_FUNCTION_PRIV, _MMU_FAULT_BUFFER_SIZE, index, _VAL) * NVC369_BUF_SIZE;
+        }
+    }
+
+    return faultBufferSize;
+}
+
+void
+kgmmuWriteMmuFaultBufferSize_TU102
+(
+    OBJGPU     *pGpu,
+    KernelGmmu *pKernelGmmu,
+    NvU32       index,
+    NvU32       value,
+    NvU32       gfid
+)
+{
+    {
+        NV_ASSERT(IS_GFID_PF(gfid));
+
+        GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_BUFFER_SIZE(index), value);
+    }
+
+    // Align the HW/SW _GET_PTR if _FAULT_BUFFER_SIZE_ENABLE is _FALSE which will be 0
+    if (FLD_TEST_DRF(_PFB_PRI, _MMU_FAULT_BUFFER_SIZE, _ENABLE, _FALSE, value))
+    {
+        pKernelGmmu->mmuFaultBuffer[gfid].hwFaultBuffers[index].cachedGetIndex = 0;
+    }
+}
+
+void
+kgmmuWriteMmuFaultBufferHiLo_TU102
+(
+    OBJGPU     *pGpu,
+    KernelGmmu *pKernelGmmu,
+    NvU32       index,
+    NvU32       loValue,
+    NvU32       hiValue,
+    NvU32       gfid
+)
+{
+    {
+        NV_ASSERT(IS_GFID_PF(gfid));
+
+        GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_BUFFER_HI(index), hiValue);
+        GPU_VREG_WR32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_BUFFER_LO(index), loValue);
+    }
+}
+
+void
+kgmmuReadMmuFaultInstHiLo_TU102
+(
+    OBJGPU     *pGpu,
+    KernelGmmu *pKernelGmmu,
+    NvU32      *pHiVal,
+    NvU32      *pLoVal
+)
+{
+    NV_ASSERT(pHiVal != NULL);
+    NV_ASSERT(pLoVal != NULL);
+    *pLoVal = GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_INST_LO);
+    *pHiVal = GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_INST_HI);
+}
+
+void
+kgmmuReadMmuFaultAddrHiLo_TU102
+(
+    OBJGPU     *pGpu,
+    KernelGmmu *pKernelGmmu,
+    NvU32      *pHiVal,
+    NvU32      *pLoVal
+)
+{
+    NV_ASSERT(pHiVal != NULL);
+    NV_ASSERT(pLoVal != NULL);
+    *pLoVal = GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_ADDR_LO);
+    *pHiVal = GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_ADDR_HI);
+}
+
+NvU32
+kgmmuReadMmuFaultInfo_TU102
+(
+    OBJGPU     *pGpu,
+    KernelGmmu *pKernelGmmu
+)
+{
+    return GPU_VREG_RD32(pGpu, NV_VIRTUAL_FUNCTION_PRIV_MMU_FAULT_INFO);
+}
+
 NvU32
 kgmmuGetEccCounts_TU102
 (
