@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -23,12 +23,14 @@
 
 #include "kernel/core/core.h"
 #include "kernel/core/locks.h"
+#include "gpu/subdevice/subdevice.h"
 #include "kernel/gpu/mem_mgr/heap.h"
 #include "kernel/gpu/mem_mgr/mem_mgr.h"
 #include "kernel/gpu/mig_mgr/kernel_mig_manager.h"
 #include "kernel/gpu/rc/kernel_rc.h"
 #include "kernel/gpu/bif/kernel_bif.h"
 #include "kernel/os/os.h"
+#include "platform/sli/sli.h"
 
 #include "class/cl0000.h" // NV01_NULL_OBJECT
 #include "class/cl0002.h" // NV01_CONTEXT_DMA
@@ -49,8 +51,10 @@
 #include "class/clc56f.h" // AMPERE_CHANNEL_GPFIFO_A
 #include "class/clc86f.h" // HOPPER_CHANNEL_GPFIFO_A
 
+#include "class/clc96f.h" // BLACKWELL_CHANNEL_GPFIFO_A
+
 #include "deprecated/rmapi_deprecated.h"
-#include "nvRmReg.h"
+#include "nvrm_registry.h"
 
 //
 // Watchdog object ids
@@ -389,9 +393,7 @@ krcWatchdogShutdown_IMPL
         return NV_OK;
 
     krcWatchdogDisable(pKernelRc);
-    osRemove1SecondRepeatingCallback(pGpu,
-                                     krcWatchdogTimerProc,
-                                     NULL /* pData */);
+    osRemove1HzCallback(pGpu, krcWatchdogTimerProc, NULL /* pData */);
 
     // This should free the client and all associated resources
     pRmApi->Free(pRmApi,
@@ -477,11 +479,23 @@ krcWatchdogInit_IMPL
 
     if (bClientUserd)
     {
-        Heap *pHeap = GPU_GET_HEAP(pGpu);
-        if (pHeap->pmaObject.bNuma)
+        MemoryManager *pMemoryManager = GPU_GET_MEMORY_MANAGER(pGpu);
+
+        if (memmgrIsPmaInitialized(pMemoryManager))
         {
-            // PMA can't be used until it's onlined
-            bClientUserd = NV_FALSE;
+            Heap *pHeap = GPU_GET_HEAP(pGpu);
+            NvU32 pmaConfig = PMA_QUERY_NUMA_ENABLED | PMA_QUERY_NUMA_ONLINED;
+
+            if (pmaQueryConfigs(&pHeap->pmaObject, &pmaConfig) == NV_OK)
+            {
+                // PMA can't be used until it's onlined
+                if (pmaConfig & PMA_QUERY_NUMA_ENABLED)
+                {
+                    // PMA can't be used until it's onlined
+                    if (!(pmaConfig & PMA_QUERY_NUMA_ONLINED))
+                        bClientUserd = NV_FALSE;
+                }
+            }
         }
     }
 
@@ -534,6 +548,9 @@ krcWatchdogInit_IMPL
         bAcquireLock = NV_FALSE;
     }
 
+    // Due to the Bug 4088184 WAR above, watchdog has to be explicitly set to internal
+    rmclientSetClientFlagsByHandle(hClient, RMAPI_CLIENT_FLAG_RM_INTERNAL_CLIENT);
+
     // Alloc device
     {
         NV0080_ALLOC_PARAMETERS *pNv0080 = &pParams->nv0080;
@@ -609,6 +626,7 @@ krcWatchdogInit_IMPL
             , {TURING_CHANNEL_GPFIFO_A,  sizeof(Nvc46fControl)}
             , {AMPERE_CHANNEL_GPFIFO_A,  sizeof(Nvc56fControl)}
             , {HOPPER_CHANNEL_GPFIFO_A,  sizeof(Nvc86fControl)}
+            , {BLACKWELL_CHANNEL_GPFIFO_A,  sizeof(Nvc96fControl)}
         };
 
         NvU32 i;
@@ -941,6 +959,7 @@ krcWatchdogInit_IMPL
         // Using device handle since VGPU doesnt support subdevice memory
         // allocations
         //
+        pMem->attr |= DRF_DEF(OS32, _ATTR, _ALLOCATE_FROM_RESERVED_HEAP, _YES);
         status = pRmApi->AllocWithHandle(pRmApi,
             hClient                    /* hClient */,
             WATCHDOG_DEVICE_ID         /* hParent */,
@@ -1164,10 +1183,10 @@ krcWatchdogInit_IMPL
     pKernelRc->watchdog.flags |= WATCHDOG_FLAGS_INITIALIZED;
 
     // Hook into the 1 Hz OS timer
-    osSchedule1SecondCallback(pGpu,
-                              krcWatchdogTimerProc,
-                              NULL /* pData */,
-                              NV_OS_1HZ_REPEAT);
+    osSchedule1HzCallback(pGpu,
+                          krcWatchdogTimerProc,
+                          NULL /* pData */,
+                          NV_OS_1HZ_REPEAT);
 
     // Schedule next interval to run immediately
     pKernelRc->watchdogPersistent.nextRunTime = 0;
@@ -1290,8 +1309,38 @@ krcWatchdogInitPushbuffer_IMPL
               NV902D_NOTIFY, NV902D_NOTIFY_TYPE_WRITE_ONLY);
     PUSH_PAIR(pKernelRc->watchdogChannelInfo.class2dSubch,
               NV902D_NO_OPERATION, 0x0);
-    PUSH_PAIR(pKernelRc->watchdogChannelInfo.class2dSubch,
-              NV906F_SET_REFERENCE, 0x0);
+
+    if (!gpuIsClassSupported(pGpu, HOPPER_CHANNEL_GPFIFO_A))
+    {
+        //
+        // TODO Bug 4588210
+        // Ticking bomb when Hopper eventually gets deprecated and
+        // HOPPER_CHANNEL_GPFIFO_A is no longer supported.
+        //
+        // When this assert starts failing, delete this branch of the if
+        // condition above.
+        //
+        NV_ASSERT(!IsHOPPERorBetter(pGpu));
+
+        PUSH_PAIR(pKernelRc->watchdogChannelInfo.class2dSubch,
+                  NV906F_SET_REFERENCE, 0x0);
+    }
+    else
+    {
+        PUSH_PAIR(pKernelRc->watchdogChannelInfo.class2dSubch, NVC86F_WFI, 0);
+        PUSH_PAIR(pKernelRc->watchdogChannelInfo.class2dSubch,
+                  NVC86F_MEM_OP_A,
+                  0);
+        PUSH_PAIR(pKernelRc->watchdogChannelInfo.class2dSubch,
+                  NVC86F_MEM_OP_B,
+                  0);
+        PUSH_PAIR(pKernelRc->watchdogChannelInfo.class2dSubch,
+                  NVC86F_MEM_OP_C,
+                  0);
+        PUSH_PAIR(pKernelRc->watchdogChannelInfo.class2dSubch,
+                  NVC86F_MEM_OP_D,
+                  DRF_DEF(C86F, _MEM_OP_D, _OPERATION, _MEMBAR));
+    }
 
     // Pushbuffer 1
     {
