@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2004-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2004-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -103,6 +103,28 @@ BRINFO upstreamPortInfo[] =
     {0, 0, NULL}
 };
 
+static NV_STATUS
+AMD_14D8_setupFunc
+(
+    OBJCL *pCl
+)
+{
+    pCl->setProperty(pCl, PDB_PROP_CL_WAR_4802761_ENABLED, NV_TRUE);
+
+    return NV_OK;
+}
+
+static NV_STATUS
+Intel_A70D_setupFunc
+(
+    OBJCL *pCl
+)
+{
+    pCl->setProperty(pCl, PDB_PROP_CL_WAR_4802761_ENABLED, NV_TRUE);
+
+    return NV_OK;
+}
+
 static NV_STATUS
 Intel_25XX_setupFunc
 (
@@ -758,15 +780,6 @@ Intel_8C4B_setupFunc
     OBJCL *pCl
 )
 {
-    switch (pCl->FHBBusInfo.deviceID)
-    {
-        case DEVICE_ID_INTEL_0C00_HASWELL_HOST_BRIDGE:
-        case DEVICE_ID_INTEL_0C04_HASWELL_HOST_BRIDGE:
-            pCl->setProperty(pCl, PDB_PROP_CL_ON_HASWELL_HOST_BRIDGE, NV_TRUE);
-            break;
-        default:
-            break;
-    }
 
     // Set ASPM L0S\L1 properties
     _Set_ASPM_L0S_L1(pCl, NV_TRUE, NV_FALSE);
@@ -903,6 +916,29 @@ Intel_7A04_setupFunc
     return NV_OK;
 }
 
+static NV_STATUS
+Intel_5795_setupFunc
+(
+    OBJCL *pCl
+)
+{
+    pCl->setProperty(pCl, PDB_PROP_CL_RELAXED_ORDERING_NOT_CAPABLE, NV_TRUE);
+
+    return NV_OK;
+}
+
+static NV_STATUS
+Intel_1B81_setupFunc
+(
+    OBJCL *pCl
+)
+{
+    pCl->setProperty(pCl, PDB_PROP_CL_RELAXED_ORDERING_NOT_CAPABLE, NV_TRUE);
+
+    return NV_OK;
+}
+
+
 static NV_STATUS
 Nvidia_T210_setupFunc
 (
@@ -1234,6 +1270,19 @@ Mellanox_BlueField_setupFunc
     return NV_OK;
 }
 
+// Mellanox BlueField3 Setup Function
+static NV_STATUS
+Mellanox_BlueField3_setupFunc
+(
+    OBJCL *pCl
+)
+{
+    // Bug 4151565: BlueField 3 does not support WC mapping 
+    pCl->setProperty(pCl, PDB_PROP_CL_DISABLE_IOMAP_WC, NV_TRUE);
+    return NV_OK;
+}
+
+
 // Amazon Gravitron2 Setup Function
 static NV_STATUS
 Amazon_Gravitron2_setupFunc
@@ -1288,6 +1337,40 @@ Arm_NeoverseN1_setupFunc
     return NV_OK;
 }
 
+static NV_STATUS
+Riscv_generic_setupFunc
+(
+    OBJCL *pCl
+)
+{
+    pCl->setProperty(pCl, PDB_PROP_CL_IS_CHIPSET_IO_COHERENT, NV_TRUE);
+    return NV_OK;
+}
+
+static NV_STATUS
+PLDA_XpressRichAXI_setupFunc
+(
+    OBJCL *pCl
+)
+{
+#if NVCPU_IS_RISCV64
+    pCl->setProperty(pCl, PDB_PROP_CL_IS_CHIPSET_IO_COHERENT, NV_TRUE);
+#endif
+    return NV_OK;
+}
+
+// Ampere AmpereOne Setup Function
+static NV_STATUS
+Ampere_AmpereOne_setupFunc
+(
+    OBJCL *pCl
+)
+{
+    // TODO Need to check if any more PDB properties should be set
+    pCl->setProperty(pCl, PDB_PROP_CL_IS_CHIPSET_IO_COHERENT, NV_TRUE);
+    return NV_OK;
+}
+
 void
 csGetInfoStrings
 (
@@ -1301,7 +1384,7 @@ csGetInfoStrings
 {
     NvU32 i;
     const char* pszUnknown = "Unknown";
-    NvU32 szUnknownLen = portStringLength(pszUnknown);
+    NvU32 szUnknownLen = portStringLength(pszUnknown) + 1;
 
     if (!pCl->chipsetIDBusAddr.valid)
     {
